/* Copyright (C) Jean-Marc Valin */
/**
   @file speex_echo.h
   @brief Echo cancellation
*/
/*
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

   1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

   3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
   ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef SPEEX_ECHO_H
#define SPEEX_ECHO_H
/** @defgroup SpeexEchoState SpeexEchoState: Acoustic echo canceller
 *  This is the acoustic echo canceller module.
 *  @{
 */
#include "speex/speex_types.h"
#include <CL/cl.h>
#include "appleOclFFT/clFFT.h"
#include <fftw3.h>

#ifdef __cplusplus
extern "C" {
#endif

/** Obtain frame size used by the AEC */
#define SPEEX_ECHO_GET_FRAME_SIZE 3

/** Set sampling rate */
#define SPEEX_ECHO_SET_SAMPLING_RATE 24
/** Get sampling rate */
#define SPEEX_ECHO_GET_SAMPLING_RATE 25

/* Can't set window sizes */
/** Get size of impulse response (int32) */
#define SPEEX_ECHO_GET_IMPULSE_RESPONSE_SIZE 27

/* Can't set window content */
/** Get impulse response (int32[]) */
#define SPEEX_ECHO_GET_IMPULSE_RESPONSE 29

/** Internal echo canceller state. Should never be accessed directly. */
struct SpeexEchoState_;

/** @class SpeexEchoState
 * This holds the state of the echo canceller. You need one per channel. 
*/

/** Internal echo canceller state. Should never be accessed directly. */
typedef struct SpeexEchoState_ SpeexEchoState;

/** Creates a new echo canceller state
 * @param frame_size Number of samples to process at one time (should correspond to 10-20 ms)
 * @param filter_length Number of samples of echo to cancel (should generally correspond to 100-500 ms)
 * @return Newly-created echo canceller state
 */
SpeexEchoState *speex_echo_state_init(int frame_size, int filter_length);

/** Creates a new multi-channel echo canceller state
 * @param frame_size Number of samples to process at one time (should correspond to 10-20 ms)
 * @param filter_length Number of samples of echo to cancel (should generally correspond to 100-500 ms)
 * @param nb_mic Number of microphone channels
 * @param nb_speakers Number of speaker channels
 * @return Newly-created echo canceller state
 */
SpeexEchoState *speex_echo_state_init_mc(int frame_size, int filter_length, int nb_mic, int nb_speakers);

/** Destroys an echo canceller state 
 * @param st Echo canceller state
*/
void speex_echo_state_destroy(SpeexEchoState *st);

int speex_echo_opencl_fft_initialize_dev(unsigned int fft_size, cl_command_queue *commandQueue);
speex_echo_opencl_fft_initialize_mem(unsigned int max_batch_size, unsigned int max_weight_update_multiplier, unsigned int fft_size, cl_command_queue *commandQueue);
void speex_echo_opencl_fft_cleanup_mem();
void speex_echo_opencl_fft_cleanup_dev();
void speex_echo_opencl_store_execution_time(double exec_time);
void speex_echo_opencl_store_execution_time_cpu(double exec_time);
void speex_echo_opencl_print_execution_time_stats();
void speex_echo_opencl_execution_time_init();
void speex_echo_opencl_execution_time_cleanup();
void set_start_reference_opencl();
void set_start_reference_cpu();

void speex_echo_opencl_fft_execute2(cl_mem input, unsigned int input_offset, cl_mem output, unsigned int output_offset, unsigned int batch_size, cl_command_queue *commandQueue);
void speex_echo_opencl_fft_update_weights(cl_mem W, cl_mem wtmp, unsigned int M, unsigned int C, unsigned int K, cl_command_queue *commandQueue);
void speex_echo_opencl_fft_convert_error(cl_mem e, cl_mem E, cl_mem y, cl_mem Y, unsigned int C, cl_command_queue *commandQueue);
void speex_echo_opencl_fft_inner_prod_combined(cl_mem data_time_domain, cl_mem data_freq_domain, cl_mem Xf, cl_mem temp_reduction, float *Sxx, unsigned int frame_size, unsigned int K, cl_command_queue *commandQueue);
void speex_echo_opencl_fft_inverse_spectral_mul_accum_reduce(cl_mem Y, cl_mem e, cl_mem temp_acc, int M, int N, int K, cl_command_queue *commandQueue);

void speex_echo_opencl_ifft_execute(cl_command_queue *commandQueue);
void speex_echo_opencl_ifft_execute2(cl_mem input, unsigned int input_offset, cl_mem output, unsigned int output_offset, unsigned int batch_size, cl_command_queue *commandQueue);
void speex_echo_opencl_ifft_prepare(float *inData, cl_command_queue *commandQueue);
void speex_echo_opencl_ifft_prepare_GPU(cl_mem input, unsigned int input_offset, cl_command_queue *commandQueue);
void speex_echo_opencl_ifft_finalize(float *output, char validate, cl_command_queue *commandQueue);
void speex_echo_opencl_ifft_finalize_GPU(cl_mem output, unsigned int output_offset, cl_command_queue *commandQueue);
void speex_echo_opencl_ifft_finalize_GPU_test(cl_mem output, unsigned int output_offset, cl_command_queue *commandQueue);

int speex_echo_opencl_copy_to_GPU(SpeexEchoState *st, const spx_int16_t *in, const spx_int16_t *far_end, spx_int16_t *out);
int speex_echo_opencl_copy_to_CPU(SpeexEchoState *st);
void speex_echo_opencl_sync_to_GPU(SpeexEchoState *st);
void speex_echo_opencl_sync_to_CPU(SpeexEchoState *st);

/** Performs echo cancellation a frame, based on the audio sent to the speaker (no delay is added
 * to playback in this form)
 *
 * @param st Echo canceller state
 * @param rec Signal from the microphone (near end + far end echo)
 * @param play Signal played to the speaker (received from far end)
 * @param out Returns near-end signal with echo removed
 */
void speex_echo_cancellation(SpeexEchoState *st, const spx_int16_t *rec, const spx_int16_t *play, spx_int16_t *out);

void speex_echo_cancellation_opencl(SpeexEchoState *st, const spx_int16_t *in, const spx_int16_t *far_end, spx_int16_t *out);
double get_residual_error(SpeexEchoState *st1, SpeexEchoState *st2);

void speex_echo_opencl_init_dev(int frame_size);
void speex_echo_opencl_init_mem(SpeexEchoState *st);
void speex_echo_opencl_cleanup_mem();
void speex_echo_opencl_cleanup_dev();

typedef struct
{
        double *real;
        double *imag;
}clFFT_SplitComplexDouble;

typedef struct
{
        double real;
        double imag;
}clFFT_ComplexDouble;

int initializeCL(cl_command_queue *commandQueue);
void runCLKernels(void);
int convertToString(const char *filename, char *str, size_t *length);
int getFileSize(const char *filename);
void speex_echo_opencl_cleanup_common_mem(cl_command_queue *commandQueue);
void speex_echo_opencl_cleanup_common_dev(cl_command_queue *commandQueue);
double computeL2ErrorReal(const float *data_ref, const float *data, int n, int batchSize, double *max_diff, double *min_diff);
double computeL2ErrorComplex(const float *data_ref, const float *data, int n, int batchSize, double *max_diff, double *min_diff);
void real_dft_negative_frequency_generation(clFFT_Complex *complex_data, int num_points);
void print_debug(char *string, int level);
void print_on_opencl_error(char *string, cl_int err);
void print_error(char *string, int stop_condition);
void print_opencl_error(cl_int error);
void print_opencl_build_info(cl_program *program);
void init_timingpoints();
double print_and_reset_timingpoints(const char *title);
int save_new_timingpoint(const char *desc);
void cleanup_timingpoints();
void init_checkpoints();
void save_new_checkpoint(int index, const float *data, int size);
void check_against_checkpoint(int index, const void *data);
void check_against_checkpoint_opencl(int index, cl_mem cl_data, size_t cl_data_offset, int input_is_int);
void cleanup_checkpoints();
void copy_echo_state(SpeexEchoState *input, SpeexEchoState *output);

/** Performs echo cancellation a frame (deprecated) */
void speex_echo_cancel(SpeexEchoState *st, const spx_int16_t *rec, const spx_int16_t *play, spx_int16_t *out, spx_int32_t *Yout);

/** Perform echo cancellation using internal playback buffer, which is delayed by two frames
 * to account for the delay introduced by most soundcards (but it could be off!)
 * @param st Echo canceller state
 * @param rec Signal from the microphone (near end + far end echo)
 * @param out Returns near-end signal with echo removed
*/
void speex_echo_capture(SpeexEchoState *st, const spx_int16_t *rec, spx_int16_t *out);

/** Let the echo canceller know that a frame was just queued to the soundcard
 * @param st Echo canceller state
 * @param play Signal played to the speaker (received from far end)
*/
void speex_echo_playback(SpeexEchoState *st, const spx_int16_t *play);

/** Reset the echo canceller to its original state 
 * @param st Echo canceller state
 */
void speex_echo_state_reset(SpeexEchoState *st);

/** Used like the ioctl function to control the echo canceller parameters
 *
 * @param st Echo canceller state
 * @param request ioctl-type request (one of the SPEEX_ECHO_* macros)
 * @param ptr Data exchanged to-from function
 * @return 0 if no error, -1 if request in unknown
 */
int speex_echo_ctl(SpeexEchoState *st, int request, void *ptr);



struct SpeexDecorrState_;

typedef struct SpeexDecorrState_ SpeexDecorrState;


/** Create a state for the channel decorrelation algorithm
    This is useful for multi-channel echo cancellation only 
 * @param rate Sampling rate
 * @param channels Number of channels (it's a bit pointless if you don't have at least 2)
 * @param frame_size Size of the frame to process at ones (counting samples *per* channel)
*/
SpeexDecorrState *speex_decorrelate_new(int rate, int channels, int frame_size);

/** Remove correlation between the channels by modifying the phase and possibly
    adding noise in a way that is not (or little) perceptible.
 * @param st Decorrelator state
 * @param in Input audio in interleaved format
 * @param out Result of the decorrelation (out *may* alias in)
 * @param strength How much alteration of the audio to apply from 0 to 100.
*/
void speex_decorrelate(SpeexDecorrState *st, const spx_int16_t *in, spx_int16_t *out, int strength);

/** Destroy a Decorrelation state 
 * @param st State to destroy
*/
void speex_decorrelate_destroy(SpeexDecorrState *st);

#ifdef __cplusplus
}
#endif


/** @}*/
#endif
