
//
// File:       main.cpp
//
// Version:    <1.0>
//
// Disclaimer: IMPORTANT:  This Apple software is supplied to you by Apple Inc. ("Apple")
//             in consideration of your agreement to the following terms, and your use,
//             installation, modification or redistribution of this Apple software
//             constitutes acceptance of these terms.  If you do not agree with these
//             terms, please do not use, install, modify or redistribute this Apple
//             software.
//
//             In consideration of your agreement to abide by the following terms, and
//             subject to these terms, Apple grants you a personal, non - exclusive
//             license, under Apple's copyrights in this original Apple software ( the
//             "Apple Software" ), to use, reproduce, modify and redistribute the Apple
//             Software, with or without modifications, in source and / or binary forms;
//             provided that if you redistribute the Apple Software in its entirety and
//             without modifications, you must retain this notice and the following text
//             and disclaimers in all such redistributions of the Apple Software. Neither
//             the name, trademarks, service marks or logos of Apple Inc. may be used to
//             endorse or promote products derived from the Apple Software without specific
//             prior written permission from Apple.  Except as expressly stated in this
//             notice, no other rights or licenses, express or implied, are granted by
//             Apple herein, including but not limited to any patent rights that may be
//             infringed by your derivative works or by other works in which the Apple
//             Software may be incorporated.
//
//             The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
//             WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
//             WARRANTIES OF NON - INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
//             PARTICULAR PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION
//             ALONE OR IN COMBINATION WITH YOUR PRODUCTS.
//
//             IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
//             CONSEQUENTIAL DAMAGES ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
//             SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
//             INTERRUPTION ) ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION
//             AND / OR DISTRIBUTION OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER
//             UNDER THEORY OF CONTRACT, TORT ( INCLUDING NEGLIGENCE ), STRICT LIABILITY OR
//             OTHERWISE, EVEN IF APPLE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Copyright ( C ) 2008 Apple Inc. All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////////////////////////


#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <cstring>
#include <CL/cl.h>
#include "clFFT.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <float.h>
#include <fftw3.h>
#ifdef LINUX
#include <sys/time.h>
#endif
#define eps_avg 10.0

#define MAX( _a, _b)	((_a)>(_b)?(_a) : (_b))

#define test_start()
#define log_perf(_number, _higherBetter, _numType, _format, ...) printf("Performance Number " _format " (in %s, %s): %g\n",##__VA_ARGS__, _numType, _higherBetter?"higher is better":"lower is better" , _number)
#define log_info printf
#define log_error printf
#define test_finish()

typedef enum {
	clFFT_OUT_OF_PLACE,
	clFFT_IN_PLACE,
}clFFT_TestType;

typedef struct
{
	double real;
	double imag;
}clFFT_ComplexDouble;

typedef struct
{
	double *real;
	double *imag;
}clFFT_SplitComplexDouble;

cl_device_id     device_id;
cl_context       context;
cl_command_queue queue;
clFFT_SplitComplexDouble data_iref;
fftwf_complex *data_oref_reverse;
typedef unsigned long long ulongFFT;
double computeL2ErrorFFTW(fftwf_complex *data, clFFT_SplitComplexDouble *data_ref, int n, int batchSize, double *max_diff, double *min_diff);

#ifdef LINUX
// Returns the time difference in seconds
double subtractTimes(timeval end, timeval start)
{
	long long tusecs = (long long) (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec);
	return (double)(tusecs/1000000.0);
}
#endif

void computeReferenceF(fftwf_complex *out, clFFT_Dim3 n, 
					  unsigned int batchSize, clFFT_Dimension dim, clFFT_Direction dir)
{
	fftwf_plan p;
	int fftwf_dir;
	int fftwf_batchSize = batchSize;

	if(clFFT_Forward) {
		fftwf_dir = FFTW_FORWARD;
	}
	else {
		fftwf_dir = FFTW_BACKWARD;
	}
	
	unsigned int log2Nx = (int) log2(n.x);
	unsigned int log2Ny = (int) log2(n.y);
	unsigned int log2Nz = (int) log2(n.z);
	unsigned int log2N;
	
	log2N = log2Nx;
	log2N = log2N > log2Ny ? log2N : log2Ny;
	log2N = log2N > log2Nz ? log2N : log2Nz;

	int fftwf_nx = n.x;
	int fftwf_ny = n.y;
	int fftwf_nz = n.z;
	unsigned int i = 0;	
	unsigned int j = 0;

	double diff_avg, diff_max, diff_min;

	diff_avg = computeL2ErrorFFTW(out, &data_iref, fftwf_nx*fftwf_ny*fftwf_nz, batchSize, &diff_max, &diff_min);
	if(diff_avg > eps_avg)
		log_error("Pre-test of FFTW vs. CL input data failed (n=(%d, %d, %d), batchsize=%d): %s Test: rel. L2-error = %f eps (max=%f eps, min=%f eps)\n", n.x, n.y, n.z, batchSize, "out-of-place", diff_avg, diff_max, diff_min);
	else
		log_info("Pre-test of FFTW vs. CL input data passed (n=(%d, %d, %d), batchsize=%d): %s Test: rel. L2-error = %f eps (max=%f eps, min=%f eps)\n", n.x, n.y, n.z, batchSize, "out-of-place", diff_avg, diff_max, diff_min);

	switch(dim)
	{
		case clFFT_1D:	
			// Batch method
			p = fftwf_plan_many_dft(1, &fftwf_nx, fftwf_batchSize, 
						out, NULL, 
						1, fftwf_nx, 
						out, NULL, 
						1, fftwf_nx, 
						fftwf_dir, FFTW_ESTIMATE);
			if(p) {
				fftwf_execute(p);
				fftwf_destroy_plan(p);
			}
			else {
				log_error("Failed to create forward DFT plan.\n");
				break;
			}
		
			// Correct unnormalized output data of the forward transform, if reversed
			if(!clFFT_Forward) {
				for(j = 0; j < batchSize; j++) {
					for(i = 0; i < n.x; i++) {
						unsigned int index = j * fftwf_nx + i;
						out[index][0] /= fftwf_nx;
						out[index][1] /= fftwf_nx;
					}
				}
			}		
	
			p = fftwf_plan_many_dft(1, &fftwf_nx, fftwf_batchSize, 
						out, NULL, 
						1, fftwf_nx, 
						data_oref_reverse, NULL, 
						1, fftwf_nx, 
						-fftwf_dir, FFTW_ESTIMATE);
			if(p) {
				fftwf_execute(p);
				fftwf_destroy_plan(p);
			}
			else {
				log_error("Failed to create reverse DFT plan.\n");
				break;
			}

			// Correct unnormalized output of the reverse transform, if forward direction
			if(clFFT_Forward) {
				for(j = 0; j < batchSize; j++) {
					for(i = 0; i < n.x; i++) {
						unsigned int index = j * fftwf_nx + i;
						data_oref_reverse[index][0] /= fftwf_nx;
						data_oref_reverse[index][1] /= fftwf_nx;
					}
				}
			}
			
			diff_avg = computeL2ErrorFFTW(data_oref_reverse, &data_iref, fftwf_nx, batchSize, &diff_max, &diff_min);
			if(diff_avg > eps_avg)
				log_error("Reverse transform test failed (n=(%d, %d, %d), batchsize=%d): %s Test: rel. L2-error = %f eps (max=%f eps, min=%f eps)\n", n.x, n.y, n.z, batchSize, "out-of-place", diff_avg, diff_max, diff_min);
			else
				log_info("Reverse transform test passed (n=(%d, %d, %d), batchsize=%d): %s Test: rel. L2-error = %f eps (max=%f eps, min=%f eps)\n", n.x, n.y, n.z, batchSize, "out-of-place", diff_avg, diff_max, diff_min);

			break;			
		case clFFT_2D: 
			printf("2D transforms not supported with this version.\n");
			exit(1);
			break;
		case clFFT_3D:
			printf("3D transforms not supported with this version.\n");
			exit(1);
			break;
	}
	
}

double complexNormSq(clFFT_ComplexDouble a)
{
	return (a.real * a.real + a.imag * a.imag);
}

double computeL2ErrorFFTW(fftwf_complex *data, clFFT_SplitComplexDouble *data_ref, int n, int batchSize, double *max_diff, double *min_diff)
{
	int i, j;
	double avg_norm = 0.0;
	*max_diff = 0.0;
	*min_diff = 0x1.0p1000;
	
	for(j = 0; j < batchSize; j++)
	{
		double norm_ref = 0.0;
		double norm = 0.0;
	    for(i = 0; i < n; i++) 
		{
			int index = j * n + i;
		    clFFT_ComplexDouble diff = (clFFT_ComplexDouble) { data_ref->real[index] - data[index][0], data_ref->imag[index] - data[index][1] };
		    double norm_tmp = complexNormSq(diff);
		    norm += norm_tmp;
		    norm_ref += (data_ref->real[index] * data_ref->real[index] + data_ref->imag[index] * data_ref->imag[index]);
	    }
	    double curr_norm = sqrt( norm / norm_ref ) / FLT_EPSILON;
		avg_norm += curr_norm;
		*max_diff = *max_diff < curr_norm ? curr_norm : *max_diff;
		*min_diff = *min_diff > curr_norm ? curr_norm : *min_diff;
	}
	
	return avg_norm / batchSize;
}

double computeL2Error(clFFT_SplitComplex *data, fftwf_complex *data_ref, int n, int batchSize, double *max_diff, double *min_diff)
{
	int i, j;
	double avg_norm = 0.0;
	*max_diff = 0.0;
	*min_diff = 0x1.0p1000;
	
	for(j = 0; j < batchSize; j++)
	{
		double norm_ref = 0.0;
		double norm = 0.0;
	    for(i = 0; i < n; i++) 
		{
			int index = j * n + i;
		    clFFT_ComplexDouble diff = (clFFT_ComplexDouble) { data_ref[index][0] - data->real[index], data_ref[index][1] - data->imag[index] };
		    double norm_tmp = complexNormSq(diff);
		    norm += norm_tmp;
		    norm_ref += (data_ref[index][0] * data_ref[index][0] + data_ref[index][1] * data_ref[index][1]);
	    }
	    double curr_norm = sqrt( norm / norm_ref ) / FLT_EPSILON;
		avg_norm += curr_norm;
		*max_diff = *max_diff < curr_norm ? curr_norm : *max_diff;
		*min_diff = *min_diff > curr_norm ? curr_norm : *min_diff;
	}
	
	return avg_norm / batchSize;
}

void convertInterleavedToSplit(clFFT_SplitComplex *result_split, clFFT_Complex *data_cl, int length)
{
	int i;
	for(i = 0; i < length; i++) {
		result_split->real[i] = data_cl[i].real;
		result_split->imag[i] = data_cl[i].imag;
	}
}

int runTest(clFFT_Dim3 n, int batchSize, clFFT_Direction dir, clFFT_Dimension dim, 
			clFFT_DataFormat dataFormat, int numIter, clFFT_TestType testType)
{	
	cl_int err = CL_SUCCESS;
	int iter;

#ifdef LINUX
	double t;
	int mx = log2(n.x);
	int my = log2(n.y);
	int mz = log2(n.z);

	struct timeval start, end;
	double gflops = 5e-9 * ((double)mx + (double)my + (double)mz) * (double)n.x * (double)n.y * (double)n.z * (double)batchSize * (double)numIter;
#endif
	int length = n.x * n.y * n.z * batchSize;
	
	clFFT_SplitComplex data_i_split = (clFFT_SplitComplex) { NULL, NULL };
	clFFT_SplitComplex data_cl_split = (clFFT_SplitComplex) { NULL, NULL };
	data_iref = (clFFT_SplitComplexDouble) { NULL, NULL }; 
	fftwf_complex *data_oref = NULL;
	data_oref_reverse = NULL;
	
	clFFT_Plan plan = NULL;
	cl_mem data_in = NULL;
	cl_mem data_out = NULL;
	cl_mem data_in_real = NULL;
	cl_mem data_in_imag = NULL;
	cl_mem data_out_real = NULL;
	cl_mem data_out_imag = NULL;
	
	if(dataFormat == clFFT_SplitComplexFormat) {
		data_i_split.real     = (float *) malloc(sizeof(float) * length);
		data_i_split.imag     = (float *) malloc(sizeof(float) * length);
		data_cl_split.real    = (float *) malloc(sizeof(float) * length);
		data_cl_split.imag    = (float *) malloc(sizeof(float) * length);
		if(!data_i_split.real || !data_i_split.imag || !data_cl_split.real || !data_cl_split.imag)
		{
			err = -1;
			log_error("Out-of-Resources\n");
			goto cleanup;
		}
	}	
	
	data_iref.real   = (double *) malloc(sizeof(double) * length);
	data_iref.imag   = (double *) malloc(sizeof(double) * length);
	data_oref = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * length);
	data_oref_reverse = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * length);

	if(!data_iref.real || !data_iref.imag || !data_oref)
	{
		err = -3;
		log_error("Out-of-Resources\n");
		goto cleanup;
	}

	int i;
	if(dataFormat == clFFT_SplitComplexFormat) {
		for(i = 0; i < length; i++)
		{
			data_i_split.real[i] = 2.0f * (float) rand() / (float) RAND_MAX - 1.0f;
			data_i_split.imag[i] = 2.0f * (float) rand() / (float) RAND_MAX - 1.0f;
			data_cl_split.real[i] = 0.0f;
			data_cl_split.imag[i] = 0.0f;			
			data_iref.real[i] = data_i_split.real[i];
			data_iref.imag[i] = data_i_split.imag[i];
			data_oref[i][0] = data_iref.real[i];
			data_oref[i][1] = data_iref.imag[i];
		}
	}
	
	plan = clFFT_CreatePlan( context, n, dim, dataFormat, &err );
	if(!plan || err) 
	{
		log_error("clFFT_CreatePlan failed\n");
		goto cleanup;
	}
	
	if(dataFormat == clFFT_SplitComplexFormat)
	{
		data_in_real = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length*sizeof(float), data_i_split.real, &err);
	    if(!data_in_real || err) 
	    {
			log_error("clCreateBuffer failed\n");
			goto cleanup;
	    }
		
		data_in_imag = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length*sizeof(float), data_i_split.imag, &err);
	    if(!data_in_imag || err) 
	    {
			log_error("clCreateBuffer failed\n");
			goto cleanup;
	    }
		
		if(testType == clFFT_OUT_OF_PLACE)
		{
			data_out_real = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length*sizeof(float), data_cl_split.real, &err);
			if(!data_out_real || err) 
			{
				log_error("clCreateBuffer failed\n");
				goto cleanup;
			}
			
			data_out_imag = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length*sizeof(float), data_cl_split.imag, &err);
			if(!data_out_imag || err) 
			{
				log_error("clCreateBuffer failed\n");
				goto cleanup;
			}			
		}
		else
		{
			data_out_real = data_in_real;
			data_out_imag = data_in_imag;
		}
	}		
			
	err = CL_SUCCESS;
	
	log_info("Starting OpenCL execution!\n");
#ifdef LINUX
	gettimeofday(&start, NULL);
#endif

	if(dataFormat == clFFT_SplitComplexFormat)
	{
		for(iter = 0; iter < numIter; iter++)
		    err |= clFFT_ExecutePlannar(queue, plan, batchSize, dir, data_in_real, data_in_imag, data_out_real, data_out_imag, 0, NULL, NULL);
	}
	else
	{
	    for(iter = 0; iter < numIter; iter++) 
			err |= clFFT_ExecuteInterleaved(queue, plan, batchSize, dir, data_in, data_out, 0, NULL, NULL);
	}
	
	err |= clFinish(queue);
	
	if(err) 
	{
		log_error("clFFT_Execute\n");
		goto cleanup;	
	}

#ifdef LINUX	
	gettimeofday(&end, NULL);
	log_info("Stopped OpenCL execution!\n");
	t = subtractTimes(end, start);
	char temp[100];
	sprintf(temp, "GFlops achieved for n = (%d, %d, %d), batchsize = %d", n.x, n.y, n.z, batchSize);
	log_perf(gflops / (float) t, 1, "GFlops/s", "%s", temp);

	log_info("Computation time on GPU (seconds): %f\n", t);
#else
	log_info("Timing of GPU computation disabled (OS support not available)\n");
#endif

	if(dataFormat == clFFT_SplitComplexFormat)
	{	
		err |= clEnqueueReadBuffer(queue, data_out_real, CL_TRUE, 0, length*sizeof(float), data_cl_split.real, 0, NULL, NULL);
		err |= clEnqueueReadBuffer(queue, data_out_imag, CL_TRUE, 0, length*sizeof(float), data_cl_split.imag, 0, NULL, NULL);
	}	
	
	if(err) 
	{
		log_error("clEnqueueReadBuffer failed\n");
        goto cleanup;
	}	

	log_info("Starting tests.\n");

#ifdef LINUX	
	gettimeofday(&start, NULL);
#endif
	computeReferenceF(data_oref, n, batchSize, dim, dir);
#ifdef LINUX
	gettimeofday(&end, NULL);
	t = subtractTimes(end, start);
	log_info("Computation time on CPU of reference (useconds): %f\n", t);
#else
	log_info("Timing of CPU computation disabled (OS support not available)\n");
#endif

        /*unsigned int k;
        printf("CL output:\n");
        for(k=0;k<n.x;k++)
        {
                printf("%d: %f\t", k*2, data_cl_split.real[k]);
                printf("%d: %f\t", k*2+1, data_cl_split.imag[k]);
                if(k % 5 == 0)
                printf("\n");
        }
        printf("\nReference output:\n");
        for(k=0;k<n.x;k++)
        {
                printf("%d: %f\t", k*2, data_oref[k][0]);
                printf("%d: %f\t", k*2+1, data_oref[k][1]);
                if(k % 5 == 0)
                printf("\n");
        }
        printf("\n");*/

	double diff_avg, diff_max, diff_min;
	if(dataFormat == clFFT_SplitComplexFormat) {
		diff_avg = computeL2Error(&data_cl_split, data_oref, n.x*n.y*n.z, batchSize, &diff_max, &diff_min);
		if(diff_avg > eps_avg)
			log_error("Test failed (n=(%d, %d, %d), batchsize=%d): %s Test: rel. L2-error = %f eps (max=%f eps, min=%f eps)\n", n.x, n.y, n.z, batchSize, (testType == clFFT_OUT_OF_PLACE) ? "out-of-place" : "in-place", diff_avg, diff_max, diff_min);
		else
			log_info("Test passed (n=(%d, %d, %d), batchsize=%d): %s Test: rel. L2-error = %f eps (max=%f eps, min=%f eps)\n", n.x, n.y, n.z, batchSize, (testType == clFFT_OUT_OF_PLACE) ? "out-of-place" : "in-place", diff_avg, diff_max, diff_min);			
	}

	log_info("Tests done, starting cleanup.\n");
cleanup:
	clFFT_DestroyPlan(plan);	
	if(dataFormat == clFFT_SplitComplexFormat) 
	{
		if(data_i_split.real)
			free(data_i_split.real);
		if(data_i_split.imag)
			free(data_i_split.imag);
		if(data_cl_split.real)
			free(data_cl_split.real);
		if(data_cl_split.imag)
			free(data_cl_split.imag);
		
		if(data_in_real)
			clReleaseMemObject(data_in_real);
		if(data_in_imag)
			clReleaseMemObject(data_in_imag);
		if(data_out_real && testType == clFFT_OUT_OF_PLACE)
			clReleaseMemObject(data_out_real);
		if(data_out_imag && clFFT_OUT_OF_PLACE)
			clReleaseMemObject(data_out_imag);
	}
	
	if(data_iref.real)
		free(data_iref.real);
	if(data_iref.imag)
		free(data_iref.imag);		
	if(data_oref) 
		fftwf_free(data_oref);
	
	return err;
}

bool ifLineCommented(const char *line) {
	const char *Line = line;
	while(*Line != '\0')
		if((*Line == '/') && (*(Line + 1) == '/'))
			return true;
		else
			Line++;
	return false;
}

cl_device_type getGlobalDeviceType()
{
	char *force_cpu = getenv( "CL_DEVICE_TYPE" );
	if( force_cpu != NULL )
	{
		if( strcmp( force_cpu, "gpu" ) == 0 || strcmp( force_cpu, "CL_DEVICE_TYPE_GPU" ) == 0 )
			return CL_DEVICE_TYPE_GPU;
		else if( strcmp( force_cpu, "cpu" ) == 0 || strcmp( force_cpu, "CL_DEVICE_TYPE_CPU" ) == 0 )
			return CL_DEVICE_TYPE_CPU;
		else if( strcmp( force_cpu, "accelerator" ) == 0 || strcmp( force_cpu, "CL_DEVICE_TYPE_ACCELERATOR" ) == 0 )
			return CL_DEVICE_TYPE_ACCELERATOR;
		else if( strcmp( force_cpu, "CL_DEVICE_TYPE_DEFAULT" ) == 0 )
			return CL_DEVICE_TYPE_DEFAULT;
	}
	// default
	return CL_DEVICE_TYPE_GPU;
}

void 
notify_callback(const char *errinfo, const void *private_info, size_t cb, void *user_data)
{
    log_error( "%s\n", errinfo );
}

int
checkMemRequirements(clFFT_Dim3 n, int batchSize, clFFT_TestType testType, cl_ulong gMemSize)
{
	cl_ulong memReq = (testType == clFFT_OUT_OF_PLACE) ? 3 : 2;
	memReq *= n.x*n.y*n.z*sizeof(clFFT_Complex)*batchSize;
	memReq = memReq/1024/1024;
	if(memReq >= gMemSize)
		return -1;
	return 0;
}

int main (int argc, char * const argv[]) {
	
	test_start();
	
	cl_ulong gMemSize;
	clFFT_Direction dir = clFFT_Forward;
	int numIter = 1;
	clFFT_Dim3 n = { 1024, 1, 1 };
	int batchSize = 1;
	clFFT_DataFormat dataFormat = clFFT_SplitComplexFormat;
	clFFT_Dimension dim = clFFT_1D;
	clFFT_TestType testType = clFFT_OUT_OF_PLACE;
	
	FILE *paramFile;
			
	cl_int err, status;
	//unsigned int num_devices;

	log_info("The value of CL_PLATFORM_VENDOR: %d\n", CL_PLATFORM_VENDOR);
	
	/*cl_device_type device_type = getGlobalDeviceType();	
	if(device_type != CL_DEVICE_TYPE_GPU) 
	{
		log_info("Test only supported on DEVICE_TYPE_GPU\n");
		test_finish();
		exit(0);
	}*/
	cl_device_type device_type = CL_DEVICE_TYPE_GPU;

	// Modified to work on ATI
	
	cl_uint numPlatforms;
	cl_platform_id platform = NULL;
        cl_context_properties cps[3];
        cl_device_id device_id;
        size_t deviceListSize;
        //cl_device_id        device;
        cl_device_id    *devices;

/*	err = clGetPlatformIDs(20, &platform, &numPlatforms);
	if(err) 
	{		
		log_error("clGetPlatformIDs failed\n");
		test_finish();
		return -1;
	}

	// Modification STOP
	
	err = clGetDeviceIDs(platform, device_type, 1, &device_id, &num_devices);
	if(err) 
	{		
		log_error("clGetComputeDevice failed, errorcode: %d\n", err);
		test_finish();
		return -1;
	}
	
	context = clCreateContext(0, 1, &device_id, notify_callback, NULL, &err);
	if(!context || err) 
	{
		log_error("clCreateContext failed\n");
		test_finish();
		return -1;
	}
	
    queue = clCreateCommandQueue(context, device_id, 0, &err);
    if(!queue || err)
	{
        log_error("clCreateQueue() failed.\n");
		clReleaseContext(context);
        test_finish();
        return -1;
    } */ 
	
        status = clGetPlatformIDs(0, NULL, &numPlatforms);
        if(status != CL_SUCCESS)
        {
                printf("clGetPlatformIDs failed\n");
                return -1;
        }

        cl_platform_id* platforms = (cl_platform_id* )malloc(sizeof(cl_platform_id)*numPlatforms);
        
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(status)
        {
                printf("clGetPlatformIDs failed\n");
                return -1;
        }
        else 
        {
                printf("Got OpenCL platform IDs. Numplatforms: %d\n", numPlatforms);
        }

        status = CL_SUCCESS;
        unsigned int contextTries = 0;
        // Loop through platforms until we find the first with the required device type (at random)
        // Made for supporting a combined AMD/NVIDIA setup with only one GPU
        // Customize if using a multiple GPU setup
        do
        {
                platform = platforms[contextTries];
                cps[0] = CL_CONTEXT_PLATFORM;
                cps[1] = (cl_context_properties)platform;
                cps[2] = 0;
                // Create an OpenCL context
                context = clCreateContextFromType(cps,
                                              device_type,
                                              NULL,
                                              NULL,
                                              &status);
                contextTries++;
        }
        while(contextTries < numPlatforms && status != CL_SUCCESS);
          
        if(status != CL_SUCCESS)
        {
                printf("Error: Creating Context. (clCreateContextFromType)\n");
                return -1;
        }

        // First, get the size of device list data
        status = clGetContextInfo(context,
                              CL_CONTEXT_DEVICES,
                              0,
                              NULL,
                              &deviceListSize);
        if(status != CL_SUCCESS)
        {
                printf("Error: Getting Context Info (device list size, clGetContextInfo)\n");
                return -1;
        }
        
        // Detect OpenCL devices
        devices = (cl_device_id *)malloc(deviceListSize);
        if(devices == 0)
        {
                printf("Error: No devices found.\n");
                return -1;
        }

        /* Now, get the device list data */
        status = clGetContextInfo(
                             context,
                 CL_CONTEXT_DEVICES,
                 deviceListSize,
                 devices,
                 NULL);
        if(status != CL_SUCCESS)
        {
                printf("Error: Getting Context Info (device list, clGetContextInfo)\n");
                return -1;
        }
        device_id = devices[0];

        // Create an OpenCL command queue
        queue = clCreateCommandQueue(
                        context,
                        devices[0],
                        0,
                        &status);
        if(status != CL_SUCCESS)
        {
                printf("Error: Creating Command Queue. (clCreateCommandQueue)\n");
                return -1;
        }
        else
        {
                printf("SUCCESS: Creating Command Queue. (clCreateCommandQueue)\n");
        }

	err = clGetDeviceInfo(device_id, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong), &gMemSize, NULL);
	if(err)
	{
		log_error("Failed to get global mem size\n");
		clReleaseContext(context);
		clReleaseCommandQueue(queue);
		test_finish();
		return -2;
	}
	
	gMemSize /= (1024*1024);
			
	char delim[] = " \n";
	char tmpStr[100];
	char line[200];
	char *param, *val;	
	int total_errors = 0;
	if(argc == 1) {
		log_error("Need file name with list of parameters to run the test\n");
		test_finish();
		return -1;
	}
	
	if(argc == 2) {	// arguments are supplied in a file with arguments for a single run are all on the same line
		paramFile = fopen(argv[1], "r");
		if(!paramFile) {
			log_error("Cannot open the parameter file\n");
			clReleaseContext(context);
			clReleaseCommandQueue(queue);			
			test_finish();
			return -3;
		}
		while(fgets(line, 199, paramFile)) {
			if(!strcmp(line, "") || !strcmp(line, "\n") || ifLineCommented(line))
				continue;
			param = strtok(line, delim);
			while(param) {
				val = strtok(NULL, delim);
				if(!strcmp(param, "-n")) {
					sscanf(val, "%d", &n.x);
					val = strtok(NULL, delim);
					sscanf(val, "%d", &n.y);
					val = strtok(NULL, delim);
					sscanf(val, "%d", &n.z);					
				}
				else if(!strcmp(param, "-batchsize")) 
					sscanf(val, "%d", &batchSize);
				else if(!strcmp(param, "-dir")) {
					sscanf(val, "%s", tmpStr);
					if(!strcmp(tmpStr, "forward"))
						dir = clFFT_Forward;
					else if(!strcmp(tmpStr, "inverse"))
						dir = clFFT_Inverse;
				}
				else if(!strcmp(param, "-dim")) {
					sscanf(val, "%s", tmpStr);
					if(!strcmp(tmpStr, "1D"))
						dim = clFFT_1D;
					else if(!strcmp(tmpStr, "2D"))
						dim = clFFT_2D; 
					else if(!strcmp(tmpStr, "3D"))
						dim = clFFT_3D;					
				}
				else if(!strcmp(param, "-format")) {
					sscanf(val, "%s", tmpStr);
					if(!strcmp(tmpStr, "plannar"))
						dataFormat = clFFT_SplitComplexFormat;
					else if(!strcmp(tmpStr, "interleaved"))
						dataFormat = clFFT_InterleavedComplexFormat;					
				}
				else if(!strcmp(param, "-numiter"))
					sscanf(val, "%d", &numIter);
				else if(!strcmp(param, "-testtype")) {
					sscanf(val, "%s", tmpStr);
					if(!strcmp(tmpStr, "out-of-place"))
						testType = clFFT_OUT_OF_PLACE;
					else if(!strcmp(tmpStr, "in-place"))
						testType = clFFT_IN_PLACE;										
				}
				param = strtok(NULL, delim);
			}
			
			if(checkMemRequirements(n, batchSize, testType, gMemSize)) {
				log_info("This test cannot run because memory requirements canot be met by the available device\n");
				continue;
			}
				
			err = runTest(n, batchSize, dir, dim, dataFormat, numIter, testType);
			if (err)
				total_errors++;
		}
	}
	
	clReleaseContext(context);
	clReleaseCommandQueue(queue);
	
	test_finish();
	return total_errors;		
}
