/* Copyright (C) 2004 Jean-Marc Valin
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// Enables some tricks (split weight update) to work on ATI
// Potentially better performance on NVIDIA without
//#define ATI_GPU 1

struct fftw_config {
  float *in;
  float *out;
  fftwf_plan fft;
  fftwf_plan ifft;
  int N;
};

float *FFT_data_input_host = NULL;
float *FFT_data_output_host = NULL;

cl_mem data_in = NULL;
cl_mem data_out = NULL;
cl_mem data_in2 = NULL;
cl_mem data_out2 = NULL;
// Created as separate buffers because of their potentially larger size
cl_mem data_in_weights = NULL;	
cl_mem data_out_weights = NULL;

cl_program clFFT_extras_program;
cl_kernel FFT_custom_kernel;
cl_kernel FFT_update_weights_kernel;
cl_kernel FFT_update_weights_ati_pt1_kernel;
cl_kernel FFT_update_weights_ati_pt2_kernel;
cl_kernel FFT_convert_error_to_freq_domain_kernel;
cl_kernel FFT_inner_prod_combined_kernel;
cl_kernel FFT_inverse_spectral_mul_accum_reduce_kernel;

int numClFFTIter;
int clFFTBatchSize;
int clFFTLength;
int clIFFTLength;

char *print_string;

/**
 * Reads kernel source code from file, must be run before memory is allocated
 * @arg max_batch_size maximum batch size for the regular FFT functions
 * @arg max_weight_update_multiplier maximum number of C*K for weight update function
 */
EXPORT int speex_echo_opencl_fft_initialize_dev(unsigned int fft_size, cl_command_queue *commandQueue)
{
	// String used for passing information to the utility functions
	print_string = (char *)malloc(sizeof(char) * 300);
	
        print_debug("Running speex_echo_opencl_fft_initialize.", 3);

        clFFT_Dim3 n = { fft_size, 1, 1 };
        clFFT_Dim3 iN = { fft_size, 1, 1 };
        numClFFTIter = 1;
        clFFTBatchSize = 1;
        clFFTLength = n.x + 2;
        clIFFTLength = iN.x / 2;

        cl_int err = CL_SUCCESS;

        //int mx = log2(n.x);
        //int my = log2(n.y);
        //int mz = log2(n.z);

        // Load IFFT prepare kernel
	char *clFFT_extras_kernel_filename = (char *)malloc(sizeof(char) * 50);
	if(fft_size == 256)
	{
		clFFT_extras_kernel_filename = "mdf_oclfft_kernel_256.cl";
	}
	else if(fft_size == 1024)
	{
		clFFT_extras_kernel_filename = "mdf_oclfft_kernel_1024.cl";
		printf("WARNING: Reading FFT source code for window size 1024, this can take some time using the NVIDIA OpenCL implementation.\n");
	}
	else
	{
		print_error("Frame size is unsupported by OpenCL FFT implementation (only 128 or 512 supported).\n", 0);
		return -1;
	}
	printf("Opening kernel file: %s\n", clFFT_extras_kernel_filename);
	
	int clFFT_extras_kernel_filesize = getFileSize(clFFT_extras_kernel_filename);

	if(clFFT_extras_kernel_filesize < 0) {
		sprintf(print_string, "Could not find file with FFT pre-/pre-processing kernels (in %s), exiting.\n", clFFT_extras_kernel_filename);
		print_error(print_string, 0);
		return -1;
	}

        char *clFFT_extras_kernel_string = (char *)malloc(sizeof(char) * clFFT_extras_kernel_filesize);
        size_t clFFT_extras_kernel_length = 0;
        int clFFT_kernel_file_read = convertToString(clFFT_extras_kernel_filename, clFFT_extras_kernel_string, &clFFT_extras_kernel_length);

        if(clFFT_kernel_file_read == 0)
        {
                //sprintf(print_string, "Kernel string (length %d):", (int)clFFT_extras_kernel_length);
		//print_debug(print_string, 2);
		//print_debug(clFFT_extras_kernel_string, 3);

                clFFT_extras_program = clCreateProgramWithSource(context, 1, (const char **)&clFFT_extras_kernel_string, &clFFT_extras_kernel_length, &err);
		print_on_opencl_error("clCreateProgramWithSource of IFFT_prepare_kernel_string failed", err);
                if(err == CL_SUCCESS)
                        print_debug("SUCCESS: clCreateProgramWithSource", 3);

                err = clBuildProgram(clFFT_extras_program, 0, NULL, NULL, NULL, NULL);
		print_on_opencl_error("clBuildProgram of clFFT_extras_program failed.", err);
                if(err == CL_SUCCESS) 
                        printf("SUCCESS: clBuildProgram\n");

                print_opencl_build_info(&clFFT_extras_program);
		
		FFT_custom_kernel = clCreateKernel(clFFT_extras_program, "clFFT_custom", &err);
		print_on_opencl_error("clCreateKernel of clFFT_custom from clFFT_extras_program failed", err);
                if(err == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clFFT_custom\n");
		
		#ifdef ATI_GPU
		if(fft_size == 1024)
		{
			FFT_update_weights_ati_pt1_kernel = clCreateKernel(clFFT_extras_program, "clFFT_update_weights_ati_pt1", &err);
			print_on_opencl_error("clCreateKernel of clFFT_update_weights_ati_pt1 from clFFT_extras_program failed", err);
			if(err == CL_SUCCESS)
				printf("SUCCESS: clCreateKernel of clFFT_update_weights_ati_pt1\n");
			FFT_update_weights_ati_pt2_kernel = clCreateKernel(clFFT_extras_program, "clFFT_update_weights_ati_pt2", &err);
			print_on_opencl_error("clCreateKernel of clFFT_update_weights_ati_pt2 from clFFT_extras_program failed", err);
			if(err == CL_SUCCESS)
				printf("SUCCESS: clCreateKernel of clFFT_update_weights_ati_pt2\n");
		}
		else
		{
		#endif
			FFT_update_weights_kernel = clCreateKernel(clFFT_extras_program, "clFFT_update_weights", &err);
			print_on_opencl_error("clCreateKernel of clFFT_update_weights from clFFT_extras_program failed", err);
			if(err == CL_SUCCESS)
				printf("SUCCESS: clCreateKernel of clFFT_update_weights\n");
		#ifdef ATI_GPU
		}
		#endif
		
		FFT_convert_error_to_freq_domain_kernel = clCreateKernel(clFFT_extras_program, "clFFT_convert_error_to_freq_domain", &err);
		print_on_opencl_error("clCreateKernel of clFFT_convert_error_to_freq_domain from clFFT_extras_program failed", err);
                if(err == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clFFT_convert_error_to_freq_domain\n");
                        
                FFT_inner_prod_combined_kernel = clCreateKernel(clFFT_extras_program, "clFFT_inner_prod_combined", &err);
		print_on_opencl_error("clCreateKernel of clFFT_inner_prod_combined from clFFT_extras_program failed", err);
                if(err == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clFFT_inner_prod_combined\n");
		
		FFT_inverse_spectral_mul_accum_reduce_kernel = clCreateKernel(clFFT_extras_program, "clFFT_inverse_spectral_mul_accum_reduce", &err);
		print_on_opencl_error("clCreateKernel of clFFT_inverse_spectral_mul_accum_reduce from clFFT_extras_program failed", err);
                if(err == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clFFT_inverse_spectral_mul_accum_reduce\n");
        }
        else
        {
                sprintf(print_string, "Could not open file with FFT pre-/pre-processing kernels (in %s), exiting.\n", clFFT_extras_kernel_filename);
		print_error(print_string, 1);
        }
        
        printf("Cleaning up FFT kernel string.\n");
        free(clFFT_extras_kernel_string);
	printf("Cleaned up FFT kernel string.\n");
	
        return 0;
}

/**
 * Allocates memory
 * @arg max_batch_size maximum batch size for the regular FFT functions
 * @arg max_weight_update_multiplier maximum number of C*K for weight update function
 */
EXPORT int speex_echo_opencl_fft_initialize_mem(unsigned int max_batch_size, unsigned int max_weight_update_multiplier, unsigned int fft_size, cl_command_queue *commandQueue)
{
	// String used for passing information to the utility functions
	print_string = (char *)malloc(sizeof(char) * 300);
	
        print_debug("Running speex_echo_opencl_fft_initialize.", 3);

        //clFFT_Dim3 n = { fft_size, 1, 1 };
        //clFFT_Dim3 iN = { fft_size, 1, 1 };
        //numClFFTIter = 1;
        //clFFTBatchSize = 1;
        //clFFTLength = n.x + 2;
        //clIFFTLength = iN.x / 2;

        cl_int err = CL_SUCCESS;

        //int mx = log2(n.x);
        //int my = log2(n.y);
        //int mz = log2(n.z);
        //double gflops = 5e-9 * ((double)mx + (double)my + (double)mz) * (double)n.x * (double)n.y * (double)n.z * (double)clFFTBatchSize * (double)numClFFTIter;

        FFT_data_input_host  = (float *) malloc(sizeof(float)*max_batch_size*clFFTLength);
        FFT_data_output_host = (float *) malloc(sizeof(float)*max_batch_size*clFFTLength);
        if(!FFT_data_input_host || !FFT_data_output_host)
        {
                print_error("Out-of-Resouces", 0);
                return -1;
        }

        data_in = clCreateBuffer(context, CL_MEM_READ_WRITE, clFFTLength*sizeof(cl_float2)*max_batch_size, NULL, &err);
        if(!data_in || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }

        data_out = clCreateBuffer(context, CL_MEM_READ_WRITE, clFFTLength*sizeof(cl_float2)*max_batch_size, NULL, &err);
        if(!data_out || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }
        
        // Temporary buffers for weight update functions
	// Need to accomodate M*C*K*FFT_SIZE elements
	// Separate buffers because of their potentially large size
        data_in_weights = clCreateBuffer(context, CL_MEM_READ_WRITE, clFFTLength*sizeof(cl_float2)*max_batch_size*max_weight_update_multiplier, NULL, &err);
        if(!data_in || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }

        data_out_weights = clCreateBuffer(context, CL_MEM_READ_WRITE, clFFTLength*sizeof(cl_float2)*max_batch_size*max_weight_update_multiplier, NULL, &err);
        if(!data_out || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }
	
	data_in2 = clCreateBuffer(context, CL_MEM_READ_WRITE, clFFTLength*sizeof(cl_float2)*max_batch_size, NULL, &err);
        if(!data_in2 || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }

        data_out2 = clCreateBuffer(context, CL_MEM_READ_WRITE, clFFTLength*sizeof(cl_float2)*max_batch_size, NULL, &err);
        if(!data_out2 || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }

        return 0;
}

/**
 * Cleans up memory only (not kernel functions)
 */
EXPORT void speex_echo_opencl_fft_cleanup_mem()
{
        printf("Running speex_echo_opencl_fft_cleanup_mem.\n");

        cl_int status;
	
        if(FFT_data_input_host)
                free(FFT_data_input_host);
        if(FFT_data_output_host)
                free(FFT_data_output_host);

        if(data_in)
                clReleaseMemObject(data_in);
	if(data_in_weights)
                clReleaseMemObject(data_in_weights);
        if(data_out)
                clReleaseMemObject(data_out);
	if(data_out_weights)
                clReleaseMemObject(data_out_weights);
	if(data_in2)
                clReleaseMemObject(data_in2);
        if(data_out2)
                clReleaseMemObject(data_out2);
	
	printf("Cleaned up FFT buffers.\n");
}

/**
 * Cleans up kiernel functions only (not memory)
 */
EXPORT void speex_echo_opencl_fft_cleanup_dev()
{
        printf("Running speex_echo_opencl_fft_cleanup.\n");

        cl_int status;
	
        if(FFT_custom_kernel)
                clReleaseKernel(FFT_custom_kernel);
	if(FFT_update_weights_kernel)
                clReleaseKernel(FFT_update_weights_kernel);
	if(FFT_convert_error_to_freq_domain_kernel)
		clReleaseKernel(FFT_convert_error_to_freq_domain_kernel);
	if(FFT_inner_prod_combined_kernel)
                clReleaseKernel(FFT_inner_prod_combined_kernel);
	if(FFT_inverse_spectral_mul_accum_reduce_kernel)
                clReleaseKernel(FFT_inverse_spectral_mul_accum_reduce_kernel);
	if(FFT_update_weights_ati_pt1_kernel)
		clReleaseKernel(FFT_update_weights_ati_pt1_kernel);
	if(FFT_update_weights_ati_pt2_kernel)
		clReleaseKernel(FFT_update_weights_ati_pt2_kernel);
	
        if(clFFT_extras_program)
                clReleaseProgram(clFFT_extras_program);

	printf("Cleaned up FFT kernels/programs.\n");
	
	if(print_string)
		free(print_string);
}


/** 
 * Specialized function for the combined execution of the entire FFT in a single kernel.
 * Uses a hard-coded kernel with dimensions for Geforce 200-series GPUs.
 */
EXPORT void speex_echo_opencl_fft_execute2(cl_mem input, unsigned int input_offset, cl_mem output, unsigned int output_offset, unsigned int batch_size, cl_command_queue *commandQueue)
{
        double t;
        cl_int err = CL_SUCCESS;
	
        int s;
        size_t gWorkItems, lWorkItems;
        int inPlaceDone;
	int dir = clFFT_Forward;
	
	if(clFFTLength == 258)
	{
		lWorkItems = 64;
		gWorkItems = 64*batch_size;
	}
	else	// window size 1024
	{
		lWorkItems = 128;
		gWorkItems = 128*batch_size;
	}

	err = clSetKernelArg(FFT_custom_kernel, 0, sizeof(cl_mem), &input);		// Input from Speex
	print_on_opencl_error("Failed to set kernel argument 0 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 1, sizeof(cl_uint), &input_offset);	// Reformatted input
	print_on_opencl_error("Failed to set kernel argument 1 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 2, sizeof(cl_mem), &data_in);
	print_on_opencl_error("Failed to set kernel argument 2 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 3, sizeof(cl_mem), &data_out);
	print_on_opencl_error("Failed to set kernel argument 3 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 4, sizeof(cl_mem), &output);
	print_on_opencl_error("Failed to set kernel argument 4 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 5, sizeof(cl_uint), &output_offset);
	print_on_opencl_error("Failed to set kernel argument 5 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 6, sizeof(cl_int), &dir);
	print_on_opencl_error("Failed to set kernel argument 6 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 7, sizeof(cl_int), &batch_size);
	print_on_opencl_error("Failed to set kernel argument 7 of clFFT_custom.", err);

	err = clEnqueueNDRangeKernel(*commandQueue, FFT_custom_kernel, 1, NULL, &gWorkItems, &lWorkItems, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel clFFT_custom.", err);
	//if(err == CL_SUCCESS)
	//	print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_custom_kernel", 3);

	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clFinish failed for FFT_custom_kernel", err);
}


/** 
 * Specialized function for the combined execution of the entire IFFT in a single kernel.
 * Uses a hard-coded kernel with dimensions for Geforce 200-series GPUs.
 */
EXPORT void speex_echo_opencl_ifft_execute2(cl_mem input, unsigned int input_offset, cl_mem output, unsigned int output_offset, unsigned int batch_size, cl_command_queue *commandQueue)
{
        double t;
        cl_int err = CL_SUCCESS;
	
        int s;
        size_t gWorkItems, lWorkItems;
        int inPlaceDone;
	int dir = clFFT_Inverse;
	
	if(clFFTLength == 258)
	{
		lWorkItems = 64;
		gWorkItems = 64*batch_size;
	}
	else	// window size 1024
	{
		lWorkItems = 128;
		gWorkItems = 128*batch_size;
	}

	err = clSetKernelArg(FFT_custom_kernel, 0, sizeof(cl_mem), &input);		// Input from Speex
	print_on_opencl_error("Failed to set kernel argument 0 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 1, sizeof(cl_uint), &input_offset);	// Reformatted input
	print_on_opencl_error("Failed to set kernel argument 1 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 2, sizeof(cl_mem), &data_in);
	print_on_opencl_error("Failed to set kernel argument 2 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 3, sizeof(cl_mem), &data_out);
	print_on_opencl_error("Failed to set kernel argument 3 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 4, sizeof(cl_mem), &output);
	print_on_opencl_error("Failed to set kernel argument 4 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 5, sizeof(cl_uint), &output_offset);
	print_on_opencl_error("Failed to set kernel argument 5 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 6, sizeof(cl_int), &dir);
	print_on_opencl_error("Failed to set kernel argument 6 of clFFT_custom.", err);
	err = clSetKernelArg(FFT_custom_kernel, 7, sizeof(cl_int), &batch_size);
	print_on_opencl_error("Failed to set kernel argument 7 of clFFT_custom.", err);

	err = clEnqueueNDRangeKernel(*commandQueue, FFT_custom_kernel, 1, NULL, &gWorkItems, &lWorkItems, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel clFFT_custom.", err);
	//if(err == CL_SUCCESS)
	//	print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_custom_kernel", 3);

	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clFinish failed for FFT_custom_kernel", err);
}

/** 
 * Specialized function for updating weights for the entire echo tail in batch.
 * Performs an IFFT, then sets a couple of values to 0, then performs an FFT
 * Uses a hard-coded kernel with dimensions for Geforce 200-series GPUs.
 */
EXPORT void speex_echo_opencl_fft_update_weights(cl_mem W, cl_mem wtmp, unsigned int M, unsigned int C, unsigned int K, cl_command_queue *commandQueue)
{
        cl_int err = CL_SUCCESS;
	
        int s;  
        int inPlaceDone;
	int dir = clFFT_Forward;
	
	//lWorkItems = 64;
	//gWorkItems = 64*M;
	
	size_t gWorkItems[3];
	size_t lWorkItems[3];
	
	if(clFFTLength == 258)
	{
		gWorkItems[0] = 64*M;
		gWorkItems[1] = C;
		gWorkItems[2] = K;
		lWorkItems[0] = 64;
		lWorkItems[1] = 1;
		lWorkItems[2] = 1;
	}
	else	// window size 1024
	{
		gWorkItems[0] = 128*M;
		gWorkItems[1] = C;
		gWorkItems[2] = K;
		lWorkItems[0] = 128;
		lWorkItems[1] = 1;
		lWorkItems[2] = 1;
	}
	
	// Use a two-part kernel for ATI, because it segfaults on the large kernel
	#ifdef ATI_GPU
	if(clFFTLength == 1026)
	{
		// PART 1
		err = clSetKernelArg(FFT_update_weights_ati_pt1_kernel, 0, sizeof(cl_mem), &W);		// Input from Speex
		print_on_opencl_error("Failed to set kernel argument 0 of clFFT_update_weights_ati_pt1.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt1_kernel, 1, sizeof(cl_mem), &data_in_weights);
		print_on_opencl_error("Failed to set kernel argument 1 of clFFT_update_weights_ati_pt1.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt1_kernel, 2, sizeof(cl_mem), &data_out_weights);
		print_on_opencl_error("Failed to set kernel argument 2 of clFFT_update_weights_ati_pt1.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt1_kernel, 3, sizeof(cl_mem), &wtmp);
		print_on_opencl_error("Failed to set kernel argument 3 of clFFT_update_weights_ati_pt1.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt1_kernel, 4, sizeof(cl_int), &M);
		print_on_opencl_error("Failed to set kernel argument 4 of clFFT_update_weights_ati_pt1.", err);

		err = clEnqueueNDRangeKernel(*commandQueue, FFT_update_weights_ati_pt1_kernel, 3, NULL, gWorkItems, lWorkItems, 0, NULL, NULL);
		print_on_opencl_error("Failed to execute kernel clFFT_update_weights_ati_pt1.", err);
		if(err == CL_SUCCESS)
			print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_update_weights_ati_pt1_kernel", 3);

		err = clEnqueueBarrier(*commandQueue);
		print_on_opencl_error("clFinish failed for FFT_update_weights_ati_pt1_kernel", err);
		
		// PART 2
		err = clSetKernelArg(FFT_update_weights_ati_pt2_kernel, 0, sizeof(cl_mem), &W);		// Input from Speex
		print_on_opencl_error("Failed to set kernel argument 0 of clFFT_update_weights_ati_pt2.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt2_kernel, 1, sizeof(cl_mem), &data_in_weights);
		print_on_opencl_error("Failed to set kernel argument 1 of clFFT_update_weights_ati_pt2.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt2_kernel, 2, sizeof(cl_mem), &data_out_weights);
		print_on_opencl_error("Failed to set kernel argument 2 of clFFT_update_weights_ati_pt2.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt2_kernel, 3, sizeof(cl_mem), &wtmp);
		print_on_opencl_error("Failed to set kernel argument 3 of clFFT_update_weights_ati_pt2.", err);
		err = clSetKernelArg(FFT_update_weights_ati_pt2_kernel, 4, sizeof(cl_int), &M);
		print_on_opencl_error("Failed to set kernel argument 4 of clFFT_update_weights_ati_pt2.", err);

		err = clEnqueueNDRangeKernel(*commandQueue, FFT_update_weights_ati_pt2_kernel, 3, NULL, gWorkItems, lWorkItems, 0, NULL, NULL);
		print_on_opencl_error("Failed to execute kernel clFFT_update_weights_ati_pt2.", err);
		if(err == CL_SUCCESS)
			print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_update_weights_ati_pt2_kernel", 3);

		err = clEnqueueBarrier(*commandQueue);
		print_on_opencl_error("clFinish failed for FFT_update_weights_ati_pt2_kernel", err);
	}
	else
	{
	#endif
		err = clSetKernelArg(FFT_update_weights_kernel, 0, sizeof(cl_mem), &W);		// Input from Speex
		print_on_opencl_error("Failed to set kernel argument 0 of clFFT_update_weights.", err);
		err = clSetKernelArg(FFT_update_weights_kernel, 1, sizeof(cl_mem), &data_in_weights);
		print_on_opencl_error("Failed to set kernel argument 1 of clFFT_update_weights.", err);
		err = clSetKernelArg(FFT_update_weights_kernel, 2, sizeof(cl_mem), &data_out_weights);
		print_on_opencl_error("Failed to set kernel argument 2 of clFFT_update_weights.", err);
		err = clSetKernelArg(FFT_update_weights_kernel, 3, sizeof(cl_mem), &wtmp);
		print_on_opencl_error("Failed to set kernel argument 3 of clFFT_update_weights.", err);
		err = clSetKernelArg(FFT_update_weights_kernel, 4, sizeof(cl_int), &M);
		print_on_opencl_error("Failed to set kernel argument 4 of clFFT_update_weights.", err);

		err = clEnqueueNDRangeKernel(*commandQueue, FFT_update_weights_kernel, 3, NULL, gWorkItems, lWorkItems, 0, NULL, NULL);
		print_on_opencl_error("Failed to execute kernel clFFT_update_weights.", err);
		if(err == CL_SUCCESS)
			print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_update_weights_kernel", 3);

		err = clEnqueueBarrier(*commandQueue);
		print_on_opencl_error("clFinish failed for FFT_update_weights_kernel", err);
	#ifdef ATI_GPU
	}
	#endif
}

/** 
 * Specialized function for doing two FFTs in parallell
 * Also sets the value of the last half of y
 * Uses a hard-coded kernel with dimensions for Geforce 200-series GPUs.
 */
EXPORT void speex_echo_opencl_fft_convert_error(cl_mem e, cl_mem E, cl_mem y, cl_mem Y, unsigned int C, cl_command_queue *commandQueue)
{
        cl_int err = CL_SUCCESS;
	
        int s;
	size_t gWorkItems[2];
	size_t lWorkItems[2];
	
	if(clFFTLength == 258)
	{
		gWorkItems[0] = 64*C;
		gWorkItems[1] = 2;	// Two independant FFTs in C batches
		lWorkItems[0] = 64;
		lWorkItems[1] = 1;
	}
	else	// window size 1024
	{
		gWorkItems[0] = 128*C;
		gWorkItems[1] = 2;
		lWorkItems[0] = 128;
		lWorkItems[1] = 1;
	}
        int inPlaceDone;
	
	/*lWorkItems[0] = 64;
	lWorkItems[1] = 1;
	gWorkItems[0] = 64*C;
	gWorkItems[1] = 2;	// Two independant FFTs*/

	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 0, sizeof(cl_mem), &e);
	print_on_opencl_error("Failed to set kernel argument 0 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 1, sizeof(cl_mem), &E);
	print_on_opencl_error("Failed to set kernel argument 1 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 2, sizeof(cl_mem), &y);
	print_on_opencl_error("Failed to set kernel argument 2 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 3, sizeof(cl_mem), &Y);
	print_on_opencl_error("Failed to set kernel argument 3 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 4, sizeof(cl_mem), &data_in);
	print_on_opencl_error("Failed to set kernel argument 4 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 5, sizeof(cl_mem), &data_in2);
	print_on_opencl_error("Failed to set kernel argument 5 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 6, sizeof(cl_mem), &data_out);
	print_on_opencl_error("Failed to set kernel argument 6 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 7, sizeof(cl_mem), &data_out2);
	print_on_opencl_error("Failed to set kernel argument 7 of clFFT_convert_error_to_freq_domain.", err);
	err = clSetKernelArg(FFT_convert_error_to_freq_domain_kernel, 8, sizeof(cl_uint), &C);
	print_on_opencl_error("Failed to set kernel argument 8 of clFFT_convert_error_to_freq_domain.", err);

	err = clEnqueueNDRangeKernel(*commandQueue, FFT_convert_error_to_freq_domain_kernel, 2, NULL, gWorkItems, lWorkItems, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel clFFT_convert_error_to_freq_domain.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_convert_error_to_freq_domain_kernel", 3);

	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for FFT_convert_error_to_freq_domain_kernel", err);
}

/** 
 * Specialized function for doing both FFT and inner product
 * Gives overall poorer performance than the separate functions.
 * Uses a hard-coded kernel with dimensions for Geforce 200-series GPUs.
 */
EXPORT void speex_echo_opencl_fft_inner_prod_combined(cl_mem data_time_domain, cl_mem data_freq_domain, cl_mem Xf, cl_mem temp_reduction, float *Sxx, unsigned int frame_size, unsigned int K, cl_command_queue *commandQueue)
{
        cl_int err = CL_SUCCESS;
	float *block_sums = (float *)malloc(sizeof(cl_float) * K);
	int i;
	
	size_t gWorkItems[2];
	size_t lWorkItems[2];
	
	if(clFFTLength == 258)
	{
		gWorkItems[0] = 64*K;
		gWorkItems[1] = 1;
		lWorkItems[0] = 64;
		lWorkItems[1] = 1;
	}
	else	// window size 1024
	{
		gWorkItems[0] = 128*K;
		gWorkItems[1] = 1;
		lWorkItems[0] = 128;
		lWorkItems[1] = 1;
	}

	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 0, sizeof(cl_mem), &data_time_domain);
	print_on_opencl_error("Failed to set kernel argument 0 of clFFT_inner_prod_combined.", err);
	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 1, sizeof(cl_mem), &data_in);
	print_on_opencl_error("Failed to set kernel argument 1 of clFFT_inner_prod_combined.", err);
	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 2, sizeof(cl_mem), &data_out);
	print_on_opencl_error("Failed to set kernel argument 2 of clFFT_inner_prod_combined.", err);
	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 3, sizeof(cl_mem), &data_freq_domain);
	print_on_opencl_error("Failed to set kernel argument 3 of clFFT_inner_prod_combined.", err);
	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 4, sizeof(cl_mem), &Xf);
	print_on_opencl_error("Failed to set kernel argument 4 of clFFT_inner_prod_combined.", err);
	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 5, sizeof(cl_mem), &temp_reduction);
	print_on_opencl_error("Failed to set kernel argument 5 of clFFT_inner_prod_combined.", err);
	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 6, sizeof(cl_int), &frame_size);
	print_on_opencl_error("Failed to set kernel argument 6 of clFFT_inner_prod_combined.", err);
	err = clSetKernelArg(FFT_inner_prod_combined_kernel, 7, sizeof(cl_int), &K);
	print_on_opencl_error("Failed to set kernel argument 7 of clFFT_inner_prod_combined.", err);

	err = clEnqueueNDRangeKernel(*commandQueue, FFT_inner_prod_combined_kernel, 2, NULL, gWorkItems, lWorkItems, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel clFFT_inner_prod_combined.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_inner_prod_combined_kernel", 3);

	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for FFT_inner_prod_combined_kernel", err);
	
	// Do a reduction of the sum for each block on the CPU
	err = clEnqueueReadBuffer(*commandQueue, temp_reduction, CL_TRUE, 0, K*sizeof(cl_float), block_sums, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy to temporary frequency block sum array in the opencl inner product function.", err);
	for(i=0;i<K;i++)
	{
		*Sxx += block_sums[i];
		//*Sxx += 1000.0f;
		//printf("Adding %f to Sxx\n", block_sums[i]);
	}

	free(block_sums);
}

/** 
 * Specialized function for doing both IFFT and last phase of spectral mul accum
 * Uses a hard-coded kernel with dimensions for Geforce 200-series GPUs.
 */
EXPORT void speex_echo_opencl_fft_inverse_spectral_mul_accum_reduce(cl_mem Y, cl_mem e, cl_mem temp_acc, int M, int N, int K, cl_command_queue *commandQueue)
{
        cl_int err = CL_SUCCESS;
	int i;
	
	size_t gWorkItems[2];
	size_t lWorkItems[2];
	
	if(clFFTLength == 258)
	{
		gWorkItems[0] = 64*K;
		gWorkItems[1] = 1;
		lWorkItems[0] = 64;
		lWorkItems[1] = 1;
	}
	else	// window size 1024
	{
		gWorkItems[0] = 128*K;
		gWorkItems[1] = 1;
		lWorkItems[0] = 128;
		lWorkItems[1] = 1;
	}

	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 0, sizeof(cl_mem), &Y);
	print_on_opencl_error("Failed to set kernel argument 0 of clFFT_inverse_spectral_mul_accum_reduce.", err);
	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 1, sizeof(cl_mem), &data_in);
	print_on_opencl_error("Failed to set kernel argument 1 of clFFT_inverse_spectral_mul_accum_reduce.", err);
	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 2, sizeof(cl_mem), &data_out);
	print_on_opencl_error("Failed to set kernel argument 2 of clFFT_inverse_spectral_mul_accum_reduce.", err);
	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 3, sizeof(cl_mem), &e);
	print_on_opencl_error("Failed to set kernel argument 3 of clFFT_inverse_spectral_mul_accum_reduce.", err);
	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 4, sizeof(cl_mem), &temp_acc);
	print_on_opencl_error("Failed to set kernel argument 4 of clFFT_inverse_spectral_mul_accum_reduce.", err);
	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 5, sizeof(cl_int), &M);
	print_on_opencl_error("Failed to set kernel argument 5 of clFFT_inverse_spectral_mul_accum_reduce.", err);
	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 6, sizeof(cl_int), &N);
	print_on_opencl_error("Failed to set kernel argument 6 of clFFT_inverse_spectral_mul_accum_reduce.", err);
	err = clSetKernelArg(FFT_inverse_spectral_mul_accum_reduce_kernel, 7, sizeof(cl_int), &K);
	print_on_opencl_error("Failed to set kernel argument 7 of clFFT_inverse_spectral_mul_accum_reduce.", err);

	err = clEnqueueNDRangeKernel(*commandQueue, FFT_inverse_spectral_mul_accum_reduce_kernel, 2, NULL, gWorkItems, lWorkItems, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel clFFT_inverse_spectral_mul_accum_reduce.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of FFT_inverse_spectral_mul_accum_reduce_kernel", 3);

	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for FFT_inverse_spectral_mul_accum_reduce_kernel", err);
}

// Generates the symmetric negative frequencies needed to calculate and
// inverse real DFT using a complex DFT (like the OpenCL FFT library).
//
// complex_data is assumed to be an array of interleaved complex numbers (real0,imag0,real1,imag1...realN,imagN)
//
// Source: http://www.dspguide.com/ch12/1.htm
void real_dft_negative_frequency_generation(clFFT_Complex *complex_data, int num_points)
{
        int i;
        for(i = (num_points/2+1-1);i < (num_points); i++)
        {
                complex_data[i].real = complex_data[num_points-i].real;
                complex_data[i].imag = -complex_data[num_points-i].imag;
        }
}

double complexNormSq(clFFT_ComplexDouble a)
{
        return (a.real * a.real + a.imag * a.imag);
}

double computeL2ErrorReal(const float *data_ref, const float *data, int n, int batchSize, double *max_diff, double *min_diff)
{
        int i, j;
        double avg_norm = 0.0;
        *max_diff = 0.0;
        *min_diff = 0x1.0p1000;

        for(j = 0; j < batchSize; j++)
        {
                double norm_ref = 0.0;
                double norm = 0.0;
                for(i = 0; i < n; i++)
                {
                        int index = j * n + i;
                        double diff = data_ref[index] - data[index];
                        double norm_tmp = diff * diff;
                        norm += norm_tmp;
                        norm_ref += (data_ref[index] * data_ref[index]);
                }
                double curr_norm = sqrt( norm / norm_ref ) / FLT_EPSILON;
                avg_norm += curr_norm;
                *max_diff = *max_diff < curr_norm ? curr_norm : *max_diff;
                *min_diff = *min_diff > curr_norm ? curr_norm : *min_diff;
        }

        return avg_norm / batchSize;
}

double computeL2ErrorComplex(const float *data_ref, const float *data, int n, int batchSize, double *max_diff, double *min_diff)
{
        int i, j;
        double avg_norm = 0.0;
        *max_diff = 0.0;
        *min_diff = 0x1.0p1000;

        for(j = 0; j < batchSize; j++)
        {
                double norm_ref = 0.0;
                double norm = 0.0;
                for(i = 0; i < n; i++)
                {
                        int index = j * n + i * 2;
                        //if(index > 240)
                        //      printf("Complex num %d, real %2.3f/%2.3f, imag %2.3f/%2.3f\n", i, data_ref[index], data[index], data_ref[index+1], data[index+1]);
                        clFFT_ComplexDouble diff = (clFFT_ComplexDouble) { data_ref[index] - data[index], data_ref[index+1] - data[index+1]};
                        double norm_tmp = complexNormSq(diff);
                        norm += norm_tmp;
                        norm_ref += (data_ref[index] * data_ref[index] + data_ref[index+1] * data_ref[index+1]);
                }
                double curr_norm = sqrt( norm / norm_ref ) / FLT_EPSILON;
                avg_norm += curr_norm;
                *max_diff = *max_diff < curr_norm ? curr_norm : *max_diff;
                *min_diff = *min_diff > curr_norm ? curr_norm : *min_diff;
        }

        return avg_norm / batchSize;
}

