// This file contains FFT-related kernels:
// * Conversion between Speex format and Apple OpenCL FFT format
// * Hardcoded FFT kernels based on Apple OpenCL FFT (to avoid library overhead)
// * Various kernels where the FFT is combined with other operations

__kernel void clFFT_forward_prepare_data(__global float *in, unsigned int in_offset, __global float *out, float m, unsigned int fftLength)
{                                                                      
	unsigned int i = get_global_id(0);
	
	out[i*2] = in[i + in_offset] * m;
	out[i*2 + 1] = 0.0f;
}

__kernel void clFFT_inverse_prepare_data(__global float *in, unsigned int in_offset, __global float *out, unsigned int fftLength) 
{                                                                      
	unsigned int i = get_global_id(0);
	unsigned int j = fftLength + i*2;
	unsigned int arraySize = fftLength * 2;
	
	if(i == 0)
	{
		out[0] = in[in_offset + 0];
		out[1] = 0.0f;
		out[fftLength] = in[in_offset + fftLength - 1];
		//out[fftLength + 1] = 0.0f;
		out[257] = 0.0f;
	}	
	else if(i < (fftLength/2)) {
		out[i*2] = in[in_offset + (i*2) - 1]; 
		out[i*2+1] = in[in_offset + (i*2)];
	}

	barrier(CLK_GLOBAL_MEM_FENCE);
	barrier(CLK_LOCAL_MEM_FENCE);

	// Generate negative frequencies for symmetric input
	// Source: http://www.dspguide.com/ch12/1.htm
	// For the remaining n/2 threads that needs to mirror 
	if(i < (fftLength/2)) {
		out[fftLength*2 - i*2] = out[i*2];
		out[fftLength*2 - i*2 + 1] = -out[i*2+1];
	}
}

__kernel void clFFT_inverse_prepare_data_test(__global float *in, unsigned int in_offset, __global float2 *out, unsigned int fftLength)
{                                                                      
	unsigned int i = get_global_id(0);
	unsigned int j = fftLength + i*2;
	unsigned int arraySize = fftLength * 2;
	unsigned int lId = get_local_id(0);
	
	/*if(lId == 0)
	{
		out[0].x = in[in_offset + 0];
		out[0].y = 0.0f;
		out[1].x = in[in_offset + 1];
		out[1].y = in[in_offset + 2];
		out[128].x = in[in_offset + 256 - 1];
		out[128].y = 0.0f;
	}
	else
	{
		out[2*lId].x = in[in_offset + lId*4 - 1];
		out[2*lId].y = in[in_offset + lId*4];
		out[2*lId + 1].x = in[in_offset + lId*4 + 1];
		out[2*lId + 1].y = in[in_offset + lId*4 + 2];
		//out[lId].x = 9000.0f;
		//out[lId].y = 8000.0f;
	}

	//in[2*lId].x = 9000.0f;  
	//in[2*lId].y = 9000.0f;
	//in[2*lId + 1].x = 9000.0f;  
	//in[2*lId + 1].y = 9000.0f;

	barrier(CLK_GLOBAL_MEM_FENCE);
	barrier(CLK_LOCAL_MEM_FENCE);

	// Generate negative frequencies for symmetric input 
	// Source: http://www.dspguide.com/ch12/1.htm
	// For the remaining n/2 threads that needs to mirror
	out[256 - lId*2].x = out[lId*2].x;
	out[256 - lId*2].y = -out[lId*2].y;
	out[256 - lId*2 - 1].x = out[lId*2 + 1].x;
	out[256 - lId*2 - 1].y = -out[lId*2 + 1].y;*/
}

__kernel void clFFT_forward_finalize_data(__global float *in, __global float *out, unsigned int out_offset, unsigned int fftLength)
{                                                                      
	unsigned int i = get_global_id(0);
	
	if(i == 0)
	{
		out[out_offset + 0] = in[0];
		out[out_offset + 1] = in[2];
		//out[2] = in[2];
	}
	else
	{
		//out[i*2] = in[i*2-2];
		//out[i*2 + 1] = in[i*2-1];
		out[out_offset + i*2] = in[i*2 + 1];
		out[out_offset + i*2+1] = in[i*2 + 2];
	}	

	//out[2] = in[3];
	//out[3] = in[4];
}

__kernel void clFFT_inverse_finalize_data(__global float *in, __global float *out, unsigned int out_offset, unsigned int fftLength)
{                                                                      
	unsigned int i = get_global_id(0);

	out[i + out_offset] = in[2*i];	
}

#ifndef M_PI
#define M_PI 0x1.921fb54442d18p+1
#endif
float2 complexMul(float2 a,float2 B) { return (float2)(mad(-(a).y, (B).y, (a).x * (B).x), mad((a).y, (B).x, (a).x * (B).y));}
#define conj(a) ((float2)((a).x, -(a).y))
#define conjTransp(a) ((float2)(-(a).y, (a).x))

#define fftKernel2(a,dir) \
{ \
    float2 c = (a)[0];    \
    (a)[0] = c + (a)[1];  \
    (a)[1] = c - (a)[1];  \
}

#define fftKernel2S(d1,d2,dir) \
{ \
    float2 c = (d1);   \
    (d1) = c + (d2);   \
    (d2) = c - (d2);   \
}

#define fftKernel4(a,dir) \
{ \
    fftKernel2S((a)[0], (a)[2], dir); \
    fftKernel2S((a)[1], (a)[3], dir); \
    fftKernel2S((a)[0], (a)[1], dir); \
    (a)[3] = (float2)(dir)*(conjTransp((a)[3])); \
    fftKernel2S((a)[2], (a)[3], dir); \
    float2 c = (a)[1]; \
    (a)[1] = (a)[2]; \
    (a)[2] = c; \
}

#define fftKernel4s(a0,a1,a2,a3,dir) \
{ \
    fftKernel2S((a0), (a2), dir); \
    fftKernel2S((a1), (a3), dir); \
    fftKernel2S((a0), (a1), dir); \
    (a3) = (float2)(dir)*(conjTransp((a3))); \
    fftKernel2S((a2), (a3), dir); \
    float2 c = (a1); \
    (a1) = (a2); \
    (a2) = c; \
}

#define bitreverse8(a) \
{ \
    float2 c; \
    c = (a)[1]; \
    (a)[1] = (a)[4]; \
    (a)[4] = c; \
    c = (a)[3]; \
    (a)[3] = (a)[6]; \
    (a)[6] = c; \
}

#define fftKernel8(a,dir) \
{ \
        const float2 w1  = (float2)(0x1.6a09e6p-1f,  dir*0x1.6a09e6p-1f);  \
        const float2 w3  = (float2)(-0x1.6a09e6p-1f, dir*0x1.6a09e6p-1f);  \
        float2 c; \
        fftKernel2S((a)[0], (a)[4], dir); \
        fftKernel2S((a)[1], (a)[5], dir); \
        fftKernel2S((a)[2], (a)[6], dir); \
        fftKernel2S((a)[3], (a)[7], dir); \
        (a)[5] = complexMul(w1, (a)[5]); \
        (a)[6] = (float2)(dir)*(conjTransp((a)[6])); \
        (a)[7] = complexMul(w3, (a)[7]); \
        fftKernel2S((a)[0], (a)[2], dir); \
        fftKernel2S((a)[1], (a)[3], dir); \
        fftKernel2S((a)[4], (a)[6], dir); \
        fftKernel2S((a)[5], (a)[7], dir); \
        (a)[3] = (float2)(dir)*(conjTransp((a)[3])); \
        (a)[7] = (float2)(dir)*(conjTransp((a)[7])); \
        fftKernel2S((a)[0], (a)[1], dir); \
        fftKernel2S((a)[2], (a)[3], dir); \
        fftKernel2S((a)[4], (a)[5], dir); \
        fftKernel2S((a)[6], (a)[7], dir); \
        bitreverse8((a)); \
}

#define bitreverse4x4(a) \
{ \
        float2 c; \
        c = (a)[1];  (a)[1]  = (a)[4];  (a)[4]  = c; \
        c = (a)[2];  (a)[2]  = (a)[8];  (a)[8]  = c; \
        c = (a)[3];  (a)[3]  = (a)[12]; (a)[12] = c; \
        c = (a)[6];  (a)[6]  = (a)[9];  (a)[9]  = c; \
        c = (a)[7];  (a)[7]  = (a)[13]; (a)[13] = c; \
        c = (a)[11]; (a)[11] = (a)[14]; (a)[14] = c; \
}

#define fftKernel16(a,dir) \
{ \
    const float w0 = 0x1.d906bcp-1f; \
    const float w1 = 0x1.87de2ap-2f; \
    const float w2 = 0x1.6a09e6p-1f; \
    fftKernel4s((a)[0], (a)[4], (a)[8],  (a)[12], dir); \
    fftKernel4s((a)[1], (a)[5], (a)[9],  (a)[13], dir); \
    fftKernel4s((a)[2], (a)[6], (a)[10], (a)[14], dir); \
    fftKernel4s((a)[3], (a)[7], (a)[11], (a)[15], dir); \
    (a)[5]  = complexMul((a)[5], (float2)(w0, dir*w1)); \
    (a)[6]  = complexMul((a)[6], (float2)(w2, dir*w2)); \
    (a)[7]  = complexMul((a)[7], (float2)(w1, dir*w0)); \
    (a)[9]  = complexMul((a)[9], (float2)(w2, dir*w2)); \
    (a)[10] = (float2)(dir)*(conjTransp((a)[10])); \
    (a)[11] = complexMul((a)[11], (float2)(-w2, dir*w2)); \
    (a)[13] = complexMul((a)[13], (float2)(w1, dir*w0)); \
    (a)[14] = complexMul((a)[14], (float2)(-w2, dir*w2)); \
    (a)[15] = complexMul((a)[15], (float2)(-w0, dir*-w1)); \
    fftKernel4((a), dir); \
    fftKernel4((a) + 4, dir); \
    fftKernel4((a) + 8, dir); \
    fftKernel4((a) + 12, dir); \
    bitreverse4x4((a)); \
}

#define bitreverse32(a) \
{ \
    float2 c1, c2; \
    c1 = (a)[2];   (a)[2] = (a)[1];   c2 = (a)[4];   (a)[4] = c1;   c1 = (a)[8];   (a)[8] = c2;    c2 = (a)[16];  (a)[16] = c1;   (a)[1] = c2; \
    c1 = (a)[6];   (a)[6] = (a)[3];   c2 = (a)[12];  (a)[12] = c1;  c1 = (a)[24];  (a)[24] = c2;   c2 = (a)[17];  (a)[17] = c1;   (a)[3] = c2; \
    c1 = (a)[10];  (a)[10] = (a)[5];  c2 = (a)[20];  (a)[20] = c1;  c1 = (a)[9];   (a)[9] = c2;    c2 = (a)[18];  (a)[18] = c1;   (a)[5] = c2; \
    c1 = (a)[14];  (a)[14] = (a)[7];  c2 = (a)[28];  (a)[28] = c1;  c1 = (a)[25];  (a)[25] = c2;   c2 = (a)[19];  (a)[19] = c1;   (a)[7] = c2; \
    c1 = (a)[22];  (a)[22] = (a)[11]; c2 = (a)[13];  (a)[13] = c1;  c1 = (a)[26];  (a)[26] = c2;   c2 = (a)[21];  (a)[21] = c1;   (a)[11] = c2; \
    c1 = (a)[30];  (a)[30] = (a)[15]; c2 = (a)[29];  (a)[29] = c1;  c1 = (a)[27];  (a)[27] = c2;   c2 = (a)[23];  (a)[23] = c1;   (a)[15] = c2; \
}

#define fftKernel32(a,dir) \
{ \
    fftKernel2S((a)[0],  (a)[16], dir); \
    fftKernel2S((a)[1],  (a)[17], dir); \
    fftKernel2S((a)[2],  (a)[18], dir); \
    fftKernel2S((a)[3],  (a)[19], dir); \
    fftKernel2S((a)[4],  (a)[20], dir); \
    fftKernel2S((a)[5],  (a)[21], dir); \
    fftKernel2S((a)[6],  (a)[22], dir); \
    fftKernel2S((a)[7],  (a)[23], dir); \
    fftKernel2S((a)[8],  (a)[24], dir); \
    fftKernel2S((a)[9],  (a)[25], dir); \
    fftKernel2S((a)[10], (a)[26], dir); \
    fftKernel2S((a)[11], (a)[27], dir); \
    fftKernel2S((a)[12], (a)[28], dir); \
    fftKernel2S((a)[13], (a)[29], dir); \
    fftKernel2S((a)[14], (a)[30], dir); \
    fftKernel2S((a)[15], (a)[31], dir); \
    (a)[17] = complexMul((a)[17], (float2)(0x1.f6297cp-1f, dir*0x1.8f8b84p-3f)); \
    (a)[18] = complexMul((a)[18], (float2)(0x1.d906bcp-1f, dir*0x1.87de2ap-2f)); \
    (a)[19] = complexMul((a)[19], (float2)(0x1.a9b662p-1f, dir*0x1.1c73b4p-1f)); \
    (a)[20] = complexMul((a)[20], (float2)(0x1.6a09e6p-1f, dir*0x1.6a09e6p-1f)); \
    (a)[21] = complexMul((a)[21], (float2)(0x1.1c73b4p-1f, dir*0x1.a9b662p-1f)); \
    (a)[22] = complexMul((a)[22], (float2)(0x1.87de2ap-2f, dir*0x1.d906bcp-1f)); \
    (a)[23] = complexMul((a)[23], (float2)(0x1.8f8b84p-3f, dir*0x1.f6297cp-1f)); \
    (a)[24] = complexMul((a)[24], (float2)(0x0p+0f, dir*0x1p+0f)); \
    (a)[25] = complexMul((a)[25], (float2)(-0x1.8f8b84p-3f, dir*0x1.f6297cp-1f)); \
    (a)[26] = complexMul((a)[26], (float2)(-0x1.87de2ap-2f, dir*0x1.d906bcp-1f)); \
    (a)[27] = complexMul((a)[27], (float2)(-0x1.1c73b4p-1f, dir*0x1.a9b662p-1f)); \
    (a)[28] = complexMul((a)[28], (float2)(-0x1.6a09e6p-1f, dir*0x1.6a09e6p-1f)); \
    (a)[29] = complexMul((a)[29], (float2)(-0x1.a9b662p-1f, dir*0x1.1c73b4p-1f)); \
    (a)[30] = complexMul((a)[30], (float2)(-0x1.d906bcp-1f, dir*0x1.87de2ap-2f)); \
    (a)[31] = complexMul((a)[31], (float2)(-0x1.f6297cp-1f, dir*0x1.8f8b84p-3f)); \
    fftKernel16((a), dir); \
    fftKernel16((a) + 16, dir); \
    bitreverse32((a)); \
}

__kernel void \
clFFT_1DTwistInterleaved(__global float2 *in, unsigned int startRow, unsigned int numCols, unsigned int N, unsigned int numRowsToProcess, int dir) \
{ \
   float2 a, w; \
   float ang; \
   unsigned int j; \
        unsigned int i = get_global_id(0); \
        unsigned int startIndex = i; \
         \
        if(i < numCols) \
        { \
            for(j = 0; j < numRowsToProcess; j++) \
            { \
                a = in[startIndex]; \
                ang = 2.0f * M_PI * dir * i * (startRow + j) / N; \
                w = (float2)(native_cos(ang), native_sin(ang)); \
                a = complexMul(a, w); \
                in[startIndex] = a; \
                startIndex += numCols; \
            } \
        }        \
} \

/**
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 256.
 */ 
__kernel void clFFT_custom(__global float *speex_in, unsigned int speex_in_offset, __global float2 *in, __global float2 *out, __global float *speex_out, unsigned int speex_out_offset, int dir, int S)
{
	__local float sMem[272];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[4];
	int lId = get_local_id( 0 );
	int groupId = get_group_id( 0 );
	const float m = 1.0f / 256.0f;
	int groupOffset = groupId * 256;
	ii = lId;
	jj = 0;

	offset =  mad24(groupId, 256, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	//in += offset;
	//out += offset;
	int workItemOffset = lId * 4 + groupOffset;
	
	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	// clFFT_Forward
	if(dir == -1)
	{	
		in[workItemOffset].x = speex_in[workItemOffset + speex_in_offset] * m;
        	in[workItemOffset].y = 0.0f;
		in[workItemOffset + 1].x = speex_in[workItemOffset + 1 + speex_in_offset] * m;
        	in[workItemOffset + 1].y = 0.0f;
		in[workItemOffset + 2].x = speex_in[workItemOffset + 2 + speex_in_offset] * m;
        	in[workItemOffset + 2].y = 0.0f;
		in[workItemOffset + 3].x = speex_in[workItemOffset + 3 + speex_in_offset] * m;
        	in[workItemOffset + 3].y = 0.0f;	
		
		barrier(CLK_GLOBAL_MEM_FENCE);
	        barrier(CLK_LOCAL_MEM_FENCE);
	}
	// clFFT_Inverse
	else if(dir == 1)
	{
		if(lId == 0)
		{
			in[groupOffset + 0].x = speex_in[groupOffset + speex_in_offset + 0];
			in[groupOffset + 0].y = 0.0f;
			in[groupOffset + 1].x = speex_in[groupOffset + speex_in_offset + 1];
			in[groupOffset + 1].y = speex_in[groupOffset + speex_in_offset + 2];
			in[groupOffset + 128].x = speex_in[groupOffset + speex_in_offset + 256 - 1];
			in[groupOffset + 128].y = 0.0f;
		}
		else
		{
			in[groupOffset + 2*lId].x = speex_in[groupOffset + speex_in_offset + lId*4 - 1];
			in[groupOffset + 2*lId].y = speex_in[groupOffset + speex_in_offset + lId*4];
			in[groupOffset + 2*lId + 1].x = speex_in[groupOffset + speex_in_offset + lId*4 + 1];
			in[groupOffset + 2*lId + 1].y = speex_in[groupOffset + speex_in_offset + lId*4 + 2];
		}

		barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

		// Generate negative frequencies for symmetric input 
	        // Source: http://www.dspguide.com/ch12/1.htm
	        // For the remaining n/2 threads that needs to mirror
		/*in[groupOffset + 256 - lId*2].x = in[groupOffset + lId*2].x;
		in[groupOffset + 256 - lId*2].y = -in[groupOffset + lId*2].y;
		in[groupOffset + 256 - lId*2 - 1].x = in[groupOffset + lId*2 + 1].x;
		in[groupOffset + 256 - lId*2 - 1].y = -in[groupOffset + lId*2 + 1].y;*/
		
		in[groupOffset + 255 - lId*2].x = in[groupOffset + lId*2 + 1].x;
		in[groupOffset + 255 - lId*2].y = -in[groupOffset + lId*2 + 1].y;
		in[groupOffset + 255 - lId*2 - 1].x = in[groupOffset + lId*2 + 2].x;
		in[groupOffset + 255 - lId*2 - 1].y = -in[groupOffset + lId*2 + 2].y;
	}
      //253 - 3
      //252 - 4
      //251 - 5

	barrier(CLK_LOCAL_MEM_FENCE);

	in += offset;
	out += offset;

	// End of preparation step

	a[0] = in[0];
	a[1] = in[64];
	a[2] = in[128];
	a[3] = in[192];
	fftKernel4(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii & 3;
	i = ii >> 2;
	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 2);
	ang = dir * ( 2.0f * M_PI * 1.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	lMemStore = sMem + ii;
	j = (ii & 15) >> 2;
	i = mad24(ii >> 4, 4, ii & 3);

	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);

	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 4);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii >> 4;
	i = ii & 15;
	lMemLoad = sMem + mad24(j, 64, i);
	lMemStore[0] = a[0].x;
	lMemStore[64] = a[1].x;
	lMemStore[128] = a[2].x;
	lMemStore[192] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[64] = a[1].y;
	lMemStore[128] = a[2].y;
	lMemStore[192] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	out[0] = a[0];
	out[64] = a[1];
	out[128] = a[2];
	out[192] = a[3];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	//in -= offset;
	//out -= offset;

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length

	// clFFT_Forward
	if(dir == -1)
	{		
		int workItemOffset2 = lId * 2 + groupOffset;

		// The first thread for each FFT length
		if(lId == 0)
		{
			speex_out[speex_out_offset + groupOffset + 0] = out_orig[workItemOffset2 +  0].x;
			speex_out[speex_out_offset + groupOffset + 1] = out_orig[workItemOffset2 + 1].x;
			speex_out[speex_out_offset + groupOffset + 2] = out_orig[workItemOffset2 + 1].y;
			speex_out[speex_out_offset + groupOffset + 3] = out_orig[workItemOffset2 + 2].x;
		}
		else
		{
			speex_out[speex_out_offset + workItemOffset] = out_orig[workItemOffset2].y;
			speex_out[speex_out_offset + workItemOffset + 1] = out_orig[workItemOffset2 + 1].x;
			speex_out[speex_out_offset + workItemOffset + 2] = out_orig[workItemOffset2 + 1].y;
			speex_out[speex_out_offset + workItemOffset + 3] = out_orig[workItemOffset2 + 2].x;
		}
	}
	// clFFT_Inverse
	// HERE BE (WAS?) ATI PROBLEMS
	else if(dir == 1)
	{
		speex_out[speex_out_offset + workItemOffset] = out_orig[workItemOffset].x;
		speex_out[speex_out_offset + workItemOffset + 1] = out_orig[workItemOffset + 1].x;
		speex_out[speex_out_offset + workItemOffset + 2] = out_orig[workItemOffset + 2].x;
		speex_out[speex_out_offset + workItemOffset + 3] = out_orig[workItemOffset + 3].x;
	}
	
	// End of finalization step
}

/**
 * Specialized kernel for updating the filter weights in batches
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 256.
 */
__kernel void clFFT_update_weights(__global float *W, __global float2 *in, __global float2 *out, __global float *wtmp, int S)
{
	__local float sMem[272];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	int chan = get_global_id(1);
	int speak = get_global_id(2);
	int num_speakers = get_global_size(2);
	int num_channels = get_global_size(1);
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[4];
	int lId = get_local_id( 0 );
	int groupId = get_group_id( 0 );
	const float m = 1.0f / 256.0f;
	int groupOffset = groupId * 256 * num_speakers;		// Group offset for float
	int groupOffset2 = groupId * 128 * num_speakers;	// Group offset for float2
	unsigned int cs_offset = chan*256*num_speakers*S + speak*256; // Channel/speaker offset
	unsigned int W_offset = cs_offset;
	ii = lId;
	jj = 0;

	// Do IFFT first
	int dir = 1;
	
	offset =  mad24(groupId, 256, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	
	int workItemOffset = lId * 4 + groupOffset + cs_offset;
	
	barrier(CLK_LOCAL_MEM_FENCE);

	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	if(lId == 0)
	{
		in[groupOffset + cs_offset + 0].x = W[groupOffset + cs_offset + 0];
		in[groupOffset + cs_offset + 0].y = 0.0f;
		in[groupOffset + cs_offset + 1].x = W[groupOffset + cs_offset + 1];
		in[groupOffset + cs_offset + 1].y = W[groupOffset + cs_offset + 2];
		in[groupOffset + cs_offset + 128].x = W[groupOffset + cs_offset + 256 - 1];
		in[groupOffset + cs_offset + 128].y = 0.0f;
	}
	else
	{
		in[groupOffset + cs_offset + 2*lId].x = W[groupOffset + cs_offset + lId*4 - 1];
		in[groupOffset + cs_offset + 2*lId].y = W[groupOffset + cs_offset + lId*4];
		in[groupOffset + cs_offset + 2*lId + 1].x = W[groupOffset + cs_offset + lId*4 + 1];
		in[groupOffset + cs_offset + 2*lId + 1].y = W[groupOffset + cs_offset + lId*4 + 2];
	}

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	// Generate negative frequencies for symmetric input 
	// Source: http://www.dspguide.com/ch12/1.htm
	// For the remaining n/2 threads that needs to mirror
	in[groupOffset + cs_offset + 255 - lId*2].x = in[groupOffset + cs_offset + lId*2 + 1].x;
	in[groupOffset + cs_offset + 255 - lId*2].y = -in[groupOffset + cs_offset + lId*2 + 1].y;
	in[groupOffset + cs_offset + 255 - lId*2 - 1].x = in[groupOffset + cs_offset + lId*2 + 2].x;
	in[groupOffset + cs_offset + 255 - lId*2 - 1].y = -in[groupOffset + cs_offset + lId*2 + 2].y;

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	//barrier(CLK_LOCAL_MEM_FENCE);
	
	//offset =  mad24(groupId, 256, ii);
	offset = cs_offset + groupId * 256 * num_speakers + ii;
	in += offset;
	out += offset;

	// End of preparation step

	a[0] = in[0];
	a[1] = in[64];
	a[2] = in[128];
	a[3] = in[192];
	fftKernel4(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii & 3;
	i = ii >> 2;
	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 2);
	ang = dir * ( 2.0f * M_PI * 1.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	lMemStore = sMem + ii;
	j = (ii & 15) >> 2;
	i = mad24(ii >> 4, 4, ii & 3);

	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);

	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 4);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii >> 4;
	i = ii & 15;
	lMemLoad = sMem + mad24(j, 64, i);
	lMemStore[0] = a[0].x;
	lMemStore[64] = a[1].x;
	lMemStore[128] = a[2].x;
	lMemStore[192] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[64] = a[1].y;
	lMemStore[128] = a[2].y;
	lMemStore[192] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	out[0] = a[0];
	out[64] = a[1];
	out[128] = a[2];
	out[192] = a[3];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length

	wtmp[workItemOffset] = out_orig[workItemOffset].x;
	wtmp[workItemOffset + 1] = out_orig[workItemOffset + 1].x;
	wtmp[workItemOffset + 2] = out_orig[workItemOffset + 2].x;
	wtmp[workItemOffset + 3] = out_orig[workItemOffset + 3].x;

	// End of finalization step
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	//barrier(CLK_LOCAL_MEM_FENCE);
	
	// Set to zero
	wtmp[groupOffset + cs_offset + 128 + 2*lId] = 0.0f;
	wtmp[groupOffset + cs_offset + 128 + 2*lId + 1] = 0.0f;
	
	in = in_orig;
	out = out_orig;
	
	// Do forward FFT
	dir = -1;
	
	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	// clFFT_Forward
	in_orig[workItemOffset].x = wtmp[workItemOffset] * m;
	in_orig[workItemOffset].y = 0.0f;
	in_orig[workItemOffset + 1].x = wtmp[workItemOffset + 1] * m;
	in_orig[workItemOffset + 1].y = 0.0f;
	in_orig[workItemOffset + 2].x = wtmp[workItemOffset + 2] * m;
	in_orig[workItemOffset + 2].y = 0.0f;
	in_orig[workItemOffset + 3].x = wtmp[workItemOffset + 3] * m;
	in_orig[workItemOffset + 3].y = 0.0f;	
	
	// End of preparation step
	
	//offset =  mad24(groupId, 256, ii);
	offset = cs_offset + groupId * 256 * num_speakers + ii;
	in += offset;
	out += offset;

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
	
	a[0] = in[0];
	a[1] = in[64];
	a[2] = in[128];
	a[3] = in[192];
	fftKernel4(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii & 3;
	i = ii >> 2;
	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 2);
	ang = dir * ( 2.0f * M_PI * 1.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	lMemStore = sMem + ii;
	j = (ii & 15) >> 2;
	i = mad24(ii >> 4, 4, ii & 3);

	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);

	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 4);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii >> 4;
	i = ii & 15;
	lMemLoad = sMem + mad24(j, 64, i);
	lMemStore[0] = a[0].x;
	lMemStore[64] = a[1].x;
	lMemStore[128] = a[2].x;
	lMemStore[192] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[64] = a[1].y;
	lMemStore[128] = a[2].y;
	lMemStore[192] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	out[0] = a[0];
	out[64] = a[1];
	out[128] = a[2];
	out[192] = a[3];
	
	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
	
	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length
	// Not adapted to multi-channel filter, only mono
	// groupOffset = groupId * 256 * num_speakers;
	int workItemOffset2 = lId * 2 + groupOffset + cs_offset;
	//workItemOffset = lId * 4 + groupOffset;
	//unsigned int W_offset = speak*256 + chan*256*num_speakers*S;
	
	// The first thread for each FFT length
	if(lId == 0)
	{
		W[groupOffset + cs_offset + 0] = out_orig[workItemOffset2 + 0].x;
		W[groupOffset + cs_offset + 1] = out_orig[workItemOffset2 + 1].x;
		W[groupOffset + cs_offset + 2] = out_orig[workItemOffset2 + 1].y;
		W[groupOffset + cs_offset + 3] = out_orig[workItemOffset2 + 2].x;
	}
	else if(groupId < S)
	{
		W[workItemOffset] = out_orig[workItemOffset2].y;
		W[workItemOffset + 1] = out_orig[workItemOffset2 + 1].x;
		W[workItemOffset + 2] = out_orig[workItemOffset2 + 1].y;
		W[workItemOffset + 3] = out_orig[workItemOffset2 + 2].x;
		//W[workItemOffset + W_offset] = get_global_id(1);
		//W[workItemOffset + W_offset + 1] = get_global_id(1);
		//W[workItemOffset + W_offset + 2] = get_global_id(1);
		//W[workItemOffset + W_offset + 3] = get_global_id(1);
	}
	
	// End of finalization step 
}

/**
 * Specialized kernel for converting the error to the frequency domain in one operation 
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 256.
 */
__kernel void clFFT_convert_error_to_freq_domain(__global float *e, __global float *E, __global float *y, __global float *Y, __global float2 *in, __global float2 *in2, __global float2 *out, __global float2 *out2, int S)
{
	__local float sMem[272];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[4];
	int lId = get_local_id(0);
	int taskId = get_global_id(1);
	int groupId = get_group_id(0);
	const float m = 1.0f / 256.0f;
	int groupOffset = groupId * 256;	// Group offset for float
	int groupOffset2 = groupId * 128;	// Group offset for float2
	ii = lId;
	jj = 0;

	offset =  mad24(groupId, 256, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	__global float2 *out2_orig = out2;	// Keep this pointer for finalize code
	__global float2 *in2_orig = in2;
	
	int workItemOffset = lId * 4 + groupOffset;

	barrier(CLK_LOCAL_MEM_FENCE);

	// Do FFT
	// Equiv: speex_echo_opencl_fft_execute2(ocl_e, chan*st->window_size, ocl_E, chan*st->window_size, 1, commandQueue);
	if(taskId == 0)
	{
		// Do FFT first
		int dir = -1;

		// Do preparation step to convert from Speex FFT format
		// There is one thread for each four positions in the FFT length

		// clFFT_Forward
		in_orig[workItemOffset].x = e[workItemOffset] * m;
		in_orig[workItemOffset].y = 0.0f;
		in_orig[workItemOffset + 1].x = e[workItemOffset + 1] * m;
		in_orig[workItemOffset + 1].y = 0.0f;
		in_orig[workItemOffset + 2].x = e[workItemOffset + 2] * m;
		in_orig[workItemOffset + 2].y = 0.0f;
		in_orig[workItemOffset + 3].x = e[workItemOffset + 3] * m;
		in_orig[workItemOffset + 3].y = 0.0f;	
		
		// End of preparation step
		
		barrier(CLK_LOCAL_MEM_FENCE);
		
		offset =  mad24(groupId, 256, ii);
		in += offset;
		out += offset;

		// End of preparation step

		a[0] = in[0];
		a[1] = in[64];
		a[2] = in[128];
		a[3] = in[192];
		fftKernel4(a+0, dir);
		angf = (float) ii;
		ang = dir * ( 2.0f * M_PI * 1.0f / 256.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 256.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 256.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);

		lMemStore = sMem + ii;
		j = ii & 3;
		i = ii >> 2;
		lMemLoad = sMem + mad24(j, 68, i);
		lMemStore[0] = a[0].x;
		lMemStore[68] = a[1].x;
		lMemStore[136] = a[2].x;
		lMemStore[204] = a[3].x;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[68] = a[1].y;
		lMemStore[136] = a[2].y;
		lMemStore[204] = a[3].y;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		angf = (float) (ii >> 2);
		ang = dir * ( 2.0f * M_PI * 1.0f / 64.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 64.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 64.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		lMemStore = sMem + ii;
		j = (ii & 15) >> 2;
		i = mad24(ii >> 4, 4, ii & 3);

		lMemLoad = sMem + mad24(j, 68, i);
		lMemStore[0] = a[0].x;
		lMemStore[68] = a[1].x;
		lMemStore[136] = a[2].x;
		lMemStore[204] = a[3].x;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);

		lMemStore[0] = a[0].y;
		lMemStore[68] = a[1].y;
		lMemStore[136] = a[2].y;
		lMemStore[204] = a[3].y;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		angf = (float) (ii >> 4);
		ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);

		lMemStore = sMem + ii;
		j = ii >> 4;
		i = ii & 15;
		lMemLoad = sMem + mad24(j, 64, i);
		lMemStore[0] = a[0].x;
		lMemStore[64] = a[1].x;
		lMemStore[128] = a[2].x;
		lMemStore[192] = a[3].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[64] = a[1].y;
		lMemStore[128] = a[2].y;
		lMemStore[192] = a[3].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		out[0] = a[0];
		out[64] = a[1];
		out[128] = a[2];
		out[192] = a[3];

		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

		// Do finalization step to convert to Speex FFT format
		// There are one thread for each four positions in the FFT length
		// Not adapted to multi-channel filter, only mono

		int workItemOffset2 = lId * 2 + groupOffset;
		
		// The first thread for each FFT length
		if(lId == 0)
		{
			E[groupOffset + 0] = out_orig[workItemOffset2 + 0].x;
			E[groupOffset + 1] = out_orig[workItemOffset2 + 1].x;
			E[groupOffset + 2] = out_orig[workItemOffset2 + 1].y;
			E[groupOffset + 3] = out_orig[workItemOffset2 + 2].x;
		}
		else
		{
			E[workItemOffset] = out_orig[workItemOffset2].y;
			E[workItemOffset + 1] = out_orig[workItemOffset2 + 1].x;
			E[workItemOffset + 2] = out_orig[workItemOffset2 + 1].y;
			E[workItemOffset + 3] = out_orig[workItemOffset2 + 2].x;
		}
		
		// End of finalization step 
		
	}
	// Set to value and do FFT
	// Equiv: mdf_opencl_set_array_to_float_value(ocl_y, chan*st->window_size, 0, st->frame_size);
	// 	  speex_echo_opencl_fft_execute2(ocl_y, chan*st->window_size, ocl_Y, chan*st->window_size, 1, commandQueue);
	else
	{
		// Set to zero
		y[groupOffset + 2*lId] = 0.0f;
		y[groupOffset + 2*lId + 1] = 0.0f;
		
		// Do FFT
		int dir = -1;

		barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
		
		// Do preparation step to convert from Speex FFT format
		// There is one thread for each four positions in the FFT length

		in2[workItemOffset].x = y[workItemOffset] * m;
        	in2[workItemOffset].y = 0.0f;
		in2[workItemOffset + 1].x = y[workItemOffset + 1] * m;
        	in2[workItemOffset + 1].y = 0.0f;
		in2[workItemOffset + 2].x = y[workItemOffset + 2] * m;
        	in2[workItemOffset + 2].y = 0.0f;
		in2[workItemOffset + 3].x = y[workItemOffset + 3] * m;
        	in2[workItemOffset + 3].y = 0.0f;	

		barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
		
		offset =  mad24(groupId, 256, ii);
		in2 += offset;
		out2 += offset;

		//barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
		
		a[0] = in2[0];
		a[1] = in2[64];
		a[2] = in2[128];
		a[3] = in2[192];
		fftKernel4(a+0, dir);
		angf = (float) ii;
		ang = dir * ( 2.0f * M_PI * 1.0f / 256.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 256.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 256.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);

		lMemStore = sMem + ii;
		j = ii & 3;
		i = ii >> 2;
		lMemLoad = sMem + mad24(j, 68, i);
		lMemStore[0] = a[0].x;
		lMemStore[68] = a[1].x;
		lMemStore[136] = a[2].x;
		lMemStore[204] = a[3].x;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[68] = a[1].y;
		lMemStore[136] = a[2].y;
		lMemStore[204] = a[3].y;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		angf = (float) (ii >> 2);
		ang = dir * ( 2.0f * M_PI * 1.0f / 64.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 64.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 64.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);
		lMemStore = sMem + ii;
		j = (ii & 15) >> 2;
		i = mad24(ii >> 4, 4, ii & 3);

		lMemLoad = sMem + mad24(j, 68, i);
		lMemStore[0] = a[0].x;
		lMemStore[68] = a[1].x;
		lMemStore[136] = a[2].x;
		lMemStore[204] = a[3].x;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);

		lMemStore[0] = a[0].y;
		lMemStore[68] = a[1].y;
		lMemStore[136] = a[2].y;
		lMemStore[204] = a[3].y;
		barrier(CLK_LOCAL_MEM_FENCE);

		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		angf = (float) (ii >> 4);
		ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[1] = complexMul(a[1], w);
		ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[2] = complexMul(a[2], w);
		ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
		w = (float2)(native_cos(ang), native_sin(ang));
		a[3] = complexMul(a[3], w);

		lMemStore = sMem + ii;
		j = ii >> 4;
		i = ii & 15;
		lMemLoad = sMem + mad24(j, 64, i);
		lMemStore[0] = a[0].x;
		lMemStore[64] = a[1].x;
		lMemStore[128] = a[2].x;
		lMemStore[192] = a[3].x;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].x = lMemLoad[0];
		a[1].x = lMemLoad[16];
		a[2].x = lMemLoad[32];
		a[3].x = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		lMemStore[0] = a[0].y;
		lMemStore[64] = a[1].y;
		lMemStore[128] = a[2].y;
		lMemStore[192] = a[3].y;
		barrier(CLK_LOCAL_MEM_FENCE);
		a[0].y = lMemLoad[0];
		a[1].y = lMemLoad[16];
		a[2].y = lMemLoad[32];
		a[3].y = lMemLoad[48];
		barrier(CLK_LOCAL_MEM_FENCE);
		fftKernel4(a+0, dir);
		out2[0] = a[0];
		out2[64] = a[1];
		out2[128] = a[2];
		out2[192] = a[3];
		
		barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
		
		// Do finalization step to convert to Speex FFT format
		// There are one thread for each four positions in the FFT length

		int workItemOffset2 = lId * 2 + groupOffset;

		// The first thread for each FFT length
		if(lId == 0)
		{
			Y[groupOffset + 0] = out2_orig[workItemOffset2 + 0].x;
			Y[groupOffset + 1] = out2_orig[workItemOffset2 + 1].x;
			Y[groupOffset + 2] = out2_orig[workItemOffset2 + 1].y;
			Y[groupOffset + 3] = out2_orig[workItemOffset2 + 2].x;
		}
		else
		{
			Y[workItemOffset] = out2_orig[workItemOffset2].y;
			Y[workItemOffset + 1] = out2_orig[workItemOffset2 + 1].x;
			Y[workItemOffset + 2] = out2_orig[workItemOffset2 + 1].y;
			Y[workItemOffset + 3] = out2_orig[workItemOffset2 + 2].x;
		}

		// End of finalization step
	}
}

/**
 * Specialized kernel for doing a forward FFT and then an inner product
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060,
 * for 1D input of length 256.
 */ 
__kernel void clFFT_inner_prod_combined(__global float *data_time_domain, __global float2 *in, __global float2 *out, __global float *data_freq_domain, __global float *Xf, __global float *temp_reduction, int frame_size, int S)
{
	__local float sMem[272];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	int lId = get_local_id(0);
	int groupId = get_group_id(0);
	int groupOffset = groupId * 256;	// Group offset for float
	int groupOffset2 = groupId * 128;	// Group offset for float2
	unsigned int xi = get_local_id(0)*2 + frame_size + groupOffset;	// Used for inner product
	unsigned int yi = get_local_id(0)*2 + frame_size + groupOffset;	// Used for inner product
	float2 w;
	float ang, angf, ang1;
	__local float *lMemStore, *lMemLoad;
	float2 a[4];
	const float m = 1.0f / 256.0f;
	ii = lId;
	jj = 0;

	// Do FFT
	int dir = -1;
	
	offset =  mad24(groupId, 256, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	__global float2 *in_orig = in;
	
	int workItemOffset = lId * 4 + groupOffset;
	
	barrier(CLK_LOCAL_MEM_FENCE);

	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	in[workItemOffset].x = data_time_domain[workItemOffset] * m;
	in[workItemOffset].y = 0.0f;
	in[workItemOffset + 1].x = data_time_domain[workItemOffset + 1] * m;
	in[workItemOffset + 1].y = 0.0f;
	in[workItemOffset + 2].x = data_time_domain[workItemOffset + 2] * m;
	in[workItemOffset + 2].y = 0.0f;
	in[workItemOffset + 3].x = data_time_domain[workItemOffset + 3] * m;
	in[workItemOffset + 3].y = 0.0f;	
	
	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);
	
	offset =  mad24(groupId, 256, ii);
	in += offset;
	out += offset;

	// End of preparation step

	a[0] = in[0];
	a[1] = in[64];
	a[2] = in[128];
	a[3] = in[192];
	fftKernel4(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii & 3;
	i = ii >> 2;
	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 2);
	ang = dir * ( 2.0f * M_PI * 1.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	lMemStore = sMem + ii;
	j = (ii & 15) >> 2;
	i = mad24(ii >> 4, 4, ii & 3);

	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);

	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 4);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii >> 4;
	i = ii & 15;
	lMemLoad = sMem + mad24(j, 64, i);
	lMemStore[0] = a[0].x;
	lMemStore[64] = a[1].x;
	lMemStore[128] = a[2].x;
	lMemStore[192] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[64] = a[1].y;
	lMemStore[128] = a[2].y;
	lMemStore[192] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	out[0] = a[0];
	out[64] = a[1];
	out[128] = a[2];
	out[192] = a[3];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length

	int workItemOffset2 = lId * 2 + groupOffset;

	// The first thread for each FFT length
	if(lId == 0)
	{
		data_freq_domain[groupOffset + 0] = out_orig[workItemOffset2 +  0].x;
		data_freq_domain[groupOffset + 1] = out_orig[workItemOffset2 + 1].x;
		data_freq_domain[groupOffset + 2] = out_orig[workItemOffset2 + 1].y;
		data_freq_domain[groupOffset + 3] = out_orig[workItemOffset2 + 2].x;
	}
	else
	{
		data_freq_domain[workItemOffset] = out_orig[workItemOffset2].y;
		data_freq_domain[workItemOffset + 1] = out_orig[workItemOffset2 + 1].x;
		data_freq_domain[workItemOffset + 2] = out_orig[workItemOffset2 + 1].y;
		data_freq_domain[workItemOffset + 3] = out_orig[workItemOffset2 + 2].x;
	}

	// End of finalization step

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Calculate inner product (last reduction is done on host)
	if(get_local_id(0) < 64) {
		sMem[lId] = 0.0f;
		//sMem[lId] = fma(data_time_domain[xi], data_time_domain[yi], sMem[lId]);	
		//sMem[lId] = fma(data_time_domain[xi+1], data_time_domain[yi+1], sMem[lId]);
		//sMem[lId] += 1.0f;
		sMem[lId] += data_time_domain[xi] * data_time_domain[yi];	
		sMem[lId] += data_time_domain[xi+1] * data_time_domain[yi+1];
	}
	else
		sMem[lId] = 0.0f;

	barrier(CLK_LOCAL_MEM_FENCE);

	// do reduction in shared mem
	for(unsigned int s = get_local_size(0) / 2; s > 0; s >>= 1) 
	{
		if(lId < s)
		{
			sMem[lId] += sMem[lId + s];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	// write result for this block to global mem
	if (lId == 0) 
	{
		temp_reduction[get_group_id(0)] = sMem[0];
	}

	//barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);
	
	// Power spectrum accum - must do twice the normal work, usually run with 128 threads
	// Loop through output channels, because of lack of atomic adds 
	/*if(get_group_id(0))
	{	
		j = get_local_id(0);
		k = j * 4 - 1;
		l = j * 4 + 1;
		int speak_offset = 0;

		if(lId == 0)
		{
			for(int s=0;s<get_num_groups(0);s++)
			{
				Xf[0] += data_freq_domain[speak_offset + 0] * data_freq_domain[speak_offset + 0];
				Xf[1] += data_freq_domain[speak_offset + 1] * data_freq_domain[speak_offset + 1] + data_freq_domain[speak_offset + 2] * data_freq_domain[speak_offset + 2];
				speak_offset += 2*frame_size;
			}
		}
		else
		{
			for(int s=0;s<get_num_groups(0);s++)
			{
				Xf[2*j] += data_freq_domain[speak_offset + k] * data_freq_domain[speak_offset + k] + data_freq_domain[speak_offset + k + 1] * data_freq_domain[speak_offset + k + 1];
				Xf[2*j + 1] += data_freq_domain[speak_offset + l] * data_freq_domain[speak_offset + l] + data_freq_domain[speak_offset + l + 1] * data_freq_domain[speak_offset + l + 1];
				speak_offset += 2*frame_size;
			}
		}
		// Set the last element (frame_size + 1)
		if(lId == (get_local_size(0) - 1))
		{
			for(int s=0;s<get_num_groups(0);s++)
			{
				Xf[frame_size] += data_freq_domain[speak_offset + (frame_size*2) - 1] * data_freq_domain[speak_offset + (frame_size*2) - 1];
				speak_offset += 2*frame_size;
			}
		}
	}*/
}

/**
 * Regular inverse FFT, combined with the spectral mul accum (long) reduce kernel
 * Kernel output from the Apple OpenCL FFT lib on a NVIDIA Tesla C1060, 
 * for 1D input of length 256.
 */ 
__kernel void clFFT_inverse_spectral_mul_accum_reduce(__global float *Y, __global float2 *in, __global float2 *out, __global float *e, __global float *temp_acc, int M, int N, int S)
{
	__local float sMem[272];
	int i, j, r, indexIn, indexOut, index, tid, bNum, xNum, k, l;
	int s, ii, jj, offset;
	float2 w;
	float ang, angf, ang1;
	int dir = 1;	// Inverse
	__local float *lMemStore, *lMemLoad;
	float2 a[4];
	int lId = get_local_id( 0 );
	int groupId = get_group_id( 0 );
	const float m = 1.0f / 256.0f;
	int groupOffset = groupId * 256;
	ii = lId;
	jj = 0;

	offset =  mad24(groupId, 256, ii);
	__global float2 *out_orig = out;	// Keep this pointer for finalize code
	int workItemOffset = lId * 4 + groupOffset;
	

	// Do the spectral mul accum reduce, 4 elementents for each thread
	Y[groupOffset + lId*4] = 0.0f;
	Y[groupOffset + lId*4 + 1] = 0.0f;
	Y[groupOffset + lId*4 + 2] = 0.0f;
	Y[groupOffset + lId*4 + 3] = 0.0f;

	for(int c=0;c<M;c++)
	{
		Y[groupOffset + lId*4] += temp_acc[groupId*M*N + c*N + lId*4];
		Y[groupOffset + lId*4 + 1] += temp_acc[groupId*M*N + c*N + lId*4 + 1];
		Y[groupOffset + lId*4 + 2] += temp_acc[groupId*M*N + c*N + lId*4 + 2];
		Y[groupOffset + lId*4 + 3] += temp_acc[groupId*M*N + c*N + lId*4 + 3];
	}

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	// Do preparation step to convert from Speex FFT format
	// There is one thread for each four positions in the FFT length

	if(lId == 0)
	{
		in[groupOffset + 0].x = Y[groupOffset + 0];
		in[groupOffset + 0].y = 0.0f;
		in[groupOffset + 1].x = Y[groupOffset + 1];
		in[groupOffset + 1].y = Y[groupOffset + 2];
		in[groupOffset + 128].x = Y[groupOffset + 256 - 1];
		in[groupOffset + 128].y = 0.0f;
	}
	else
	{
		in[groupOffset + 2*lId].x = Y[groupOffset + lId*4 - 1];
		in[groupOffset + 2*lId].y = Y[groupOffset + lId*4];
		in[groupOffset + 2*lId + 1].x = Y[groupOffset + lId*4 + 1];
		in[groupOffset + 2*lId + 1].y = Y[groupOffset + lId*4 + 2];
	}

	barrier(CLK_GLOBAL_MEM_FENCE | CLK_LOCAL_MEM_FENCE);

	// Generate negative frequencies for symmetric input 
	// Source: http://www.dspguide.com/ch12/1.htm
	// For the remaining n/2 threads that needs to mirror
	/*in[groupOffset + 256 - lId*2].x = in[groupOffset + lId*2].x;
	in[groupOffset + 256 - lId*2].y = -in[groupOffset + lId*2].y;
	in[groupOffset + 256 - lId*2 - 1].x = in[groupOffset + lId*2 + 1].x;
	in[groupOffset + 256 - lId*2 - 1].y = -in[groupOffset + lId*2 + 1].y;*/
	
	in[groupOffset + 255 - lId*2].x = in[groupOffset + lId*2 + 1].x;
	in[groupOffset + 255 - lId*2].y = -in[groupOffset + lId*2 + 1].y;
	in[groupOffset + 255 - lId*2 - 1].x = in[groupOffset + lId*2 + 2].x;
	in[groupOffset + 255 - lId*2 - 1].y = -in[groupOffset + lId*2 + 2].y;

	barrier(CLK_LOCAL_MEM_FENCE);

	in += offset;
	out += offset;

	// End of preparation step

	a[0] = in[0];
	a[1] = in[64];
	a[2] = in[128];
	a[3] = in[192];
	fftKernel4(a+0, dir);
	angf = (float) ii;
	ang = dir * ( 2.0f * M_PI * 1.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 256.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii & 3;
	i = ii >> 2;
	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 2);
	ang = dir * ( 2.0f * M_PI * 1.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 64.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);
	lMemStore = sMem + ii;
	j = (ii & 15) >> 2;
	i = mad24(ii >> 4, 4, ii & 3);

	lMemLoad = sMem + mad24(j, 68, i);
	lMemStore[0] = a[0].x;
	lMemStore[68] = a[1].x;
	lMemStore[136] = a[2].x;
	lMemStore[204] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);

	lMemStore[0] = a[0].y;
	lMemStore[68] = a[1].y;
	lMemStore[136] = a[2].y;
	lMemStore[204] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);

	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	angf = (float) (ii >> 4);
	ang = dir * ( 2.0f * M_PI * 1.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[1] = complexMul(a[1], w);
	ang = dir * ( 2.0f * M_PI * 2.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[2] = complexMul(a[2], w);
	ang = dir * ( 2.0f * M_PI * 3.0f / 16.0f ) * angf;
	w = (float2)(native_cos(ang), native_sin(ang));
	a[3] = complexMul(a[3], w);

	lMemStore = sMem + ii;
	j = ii >> 4;
	i = ii & 15;
	lMemLoad = sMem + mad24(j, 64, i);
	lMemStore[0] = a[0].x;
	lMemStore[64] = a[1].x;
	lMemStore[128] = a[2].x;
	lMemStore[192] = a[3].x;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].x = lMemLoad[0];
	a[1].x = lMemLoad[16];
	a[2].x = lMemLoad[32];
	a[3].x = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	lMemStore[0] = a[0].y;
	lMemStore[64] = a[1].y;
	lMemStore[128] = a[2].y;
	lMemStore[192] = a[3].y;
	barrier(CLK_LOCAL_MEM_FENCE);
	a[0].y = lMemLoad[0];
	a[1].y = lMemLoad[16];
	a[2].y = lMemLoad[32];
	a[3].y = lMemLoad[48];
	barrier(CLK_LOCAL_MEM_FENCE);
	fftKernel4(a+0, dir);
	out[0] = a[0];
	out[64] = a[1];
	out[128] = a[2];
	out[192] = a[3];

	barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);

	// Do finalization step to convert to Speex FFT format
	// There are one thread for each four positions in the FFT length

	// clFFT_Inverse
	e[workItemOffset] = out_orig[workItemOffset].x;
	e[workItemOffset + 1] = out_orig[workItemOffset + 1].x;
	e[workItemOffset + 2] = out_orig[workItemOffset + 2].x;
	e[workItemOffset + 3] = out_orig[workItemOffset + 3].x;
	
	// End of finalization step
}
