/* Copyright (C) 2004 Jean-Marc Valin */
/**
   @file mdf_opencl.h
   @brief OpenCL implementation of the MDF algorithm
*/
/*
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <CL/cl.h>
#include <string.h>
#include <stdlib.h>
#include <fftw3.h>
#include <float.h>
#include <sys/time.h>
//#include "appleOclFFT/clFFT.h"

#define CL_IS_INCLUDED 1
//#define SKIP_OCL_STATE_SYNC 1	// Uncomment to skip memory copy of the complete state for each new frame
#define eps_avg 10.0

#define WORD2INT(x) ((x) < -32767.5f ? -32768 : ((x) > 32766.5f ? 32767 : floor(.5+(x))))  

/*** GLOBALS ***/

/*
 * Input data is stored here.
 */
cl_uint *input;

/*
 * Output data is stored here.
 */
cl_uint *output;

/*
 * Multiplier is stored in this variable
 */
cl_uint multiplier;

/* problem size for 1D algorithm and width of problem size for 2D algorithm */
cl_uint width;

/* The memory buffer that is used as input/output for OpenCL kernel */
cl_mem   inputBuffer;
cl_mem   outputBuffer;

cl_context          context;
cl_device_id        device;
//cl_command_queue    commandQueue;
cl_device_id	*devices;

cl_program program;

cl_program clMDF_program;
cl_kernel MDF_inner_product_reduce_kernel;
cl_kernel MDF_spectral_mul_accum_kernel;
cl_kernel MDF_compute_foreground_filter_kernel;
cl_kernel MDF_adjust_prop_phase1_reduce_kernel;
cl_kernel MDF_weighted_spectral_mul_conj_kernel;
cl_kernel MDF_set_array_to_float_value_kernel;
cl_kernel MDF_copy_float_array_kernel;
cl_kernel MDF_update_foreground_filter_kernel;
cl_kernel MDF_update_background_filter_kernel;
cl_kernel MDF_smooth_far_end_compute_filtered_spectra_kernel;
cl_kernel MDF_learning_rate_calc_kernel;
cl_kernel MDF_adapted_copy_kernel;
// Combined kernels
cl_kernel MDF_prepare_shift_preemph_freq_data_kernel;
cl_kernel MDF_power_spectrum_inner_product_kernel;
cl_kernel MDF_response_diff_inner_product_kernel;
cl_kernel MDF_error_signal_combined_kernel; 
cl_kernel MDF_inner_prod_power_spec_spectral_mul_kernel;

/* This program uses only one kernel and this serves as a handle to it */
cl_kernel  kernel;

cl_mem ocl_in = NULL;
cl_mem ocl_far_end = NULL;
cl_mem ocl_input = NULL;
cl_mem ocl_out = NULL;
cl_mem ocl_notch_mem = NULL;
cl_mem ocl_memD = NULL;
cl_mem ocl_memE = NULL;
cl_mem ocl_memX = NULL;
cl_mem ocl_Xf = NULL;
cl_mem ocl_Rf = NULL;
cl_mem ocl_Yf = NULL;
cl_mem ocl_Yh = NULL;
cl_mem ocl_Eh = NULL;
cl_mem ocl_Y = NULL;
cl_mem ocl_foreground = NULL;
cl_mem ocl_e = NULL;
cl_mem ocl_temp_e = NULL;
cl_mem ocl_W = NULL;
cl_mem ocl_prop = NULL; 
cl_mem ocl_power = NULL;
cl_mem ocl_power_1 = NULL;
cl_mem ocl_E = NULL;
cl_mem ocl_PHI = NULL;
cl_mem ocl_wtmp = NULL;
cl_mem ocl_y = NULL;
cl_mem ocl_window = NULL;
cl_mem ocl_saturated = NULL;
cl_mem ocl_Pey = NULL;
cl_mem ocl_Pyy = NULL;
cl_mem ocl_last_y = NULL;
cl_mem ocl_temp_acc = NULL;

cl_mem data_time_domain = NULL;
cl_mem data_freq_domain = NULL;
float *data_time_domain_host = NULL;
float *data_freq_domain_host = NULL;
cl_mem data_time_domain_temp = NULL;
cl_mem data_freq_domain_temp = NULL;
cl_mem ocl_temp_buffer = NULL;	// Temporary buffer at a convinient size
cl_mem ocl_temp_reduction_buffer = NULL;
int *out_host_dummy = NULL;	// Dummy output buffer for timing
float *Pey_Pyy_reduction_host = NULL;

float *error_signal_block_sums;
float *error_signal_temp;
int *error_signal_out_int;
float *host_temp_memD;

// Returns the time difference in seconds
double subtractTimes(struct timeval end, struct timeval start)
{
        long long tusecs = (long long) (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec);
        return (double)(tusecs/1000000.0);
}

void notify_callback(const char *errinfo, const void *private_info, size_t cb, void *user_data)
{
	printf( "%s\n", errinfo );
}

/**
 * Only initializes devices and compiles kernel source code
 * Split from memory allocation for benchmarking purposes (compiling the kernels can take a long time)
 */
EXPORT void speex_echo_opencl_init_dev(int frame_size)
{	
	commandQueue = (cl_command_queue *)malloc(sizeof(cl_command_queue));
	if(initializeCL(commandQueue) != 0)
	{
		printf("Failed to initialize OpenCL(device init and kernel compilation), exiting.\n");
		exit(-1);
	}
	
	if(speex_echo_opencl_fft_initialize_dev(frame_size*2, commandQueue) != 0)
	{
		printf("Failed to initialize OpenCL FFT (kernel compilation), exiting.\n");
		exit(-1);
	}
}

EXPORT void speex_echo_opencl_init_mem(SpeexEchoState *st)
{	
	if(speex_echo_opencl_allocate_common_memory(st, commandQueue) != 0)
	{
		printf("Failed to initialize OpenCL common memory, exiting.\n");
		exit(-1);
	}

	// Find max batch size (largest of number of input/output channels or length of echo tail)
	int max_batch_size = st->M;
	if(st->C > max_batch_size)
		max_batch_size = st->C;
	if(st->K > max_batch_size)
		max_batch_size = st->K;
	
	if(speex_echo_opencl_fft_initialize_mem(max_batch_size, st->K*st->C, st->window_size, commandQueue) != 0)
	{
		printf("Failed to initialize OpenCL FFT memory, exiting.\n");
		exit(-1);
	}
}

EXPORT void speex_echo_opencl_cleanup_mem()
{
	cl_int err;
	err=clFinish(*commandQueue);
        print_on_opencl_error("clFinish failed in cleanup function.", err);
	speex_echo_opencl_fft_cleanup_mem();
	printf("Cleaned up FFT memory.\n");
	speex_echo_opencl_cleanup_common_mem(commandQueue);
	printf("Cleaned up MDF memory.\n");
}

EXPORT void speex_echo_opencl_cleanup_dev()
{
	cl_int err;
	err=clFinish(*commandQueue);
        print_on_opencl_error("clFinish failed in cleanup function.", err);
	speex_echo_opencl_fft_cleanup_dev();
	printf("Cleaned up FFT kernels.\n");
	speex_echo_opencl_cleanup_common_dev(commandQueue);
	printf("Cleaned up MDF program, context, commandqueue and kernels.\n");
}

/**
 * OpenCL related initialization
 * Create Context, Device list, Command Queue
 * Create OpenCL memory buffer objects
 * Load CL file, compile, link CL source
 * Build program and kernel objects
 */
int initializeCL(cl_command_queue *commandQueue)
{
	cl_int status = 0;
	size_t deviceListSize;
	unsigned int i;
	unsigned int num_devices;
	int contextTries = 0;

	cl_device_type device_type = CL_DEVICE_TYPE_GPU;
        cl_device_id device_id;
	cl_platform_id platform;
        cl_uint numPlatforms;
	cl_context_properties cps[3];
	
	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	if(status != CL_SUCCESS)
	{
		printf("clGetPlatformIDs failed\n");
		return -1;
	}
	cl_platform_id* platforms = (cl_platform_id* )malloc(sizeof(cl_platform_id)*numPlatforms);
	
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(status)
        {
                printf("clGetPlatformIDs failed\n");
                return -1;
        }
        else {
		printf("Got OpenCL platform IDs. Numplatforms: %d\n", numPlatforms);
	}

	status = CL_SUCCESS;
	
	// Loop through platforms until we find the first with the required device type (at random)
	// Made for supporting a combined AMD/NVIDIA setup with only one GPU
	// Customize if using a multiple GPU setup
	do
	{
		platform = platforms[contextTries];
		cps[0] = CL_CONTEXT_PLATFORM;
		cps[1] = (cl_context_properties)platform;
		cps[2] = 0;
		// Create an OpenCL context
		context = clCreateContextFromType(cps,
					      device_type,
					      NULL,
					      NULL,
					      &status);
		contextTries++;
	}
	while(contextTries < numPlatforms && status != CL_SUCCESS);
	  
	if(status != CL_SUCCESS)
        {
                printf("Error: Creating Context. (clCreateContextFromType)\n");
                return -1;
        }

	// First, get the size of device list data
    	status = clGetContextInfo(context,
                              CL_CONTEXT_DEVICES,
                              0,
                              NULL,
                              &deviceListSize);
    	if(status != CL_SUCCESS)
        {
                printf(
                        "Error: Getting Context Info \
                    (device list size, clGetContextInfo)\n");
                return;
        }
	
        // Detect OpenCL devices
    	devices = (cl_device_id *)malloc(deviceListSize);
        if(devices == 0)
        {
                printf("Error: No devices found.\n");
                return;
        }

    	/* Now, get the device list data */
    	status = clGetContextInfo(
                             context,
                 CL_CONTEXT_DEVICES,
                 deviceListSize,
                 devices,
                 NULL);
    	if(status != CL_SUCCESS)
        {
                printf("Error: Getting Context Info \
                    (device list, clGetContextInfo)\n");
                return;
        }
	device_id = devices[0];

	// Create an OpenCL command queue
	*commandQueue = clCreateCommandQueue(
			context,
			devices[0],
			0,
			&status);
	if(status != CL_SUCCESS)
	{
		printf("Error: Creating Command Queue. (clCreateCommandQueue)\n");
		return -1;
	}
	else
	{
		printf("SUCCESS: Creating Command Queue. (clCreateCommandQueue)\n");
	}

	// Load MDF kernels
	char *print_string;
	print_string = (char *)malloc(sizeof(char) * 300);
        char *clMDF_kernel_filename = "mdf_oclmdf_kernel.cl";
        int clMDF_kernel_filesize = getFileSize(clMDF_kernel_filename);

        if(clMDF_kernel_filesize < 0) {
                sprintf(print_string, "Could not find file with MDF kernels (in %s), exiting.\n", clMDF_kernel_filename);
                print_error(print_string, 0);
                return -1;
        }

        char *clMDF_kernel_string = (char *)malloc(sizeof(char) * clMDF_kernel_filesize);
        size_t clMDF_kernel_length = 0;
        int clMDF_kernel_file_read = convertToString(clMDF_kernel_filename, clMDF_kernel_string, &clMDF_kernel_length);

        if(clMDF_kernel_file_read == 0)
        {
		// Uncomment to print kernel strings at startup
                //sprintf(print_string, "Kernel string (length %d):", (int)clMDF_kernel_length);
                //print_debug(print_string, 2);
                //print_debug(clMDF_kernel_string, 3);
		// Marker in console output
		print_debug("------------------------ READ FILES ------------------------", 2);

                clMDF_program = clCreateProgramWithSource(context, 1, (const char **)&clMDF_kernel_string, &clMDF_kernel_length, &status);
                print_on_opencl_error("clCreateProgramWithSource of clMDF_kernel_string failed", status);
                if(status == CL_SUCCESS)
                        print_debug("SUCCESS: clCreateProgramWithSource for MDF program", 3);

                //status = clBuildProgram(clMDF_program, 0, NULL, "-cl-mad-enable", NULL, NULL);
		status = clBuildProgram(clMDF_program, 0, NULL, "", NULL, NULL);
                print_on_opencl_error("clBuildProgram of clMDF_program failed.", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clBuildProgram of clMDF_program\n");

                print_opencl_build_info(&clMDF_program);

		MDF_inner_product_reduce_kernel = clCreateKernel(clMDF_program, "clMDF_inner_product_reduce", &status);
                print_on_opencl_error("clCreateKernel of clMDF_inner_product_reduce from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_inner_product_reduce\n");
 
		MDF_spectral_mul_accum_kernel = clCreateKernel(clMDF_program, "clMDF_spectral_mul_accum", &status);
                print_on_opencl_error("clCreateKernel of clMDF_spectral_mul_accum from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_spectral_mul_accum\n"); 

		MDF_compute_foreground_filter_kernel = clCreateKernel(clMDF_program, "clMDF_compute_foreground_filter", &status);
                print_on_opencl_error("clCreateKernel of clMDF_compute_foreground_filter from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_compute_foreground_filter\n");

		MDF_adjust_prop_phase1_reduce_kernel = clCreateKernel(clMDF_program, "clMDF_adjust_prop_phase1_reduce", &status);
                print_on_opencl_error("clCreateKernel of clMDF_adjust_prop_phase1_reduce from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_adjust_prop_phase1_reduce\n");
		
		MDF_weighted_spectral_mul_conj_kernel = clCreateKernel(clMDF_program, "clMDF_weighted_spectral_mul_conj", &status);
                print_on_opencl_error("clCreateKernel of clMDF_weighted_spectral_mul_conj from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_weighted_spectral_mul_conj\n");

		MDF_set_array_to_float_value_kernel = clCreateKernel(clMDF_program, "clMDF_set_array_to_float_value", &status);
                print_on_opencl_error("clCreateKernel of clMDF_set_array_to_float_value from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_set_array_to_float_value\n");
		
		MDF_copy_float_array_kernel = clCreateKernel(clMDF_program, "clMDF_copy_float_array", &status);
                print_on_opencl_error("clCreateKernel of clMDF_copy_float_array from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_copy_float_array\n");
		
		MDF_update_foreground_filter_kernel = clCreateKernel(clMDF_program, "clMDF_update_foreground_filter", &status);
                print_on_opencl_error("clCreateKernel of clMDF_update_foreground_filter from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_update_foreground_filter\n");
		
		MDF_update_background_filter_kernel = clCreateKernel(clMDF_program, "clMDF_update_background_filter", &status);
                print_on_opencl_error("clCreateKernel of clMDF_update_background_filter from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_update_background_filter\n");
		
		MDF_smooth_far_end_compute_filtered_spectra_kernel = clCreateKernel(clMDF_program, "clMDF_smooth_far_end_compute_filtered_spectra", &status);
                print_on_opencl_error("clCreateKernel of clMDF_smooth_far_end_compute_filtered_spectra from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_smooth_far_end_compute_filtered_spectra\n");
		
		MDF_learning_rate_calc_kernel = clCreateKernel(clMDF_program, "clMDF_learning_rate_calc", &status);
                print_on_opencl_error("clCreateKernel of clMDF_learning_rate_calc from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_learning_rate_calc\n");
		
		MDF_adapted_copy_kernel = clCreateKernel(clMDF_program, "clMDF_adapted_copy", &status);
                print_on_opencl_error("clCreateKernel of clMDF_adapted_copy from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_adapted_copy\n");
		
		MDF_prepare_shift_preemph_freq_data_kernel = clCreateKernel(clMDF_program, "clMDF_prepare_shift_preemph_freq_data", &status);
                print_on_opencl_error("clCreateKernel of clMDF_prepare_shift_preemph_freq_data from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_prepare_shift_preemph_freq_data\n");
		
		MDF_power_spectrum_inner_product_kernel = clCreateKernel(clMDF_program, "clMDF_power_spectrum_inner_product", &status);
                print_on_opencl_error("clCreateKernel of clMDF_power_spectrum_inner_product from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_power_spectrum_inner_product\n");

		MDF_response_diff_inner_product_kernel = clCreateKernel(clMDF_program, "clMDF_combined_response_diff_inner_product", &status);
                print_on_opencl_error("clCreateKernel of clMDF_response_diff_inner_product from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_response_diff_inner_product\n");

		MDF_error_signal_combined_kernel = clCreateKernel(clMDF_program, "clMDF_error_signal_combined", &status);
                print_on_opencl_error("clCreateKernel of clMDF_error_signal_combined from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_error_signal_combined\n");
		
		MDF_inner_prod_power_spec_spectral_mul_kernel = clCreateKernel(clMDF_program, "clMDF_inner_prod_power_spec_spectral_mul", &status);
                print_on_opencl_error("clCreateKernel of clMDF_inner_prod_power_spec_spectral_mul from clMDF_program failed", status);
                if(status == CL_SUCCESS)
                        printf("SUCCESS: clCreateKernel of clMDF_inner_prod_power_spec_spectral_mul\n");
	}
        else
        {
                sprintf(print_string, "Could not open file with MDF kernels (in %s), exiting.\n", clMDF_kernel_filename);
                print_error(print_string, 1);
        }

        free(clMDF_kernel_string);
	free(platforms);
	free(devices);
	
	return 0;
}

int speex_echo_opencl_allocate_common_memory(SpeexEchoState *st, cl_command_queue *commandQueue)
{
	cl_int err = CL_SUCCESS;

        data_time_domain = clCreateBuffer(context, CL_MEM_READ_WRITE, st->K*st->window_size*sizeof(cl_float), NULL, &err);
	if(!data_time_domain || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }

	data_time_domain_host = (float *)malloc(st->K*st->window_size*sizeof(float));

	data_time_domain_temp = clCreateBuffer(context, CL_MEM_READ_WRITE, st->K*st->window_size*sizeof(cl_float), NULL, &err);
	if(!data_time_domain_temp || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }

        data_freq_domain = clCreateBuffer(context, CL_MEM_READ_WRITE, st->K*(st->M+1)*st->window_size*sizeof(float)*2, NULL, &err);
        if(!data_freq_domain || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }

	ocl_in = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->frame_size*sizeof(cl_short), NULL, &err);
	if(!ocl_in || err)
	{
		print_error("clCreateBuffer of ocl_in failed", 0);
	}
	
	ocl_far_end = clCreateBuffer(context, CL_MEM_READ_WRITE, st->K*st->frame_size*sizeof(cl_short), NULL, &err);
	if(!ocl_far_end || err)
	{
		print_error("clCreateBuffer of ocl_far_end failed", 0);
	}

	ocl_input = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->frame_size*sizeof(cl_float), NULL, &err);
	if(!ocl_input || err)
	{
		print_error("clCreateBuffer of ocl_input failed", 0);
	}

	ocl_out = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->frame_size*sizeof(cl_int), NULL, &err);
	if(!ocl_out || err)
	{
		print_error("clCreateBuffer of ocl_out failed", 0);
	}

	ocl_notch_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, 2*st->C*sizeof(cl_float), NULL, &err);
	if(!ocl_notch_mem || err)
	{
		print_error("clCreateBuffer of ocl_mem failed", 0);
	}

	ocl_memD = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*sizeof(cl_float), NULL, &err);
	if(!ocl_memD || err)
	{
		print_error("clCreateBuffer of ocl_memD failed", 0);
	}

	ocl_memE = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*sizeof(cl_float), NULL, &err);
	if(!ocl_memE || err)
	{
		print_error("clCreateBuffer of ocl_memE failed", 0);
	}

	ocl_memX = clCreateBuffer(context, CL_MEM_READ_WRITE, st->K*sizeof(cl_float), NULL, &err);
	if(!ocl_memX || err)
	{
		print_error("clCreateBuffer of ocl_memX failed", 0);
	}

	ocl_Xf = clCreateBuffer(context, CL_MEM_READ_WRITE, (st->frame_size+1)*sizeof(cl_float), NULL, &err);
	if(!ocl_Xf || err)
	{
		print_error("clCreateBuffer of ocl_Xf failed", 0);
	}

	ocl_Rf = clCreateBuffer(context, CL_MEM_READ_WRITE, (st->frame_size+1)*sizeof(cl_float), NULL, &err);
	if(!ocl_Rf || err)
	{
		print_error("clCreateBuffer of ocl_Rf failed", 0);
	}

	ocl_Yf = clCreateBuffer(context, CL_MEM_READ_WRITE, (st->frame_size+1)*sizeof(cl_float), NULL, &err);
	if(!ocl_Yf || err)
	{
		print_error("clCreateBuffer of ocl_Yf failed", 0);
	}
	
	ocl_Yh = clCreateBuffer(context, CL_MEM_READ_WRITE, (st->frame_size+1)*sizeof(cl_float), NULL, &err);
	if(!ocl_Yh || err)
	{
		print_error("clCreateBuffer of ocl_Yh failed", 0);
	}
	
	ocl_Eh = clCreateBuffer(context, CL_MEM_READ_WRITE, (st->frame_size+1)*sizeof(cl_float), NULL, &err);
	if(!ocl_Eh || err)
	{
		print_error("clCreateBuffer of ocl_Eh failed", 0);
	}

	ocl_Y = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_Y || err)
	{
		print_error("clCreateBuffer of ocl_Y failed", 0);
	}

	ocl_foreground = clCreateBuffer(context, CL_MEM_READ_WRITE, st->M*st->window_size*st->C*st->K*sizeof(cl_float), NULL, &err);
	if(!ocl_foreground || err)
	{
		print_error("clCreateBuffer of ocl_foreground failed", 0);
	}

	ocl_e = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_e || err)
	{
		print_error("clCreateBuffer of ocl_e failed", 0);
	}

	ocl_temp_e = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_temp_e || err)
	{
		print_error("clCreateBuffer of ocl_temp_e failed", 0);
	}

	ocl_W = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->K*st->M*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_W || err)
	{
		print_error("clCreateBuffer of ocl_W failed", 0);
	}

	ocl_prop = clCreateBuffer(context, CL_MEM_READ_WRITE, st->M*sizeof(cl_float), NULL, &err);
	if(!ocl_prop || err)
	{
		print_error("clCreateBuffer of ocl_prop failed", 0);
	}
	
	ocl_power= clCreateBuffer(context, CL_MEM_READ_WRITE, (st->frame_size+1)*sizeof(cl_float), NULL, &err);
	if(!ocl_power || err)
	{
		print_error("clCreateBuffer of ocl_power failed", 0);
	}
	
	ocl_power_1 = clCreateBuffer(context, CL_MEM_READ_WRITE, (st->frame_size+1)*sizeof(cl_float), NULL, &err);
	if(!ocl_power_1 || err)
	{
		print_error("clCreateBuffer of ocl_power_1 failed", 0);
	}
	
	ocl_E = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_E || err)
	{
		print_error("clCreateBuffer of ocl_E failed", 0);
	}
	
	ocl_PHI = clCreateBuffer(context, CL_MEM_READ_WRITE, st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_PHI || err)
	{
		print_error("clCreateBuffer of ocl_PHI failed", 0);
	}

	// Size multiplied by M/C/K to allow for parallel update of the weights
	ocl_wtmp = clCreateBuffer(context, CL_MEM_READ_WRITE, st->window_size*st->M*st->K*st->C*sizeof(cl_float), NULL, &err);
	if(!ocl_wtmp || err)
	{
		print_error("clCreateBuffer of ocl_wtmp failed", 0);
	}
	
	ocl_y = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_y || err)
	{
		print_error("clCreateBuffer of ocl_y failed", 0);
	}
	
	ocl_last_y = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_last_y || err)
	{
		print_error("clCreateBuffer of ocl_last_y failed", 0);
	}
	
	ocl_window = clCreateBuffer(context, CL_MEM_READ_WRITE, st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_window || err)
	{
		print_error("clCreateBuffer of ocl_window failed", 0);
	}

	ocl_saturated = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(int), NULL, &err);
	if(!ocl_saturated || err)
	{
		print_error("clCreateBuffer of ocl_saturated failed", 0);
	}
	
	ocl_Pey = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_float), NULL, &err);
	if(!ocl_Pey || err)
	{
		print_error("clCreateBuffer of ocl_Pey failed", 0);
	}
	
	ocl_Pyy = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_float), NULL, &err);
	if(!ocl_Pyy || err)
	{
		print_error("clCreateBuffer of ocl_Pyy failed", 0);
	}

	// Temporary buffer of convinient size (here: window size (usually 256) * number of channels (usually 1))	
	ocl_temp_buffer = clCreateBuffer(context, CL_MEM_READ_WRITE, st->window_size*st->C*2*sizeof(cl_float), NULL, &err);
	if(!ocl_temp_buffer || err)
	{
		print_error("clCreateBuffer of ocl_temp_buffer failed", 0);
	}

	// Temporary buffer of convinient size for reductions (here: window size)	
	ocl_temp_reduction_buffer = clCreateBuffer(context, CL_MEM_READ_WRITE, (st->window_size+2)*sizeof(cl_float), NULL, &err);
	if(!ocl_temp_reduction_buffer || err)
	{
		print_error("clCreateBuffer of ocl_temp_reduction_buffer failed", 0);
	}

	// Temporary buffer of convinient size for reductions (here: window size)	
	ocl_temp_acc = clCreateBuffer(context, CL_MEM_READ_WRITE, st->C*st->M*st->K*st->window_size*sizeof(cl_float), NULL, &err);
	if(!ocl_temp_acc || err)
	{
		print_error("clCreateBuffer of ocl_temp_acc failed", 0);
	}

	data_freq_domain_host = (float *)malloc(st->K*(st->M+1)*st->window_size*sizeof(float)*2);
	
	out_host_dummy = (float *)malloc(st->C*st->frame_size*sizeof(cl_int));

	data_freq_domain_temp = clCreateBuffer(context, CL_MEM_READ_WRITE, st->K*(st->M+1)*st->window_size*sizeof(cl_float)*2, NULL, &err);
        if(!data_freq_domain_temp || err)
        {
                print_error("clCreateBuffer failed", 0);
                return -1;
        }
        
        Pey_Pyy_reduction_host = (float *)malloc(sizeof(float)*(st->frame_size+1)*2);

	// Temporary buffer used in error_signal_combined
	error_signal_block_sums = (float *)malloc(sizeof(float) * st->frame_size/128 * 3 * st->C);
	error_signal_temp = (float *)malloc(sizeof(float) * st->frame_size * st->C);
	error_signal_out_int = (int *)malloc(sizeof(int) * st->frame_size * st->C);
	host_temp_memD = (float *)malloc(sizeof(float)*st->C);
	
	return 0;
}

/**
 * Copies data from the host-memory st-struct to GPU memory to sync the state
 */
int speex_echo_opencl_copy_to_GPU(SpeexEchoState *st, const spx_int16_t *in, const spx_int16_t *far_end, spx_int16_t *out)
{
	cl_int err = CL_SUCCESS;

	err = clEnqueueWriteBuffer(*commandQueue, ocl_in, CL_FALSE, 0, st->C*st->frame_size*sizeof(cl_short), in, 0, NULL, NULL);
        print_on_opencl_error("clEnqueueWriteBuffer of ocl_in failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_far_end, CL_FALSE, 0, st->K*st->frame_size*sizeof(cl_short), far_end, 0, NULL, NULL);
        print_on_opencl_error("clEnqueueWriteBuffer of ocl_far_end failed", err);
	
	// Done on CPU, so must be copied over
	err = clEnqueueWriteBuffer(*commandQueue, ocl_notch_mem, CL_TRUE, 0, 2*st->C*sizeof(cl_float), st->notch_mem, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_notch_mem failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_input, CL_TRUE, 0, st->C*st->frame_size*sizeof(cl_float), st->input, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_input failed", err);

#ifdef SKIP_OCL_STATE_SYNC
	// Always copy on the first frame
	if(st->cancel_count == 1)
	{
#endif
		printf("Doing full sync of CPU/GPU state.\n");
		err = clEnqueueWriteBuffer(*commandQueue, ocl_out, CL_TRUE, 0, st->C*st->frame_size*sizeof(cl_int), out, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_out failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_memD, CL_TRUE, 0, st->C*sizeof(cl_float), st->memD, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_memD failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_memE, CL_TRUE, 0, st->C*sizeof(cl_float), st->memE, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_memE failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_memX, CL_TRUE, 0, st->K*sizeof(cl_float), st->memX, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_memX failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_Xf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Xf, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_Xf failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_Rf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Rf, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_Rf failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_Yf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Yf, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_Yf failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_Yh, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Yh, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_Yh failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_Eh, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Eh, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_Eh failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_foreground, CL_TRUE, 0, st->M*st->window_size*st->C*st->K*sizeof(cl_float), st->foreground, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_foreground failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_e, CL_TRUE, 0, st->C*st->window_size*sizeof(cl_float), st->e, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_e failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_W, CL_TRUE, 0, st->C*st->K*st->M*st->window_size*sizeof(cl_float), st->W, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_W failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_prop, CL_TRUE, 0, st->M*sizeof(cl_float), st->prop, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_prop failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_power, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->power, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_power failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_power_1, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->power_1, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_power_1 failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_E, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->E, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_E failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_PHI, CL_TRUE, 0, (st->window_size)*sizeof(cl_float), st->PHI, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_PHI failed", err);

		//err = clEnqueueWriteBuffer(*commandQueue, ocl_wtmp, CL_TRUE, 0, (st->window_size*st->M)*sizeof(cl_float), st->wtmp, 0, NULL, NULL);
		//print_on_opencl_error("clEnqueueWriteBuffer of ocl_wtmp failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_y, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->y, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_y failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_Y, CL_FALSE, 0, st->C*st->window_size*sizeof(cl_float), st->Y, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_Y failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_last_y, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->last_y, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_last_y failed", err);
		
		err = clEnqueueWriteBuffer(*commandQueue, ocl_window, CL_TRUE, 0, (st->window_size)*sizeof(cl_float), st->window, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_window failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, ocl_saturated, CL_TRUE, 0, sizeof(cl_int), &st->saturated, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of ocl_saturated failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, data_time_domain, CL_TRUE, 0, (st->K*st->window_size)*sizeof(cl_float), st->x, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of data_time_domain failed", err);

		err = clEnqueueWriteBuffer(*commandQueue, data_freq_domain, CL_TRUE, 0, st->K*(st->M+1)*st->window_size*sizeof(cl_float), st->X, 0, NULL, NULL);
		print_on_opencl_error("clEnqueueWriteBuffer of data_time_domain failed", err);
#ifdef SKIP_OCL_STATE_SYNC
	}
	/*else
	{
		printf("Skipping full sync of CPU/GPU state. Cancel count is %d\n", st->cancel_count);
	}*/
#endif
}

/**
 * Copies data from the host-memory st-struct to GPU memory to sync the state
 * This version is used on device switch by the load balancer
 */
EXPORT void speex_echo_opencl_sync_to_GPU(SpeexEchoState *st)
{
	cl_int err = CL_SUCCESS;

	err = clEnqueueWriteBuffer(*commandQueue, ocl_notch_mem, CL_TRUE, 0, 2*st->C*sizeof(cl_float), st->notch_mem, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_notch_mem failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_input, CL_TRUE, 0, st->C*st->frame_size*sizeof(cl_float), st->input, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_input failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_memD, CL_TRUE, 0, st->C*sizeof(cl_float), st->memD, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_memD failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_memE, CL_TRUE, 0, st->C*sizeof(cl_float), st->memE, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_memE failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_memX, CL_TRUE, 0, st->K*sizeof(cl_float), st->memX, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_memX failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_Xf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Xf, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Xf failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_Rf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Rf, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Rf failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_Yf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Yf, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Yf failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_Yh, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Yh, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Yh failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_Eh, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Eh, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Eh failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_foreground, CL_TRUE, 0, st->M*st->window_size*st->C*st->K*sizeof(cl_float), st->foreground, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_foreground failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_e, CL_TRUE, 0, st->C*st->window_size*sizeof(cl_float), st->e, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_e failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_W, CL_TRUE, 0, st->C*st->K*st->M*st->window_size*sizeof(cl_float), st->W, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_W failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_prop, CL_TRUE, 0, st->M*sizeof(cl_float), st->prop, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_prop failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_power, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->power, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_power failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_power_1, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->power_1, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_power_1 failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_E, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->E, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_E failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_PHI, CL_TRUE, 0, (st->window_size)*sizeof(cl_float), st->PHI, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_PHI failed", err);

	//err = clEnqueueWriteBuffer(*commandQueue, ocl_wtmp, CL_TRUE, 0, (st->window_size*st->M)*sizeof(cl_float), st->wtmp, 0, NULL, NULL);
	//print_on_opencl_error("clEnqueueWriteBuffer of ocl_wtmp failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_y, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->y, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_y failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_Y, CL_FALSE, 0, st->C*st->window_size*sizeof(cl_float), st->Y, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Y failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_last_y, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->last_y, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_last_y failed", err);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_window, CL_TRUE, 0, (st->window_size)*sizeof(cl_float), st->window, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_window failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, ocl_saturated, CL_TRUE, 0, sizeof(cl_int), &st->saturated, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_saturated failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, data_time_domain, CL_TRUE, 0, (st->K*st->window_size)*sizeof(cl_float), st->x, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of data_time_domain failed", err);

	err = clEnqueueWriteBuffer(*commandQueue, data_freq_domain, CL_TRUE, 0, st->K*(st->M+1)*st->window_size*sizeof(cl_float), st->X, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of data_time_domain failed", err);
}

/**
 * Reverse of the above, used for stop execution on GPU, and continue on CPU (non-opencl)
 */
EXPORT void speex_echo_opencl_sync_to_CPU(SpeexEchoState *st)
{
	cl_int err = CL_SUCCESS;
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_notch_mem, CL_TRUE, 0, 2*st->C*sizeof(cl_float), st->notch_mem, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_notch_mem failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_input, CL_TRUE, 0, st->C*st->frame_size*sizeof(cl_float), st->input, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_input failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_memD, CL_TRUE, 0, st->C*sizeof(cl_float), st->memD, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_memD failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_memE, CL_TRUE, 0, st->C*sizeof(cl_float), st->memE, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_memE failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_memX, CL_TRUE, 0, st->K*sizeof(cl_float), st->memX, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_memX failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_Xf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Xf, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Xf failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_Rf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Rf, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Rf failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_Yf, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Yf, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Yf failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_Yh, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Yh, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Yh failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_Eh, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->Eh, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Eh failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_foreground, CL_TRUE, 0, st->M*st->window_size*st->C*st->K*sizeof(cl_float), st->foreground, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_foreground failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_e, CL_TRUE, 0, st->C*st->window_size*sizeof(cl_float), st->e, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_e failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_W, CL_TRUE, 0, st->C*st->K*st->M*st->window_size*sizeof(cl_float), st->W, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_W failed", err);

	//
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_prop, CL_TRUE, 0, st->M*sizeof(cl_float), st->prop, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_prop failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_power, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->power, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_power failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_power_1, CL_TRUE, 0, (st->frame_size+1)*sizeof(cl_float), st->power_1, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_power_1 failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_E, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->E, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_E failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_PHI, CL_TRUE, 0, (st->window_size)*sizeof(cl_float), st->PHI, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_PHI failed", err);

	//err = clEnqueueReadBuffer(*commandQueue, ocl_wtmp, CL_TRUE, 0, (st->window_size*st->M)*sizeof(cl_float), st->wtmp, 0, NULL, NULL);
	//print_on_opencl_error("clEnqueueWriteBuffer of ocl_wtmp failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_y, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->y, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_y failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_Y, CL_FALSE, 0, st->C*st->window_size*sizeof(cl_float), st->Y, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_Y failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_last_y, CL_TRUE, 0, (st->C*st->window_size)*sizeof(cl_float), st->last_y, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_last_y failed", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_window, CL_TRUE, 0, (st->window_size)*sizeof(cl_float), st->window, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_window failed", err);

	err = clEnqueueReadBuffer(*commandQueue, ocl_saturated, CL_TRUE, 0, sizeof(cl_int), &st->saturated, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_saturated failed", err);

	err = clEnqueueReadBuffer(*commandQueue, data_time_domain, CL_TRUE, 0, (st->K*st->window_size)*sizeof(cl_float), st->x, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of data_time_domain failed", err);

	err = clEnqueueReadBuffer(*commandQueue, data_freq_domain, CL_TRUE, 0, st->K*(st->M+1)*st->window_size*sizeof(cl_float), st->X, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of data_time_domain failed", err);
}

/**
 * Release OpenCL memory
 */
void speex_echo_opencl_cleanup_common_mem(cl_command_queue *commandQueue)
{
	cl_int status;

	printf("Releasing OpenCL buffers.\n");

	if(error_signal_block_sums)
		free(error_signal_block_sums);
	if(error_signal_temp)
		free(error_signal_temp);
	if(error_signal_out_int)
		free(error_signal_out_int);
	if(host_temp_memD)
		free(host_temp_memD);
	
	if(Pey_Pyy_reduction_host)
		free(Pey_Pyy_reduction_host);
        if(data_time_domain)
                clReleaseMemObject(data_time_domain);
	if(data_time_domain_host)
		free(data_time_domain_host);
        if(data_time_domain_temp)
                clReleaseMemObject(data_time_domain_temp);
        if(data_freq_domain)
                clReleaseMemObject(data_freq_domain);
	if(data_freq_domain_host)
		free(data_freq_domain_host);
        if(data_freq_domain_temp)
                clReleaseMemObject(data_freq_domain_temp);
	if(ocl_in)
		clReleaseMemObject(ocl_in);
	if(ocl_far_end)
		clReleaseMemObject(ocl_far_end);
	if(ocl_input)
		clReleaseMemObject(ocl_input);
	if(ocl_out)
		clReleaseMemObject(ocl_out);
	if(ocl_notch_mem)
		clReleaseMemObject(ocl_notch_mem);
	if(ocl_memD)
		clReleaseMemObject(ocl_memD);
	if(ocl_memE)
		clReleaseMemObject(ocl_memE);
	if(ocl_memX)
		clReleaseMemObject(ocl_memX);
	if(ocl_Xf)
		clReleaseMemObject(ocl_Xf);
	if(ocl_Rf)
		clReleaseMemObject(ocl_Rf);
	if(ocl_Yf)
		clReleaseMemObject(ocl_Yf);
	if(ocl_Yh)
		clReleaseMemObject(ocl_Yh);
	if(ocl_Y)
		clReleaseMemObject(ocl_Y);
	if(ocl_power_1)
		clReleaseMemObject(ocl_power_1);
	if(ocl_e)
		clReleaseMemObject(ocl_e);
	if(ocl_temp_e)
		clReleaseMemObject(ocl_temp_e);
	if(ocl_E)
		clReleaseMemObject(ocl_E);
	if(ocl_PHI)
		clReleaseMemObject(ocl_PHI);
	if(ocl_wtmp)
		clReleaseMemObject(ocl_wtmp);
	if(ocl_y)
		clReleaseMemObject(ocl_y);
	if(ocl_last_y)
		clReleaseMemObject(ocl_last_y);
	if(ocl_window)
		clReleaseMemObject(ocl_window);
	if(ocl_saturated)
		clReleaseMemObject(ocl_saturated);
	if(ocl_Pey)
		clReleaseMemObject(ocl_Pey);
	if(ocl_Pyy)
		clReleaseMemObject(ocl_Pyy);
	if(ocl_temp_reduction_buffer)
		clReleaseMemObject(ocl_temp_reduction_buffer);
	if(ocl_temp_acc)
		clReleaseMemObject(ocl_temp_acc);

	printf("Released OpenCL buffers.\n");
}

/*
 * \brief Release OpenCL resources (Context, Program, Kernels etc.)
 */
void speex_echo_opencl_cleanup_common_dev(cl_command_queue *commandQueue)
{
	cl_int status;
	
	if(MDF_inner_product_reduce_kernel)
                clReleaseKernel(MDF_inner_product_reduce_kernel);
	if(MDF_spectral_mul_accum_kernel)
                clReleaseKernel(MDF_spectral_mul_accum_kernel);
	if(MDF_compute_foreground_filter_kernel)
                clReleaseKernel(MDF_compute_foreground_filter_kernel);
	if(MDF_weighted_spectral_mul_conj_kernel)
                clReleaseKernel(MDF_weighted_spectral_mul_conj_kernel);
	if(MDF_set_array_to_float_value_kernel)
                clReleaseKernel(MDF_set_array_to_float_value_kernel);
	if(MDF_copy_float_array_kernel)
                clReleaseKernel(MDF_copy_float_array_kernel);
	if(MDF_update_foreground_filter_kernel)
		clReleaseKernel(MDF_update_foreground_filter_kernel);
	if(MDF_update_background_filter_kernel)
		clReleaseKernel(MDF_update_background_filter_kernel);
	if(MDF_smooth_far_end_compute_filtered_spectra_kernel)
		clReleaseKernel(MDF_smooth_far_end_compute_filtered_spectra_kernel);
	if(MDF_learning_rate_calc_kernel)
		clReleaseKernel(MDF_learning_rate_calc_kernel);
	if(MDF_adapted_copy_kernel)
		clReleaseKernel(MDF_adapted_copy_kernel);
	if(MDF_prepare_shift_preemph_freq_data_kernel)
		clReleaseKernel(MDF_prepare_shift_preemph_freq_data_kernel);
	if(MDF_power_spectrum_inner_product_kernel)
		clReleaseKernel(MDF_power_spectrum_inner_product_kernel);
	if(MDF_response_diff_inner_product_kernel)
		clReleaseKernel(MDF_response_diff_inner_product_kernel);
	if(MDF_error_signal_combined_kernel)
		clReleaseKernel(MDF_error_signal_combined_kernel);
	if(MDF_inner_prod_power_spec_spectral_mul_kernel)
		clReleaseKernel(MDF_inner_prod_power_spec_spectral_mul_kernel);

	printf("Released OpenCL kernels.\n");
	
	if(clMDF_program)
                clReleaseProgram(clMDF_program);

	if(*commandQueue != NULL)
        {
                status = clReleaseCommandQueue(*commandQueue);
                if(status != CL_SUCCESS)
                {
                        printf("Error: In clReleaseCommandQueue\n");
                        return;
                }
        }
        if(context != NULL)
	{
		status = clReleaseContext(context);
		if(status != CL_SUCCESS)
		{
			printf("Error: In clReleaseContext\n");
			return;
		}
	}
        if(device != NULL)
        {
                //free(devices);
                device = NULL;
        }
        
        printf("Done cleaning up after OpenCL.\n");
}

/*
 * Calculates error by finding the sum of the squared difference between residuals of two states
 */
EXPORT double get_residual_error(SpeexEchoState *st1, SpeexEchoState *st2)
{
	double sqr_sum = 0.0;
	int i;
	spx_word32_t *residual1 = (spx_word32_t *)malloc(sizeof(spx_word32_t)*st1->frame_size+1);
	spx_word32_t *residual2 = (spx_word32_t *)malloc(sizeof(spx_word32_t)*st1->frame_size+1);
	
	speex_echo_get_residual(st1, residual1, st1->frame_size);
	speex_echo_get_residual(st2, residual2, st1->frame_size);
	
	//printf("Diff for residuals: \n");
	for(i=0;i<=st1->frame_size;i++)
	{
		sqr_sum += (residual2[i] - residual1[i]) * (residual2[i] - residual1[i]);
		//printf("%d: %f\t", i, (residual2[i] - residual1[i]) * (residual2[i] - residual1[i]));
		//if(i % 5 == 0)
		//	printf("\n");
	}
	//printf("\nSum residual error: %f\n", sqr_sum);
	
	return sqr_sum;
}

#include "opencl_utility.c"
#include "mdf_oclfft.c"

unsigned int nextPow2( unsigned int x ) {
    --x;
    x |= x >> 1;
    x |= x >> 2;
    x |= x >> 4;
    x |= x >> 8;
    x |= x >> 16;
    return ++x;
}

/**
 * Precondition: len >= 128
 */
float mdf_opencl_inner_prod(cl_mem x, unsigned int x_offset, cl_mem y, unsigned int y_offset, int len, int batch_size, int batch_pitch)
{
	size_t szLocalWorkSize;
        size_t szGlobalWorkSize;
	cl_int err = CL_SUCCESS;
	int n = len >> 1;
	int i;
	float final_sum = 0.0;
	//int cl_threads = (n < 256) ? nextPow2(n) : 128;
	//int cl_blocks = (n + cl_threads - 1) / (cl_threads);
	int cl_threads = n;
	//int cl_blocks = len / cl_threads;
	int cl_blocks = 1;
	szLocalWorkSize = cl_threads;
	szGlobalWorkSize = cl_threads * cl_blocks;

	//printf("Running MDF inner product with %d threads in %d blocks.\n", cl_threads, cl_blocks);

	float *block_sums = (float *)malloc(sizeof(float) * cl_blocks);

	clSetKernelArg(MDF_inner_product_reduce_kernel, 0, sizeof(cl_mem), &x);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 1, sizeof(cl_uint), &x_offset);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 2, sizeof(cl_mem), &y);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 3, sizeof(cl_uint), &y_offset);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 4, sizeof(cl_mem), &ocl_temp_buffer);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 5, sizeof(cl_int), &n);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 6, sizeof(cl_int), &batch_size);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 7, sizeof(cl_int), &batch_pitch);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_inner_product_reduce_kernel.", err);
	clSetKernelArg(MDF_inner_product_reduce_kernel, 8, sizeof(float) * cl_threads, NULL); //Shared mem
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_inner_product_reduce_kernel.", err);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_inner_product_reduce_kernel, 1, NULL, &szGlobalWorkSize, &szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_inner_product_reduce_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_inner_product_reduce_kernel", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_inner_product_reduce_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_inner_product_reduce_kernel", err);
	// Do a reduction of the sum for each block on the CPU
	err = clEnqueueReadBuffer(*commandQueue, ocl_temp_buffer, CL_TRUE, 0, cl_blocks*sizeof(cl_float), block_sums, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy to temporary frequency block sum array in the opencl inner product function.", err);
	for(i=0;i<cl_blocks;i++)
	{
		final_sum += block_sums[i];
	}

	free(block_sums);
	return final_sum;
}

float mdf_opencl_power_spectrum_inner_prod(cl_mem x, cl_mem E, cl_mem Y, cl_mem X, cl_mem Rf, cl_mem Yf, cl_mem Xf, unsigned int E_offset, unsigned int Y_offset, unsigned int X_offset, unsigned int C, unsigned int K, unsigned int window_size, unsigned int x1_offset, unsigned int x2_offset, unsigned int N_power_spectrum, int n_inner_product)
{
	size_t szLocalWorkSize[1];
        size_t szGlobalWorkSize[1];
	cl_int err = CL_SUCCESS;
	int n = n_inner_product >> 1;
	int i, speak;
	float final_sum = 0.0;
	//int inner_prod_threads = (n < 256) ? nextPow2(n) : 128;
	int inner_prod_threads = 64;
	int inner_prod_blocks = n_inner_product / inner_prod_threads;
	//int inner_prod_blocks = (n + inner_prod_threads - 1) / (inner_prod_threads);
	szLocalWorkSize[0] = inner_prod_threads;
	//szGlobalWorkSize[0] = n_inner_product / 2;	// Size for power spectrum
	szGlobalWorkSize[0] = n_inner_product;	// Size for power spectrum

	//printf("Running MDF inner product with %d threads in %d blocks.\n", cl_threads, cl_blocks);

	float *block_sums = (float *)malloc(sizeof(float) * inner_prod_blocks * K);

	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 0, sizeof(cl_mem), &x);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 1, sizeof(cl_mem), &E);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 2, sizeof(cl_mem), &Y);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 3, sizeof(cl_mem), &X);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 4, sizeof(cl_mem), &Rf);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 5, sizeof(cl_mem), &Yf);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 6, sizeof(cl_mem), &Xf);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 7, sizeof(cl_uint), &E_offset);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 8, sizeof(cl_uint), &Y_offset);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 9, sizeof(cl_uint), &X_offset);
	print_on_opencl_error("Failed to set kernel argument 9 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 10, sizeof(cl_uint), &x1_offset);
	print_on_opencl_error("Failed to set kernel argument 10 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 11, sizeof(cl_uint), &x2_offset);
	print_on_opencl_error("Failed to set kernel argument 11 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 12, sizeof(cl_uint), &C);
	print_on_opencl_error("Failed to set kernel argument 12 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 13, sizeof(cl_uint), &K);
	print_on_opencl_error("Failed to set kernel argument 13 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 14, sizeof(cl_uint), &window_size);
	print_on_opencl_error("Failed to set kernel argument 14 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 15, sizeof(cl_uint), &N_power_spectrum);
	print_on_opencl_error("Failed to set kernel argument 15 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 16, sizeof(cl_uint), &n_inner_product);
	print_on_opencl_error("Failed to set kernel argument 16 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 17, sizeof(cl_mem), &ocl_temp_buffer);
	print_on_opencl_error("Failed to set kernel argument 17 of MDF_power_spectrum_inner_product_kernel.", err);
	clSetKernelArg(MDF_power_spectrum_inner_product_kernel, 18, sizeof(float) * inner_prod_threads, NULL); //Local mem
	print_on_opencl_error("Failed to set kernel argument 18 of MDF_power_spectrum_inner_product_kernel.", err);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_power_spectrum_inner_product_kernel, 1, NULL, &szGlobalWorkSize, &szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_power_spectrum_inner_product_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_power_spectrum_inner_product_kernel", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_inner_product_reduce_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_power_spectrum_inner_product_kernel", err);

	// Do a reduction of the sum for each block on the CPU
	err = clEnqueueReadBuffer(*commandQueue, ocl_temp_buffer, CL_TRUE, 0, inner_prod_blocks*K*sizeof(float), block_sums, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy to temporary frequency block sum array in the opencl inner product function.", err);

	for(speak=0;speak<K;speak++)
	{
		for(i=0;i<inner_prod_blocks;i++)
		{
			//printf("Adding %f to Sxx\n", block_sums[inner_prod_blocks*speak + i]);
			final_sum += block_sums[inner_prod_blocks*speak + i];
		}
	}

	free(block_sums);

	return final_sum;
}

// Now special case looping through channels as well
void mdf_opencl_spectral_mul_accum(cl_mem X, cl_mem Y, cl_mem acc, cl_mem Rf, cl_mem Yf, cl_mem Xf, unsigned int N, unsigned int M, unsigned int K, unsigned int C)
{
	size_t szLocalWorkSize[2];
        size_t szGlobalWorkSize[2];
	cl_int err = CL_SUCCESS;
	szLocalWorkSize[0] = 32;
	szLocalWorkSize[1] = 1;
	szGlobalWorkSize[0] = N/2;
	szGlobalWorkSize[1] = C;

	clSetKernelArg(MDF_spectral_mul_accum_kernel, 0, sizeof(cl_mem), &X);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 1, sizeof(cl_mem), &Y);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 2, sizeof(cl_mem), &acc);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 3, sizeof(cl_mem), &Rf);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 4, sizeof(cl_mem), &Yf);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 5, sizeof(cl_mem), &Xf);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 6, sizeof(cl_uint), &N);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 7, sizeof(cl_uint), &K);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_spectral_mul_accum_kernel.", err);
	clSetKernelArg(MDF_spectral_mul_accum_kernel, 8, sizeof(cl_uint), &M);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_spectral_mul_accum_kernel.", err);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_spectral_mul_accum_kernel, 2, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_spectral_mul_accum_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_spectral_mul_accum_kernel", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_spectral_mul_accum_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_spectral_mul_accum_kernel", err);
}

/**
 * Adjust proportional adaption rate
 * Some calculations done on GPU, but some reductions and stuff are done on the CPU
 * Could be fully implemented on GPU with some added complexity
 */
void mdf_opencl_adjust_prop(cl_mem W, int N, int M, int P, cl_mem ocl_prop)
{
	int i, j, p;
	spx_word16_t max_sum = 1;
	spx_word32_t prop_sum = 1;
	float *prop = (float *)malloc(sizeof(float) * M);
	
	cl_int err = CL_SUCCESS;
	size_t szLocalWorkSize[3];
        size_t szGlobalWorkSize[3];
	int cl_threads = 128;
	//int cl_blocks = (P*N + cl_threads - 1) / (cl_threads);
	int cl_blocks = N / (cl_threads);
	int n = P*N;
	//float *block_sums = (float *)malloc(sizeof(float) * cl_blocks * M * P);
	float *block_sums = (float *)malloc(sizeof(float) * cl_blocks * M);
	float temp_sum;

	szLocalWorkSize[0] = cl_threads;
	szLocalWorkSize[1] = 1;
	szLocalWorkSize[2] = 1;
	szGlobalWorkSize[0] = N;
	szGlobalWorkSize[1] = 1;
	szGlobalWorkSize[2] = M;
	temp_sum = 1.0;

	// GPU
	clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 0, sizeof(cl_mem), &W);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_adjust_prop_phase1_reduce_kernel.", err);
	clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 1, sizeof(cl_mem), &ocl_temp_buffer);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_adjust_prop_phase1_reduce_kernel.", err);
	//clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 2, sizeof(cl_uint), &i);
	//print_on_opencl_error("Failed to set kernel argument 2 of MDF_adjust_prop_phase1_reduce_kernel.", err);
	clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 2, sizeof(cl_uint), &M);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_adjust_prop_phase1_reduce_kernel.", err);
	clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 3, sizeof(cl_uint), &N);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_adjust_prop_phase1_reduce_kernel.", err);
	clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 4, sizeof(cl_uint), &n);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_adjust_prop_phase1_reduce_kernel.", err);
	clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 5, sizeof(cl_uint), &P);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_adjust_prop_phase1_reduce_kernel.", err);
	clSetKernelArg(MDF_adjust_prop_phase1_reduce_kernel, 6, sizeof(float) * szLocalWorkSize[0], NULL);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_adjust_prop_phase1_reduce_kernel.", err);

	err = clEnqueueNDRangeKernel(*commandQueue, MDF_adjust_prop_phase1_reduce_kernel, 3, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel MDF_adjust_prop_phase1_reduce_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_adjust_prop_phase1_reduce_kernel", 3);
		//printf("SUCCESS: clEnqueueNDRangeKernel of MDF_adjust_prop_phase1_reduce_kernel (i is %d, M is %d)\n", i, M);

	//err=clFinish(*commandQueue);
	//print_on_opencl_error("clFinish failed for MDF_adjust_prop_phase1_reduce_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_adjust_prop_phase1_reduce_kernel", err);

	// Do a reduction of the sum for each block on the CPU
	err = clEnqueueReadBuffer(*commandQueue, ocl_temp_buffer, CL_TRUE, 0, cl_blocks*M*sizeof(float), block_sums, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy to temporary frequency block sum array in the opencl inner product function.", err);
	
	//printf("Priting temporary block data, stats: M (%d), P(%d), cl_blocks(%d)\n", M, P, cl_blocks);
	//for(i=0;i<M*cl_blocks;i++)
	//{
	//	printf("block %d: %f\t", i, block_sums[i]);
	//}
	//printf("\n");

	for (i=0;i<M;i++)
	{
		temp_sum = 1.0;
		
		for(j=0;j<cl_blocks;j++)
		{
			temp_sum += block_sums[i*cl_blocks + j];
			//printf("Adding %f (block_sums[%d + %d] = %d) to temp_sum\n", block_sums[i*cl_blocks + j], i*cl_blocks, j, i*cl_blocks + j);
			//temp_sum += 9000.0f;
		}
		//printf("prop[%d] was %f, temp_sum is %f, root of temp_sum is %f\n", i, prop[i], temp_sum, sqrtf(temp_sum));
		prop[i] = spx_sqrt(temp_sum);
		//printf("Setting prop[%d] to %f\n", i, prop[i]);
		if (prop[i] > max_sum)
			max_sum = prop[i];
	}
	
	free(block_sums);
	
	for (i=0;i<M;i++)
	{
		prop[i] += 0.1f * max_sum;
		prop_sum += prop[i];
	}
	for (i=0;i<M;i++)
	{
		prop[i] = 0.99f * prop[i] / prop_sum;
	}

	// Copy the final result of prop back to the GPU
	err = clEnqueueWriteBuffer(*commandQueue, ocl_prop, CL_TRUE, 0, M*sizeof(float), prop, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer failed on copy to GPU result in mdf_opencl_adjust_prop.", err);

}

/**
 * TODO: Parallelize more
 */
void mdf_opencl_weighted_spectral_mul_conj(unsigned int C, unsigned int K, unsigned int M, cl_mem power1, cl_mem prop, cl_mem X, cl_mem E, cl_mem PHI, unsigned int N, cl_mem W)
{
	unsigned int chan, speak, j;
	unsigned int prop_offset, E_offset;
	
	int u;

	u = M-1;

	size_t szLocalWorkSize[3];
	size_t szGlobalWorkSize[3];
	cl_int err = CL_SUCCESS;
	szLocalWorkSize[0] = 128;
	szLocalWorkSize[1] = 1;
	szLocalWorkSize[2] = 1;
	szGlobalWorkSize[0] = N/2;
	szGlobalWorkSize[1] = M;
	szGlobalWorkSize[2] = C*K;
	prop_offset = u;
	E_offset = chan * N;

	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 0, sizeof(cl_mem), &power1);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 1, sizeof(cl_mem), &prop);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 2, sizeof(cl_mem), &X);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 3, sizeof(cl_mem), &E);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 4, sizeof(cl_uint), &E_offset);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 5, sizeof(cl_mem), &PHI);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 6, sizeof(cl_mem), &W);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 7, sizeof(cl_uint), &N);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 8, sizeof(cl_uint), &C);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_weighted_spectral_mul_conj_kernel.", err);
	clSetKernelArg(MDF_weighted_spectral_mul_conj_kernel, 9, sizeof(cl_uint), &K);
	print_on_opencl_error("Failed to set kernel argument 9 of MDF_weighted_spectral_mul_conj_kernel.", err);

	err = clEnqueueNDRangeKernel(*commandQueue, MDF_weighted_spectral_mul_conj_kernel, 3, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel MDF_weighted_spectral_mul_conj_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_weighted_spectral_mul_conj_kernel", 3);

	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_weighted_spectral_mul_conj_kernel", err);
}

void mdf_opencl_copy_float_array(cl_mem input, int input_offset, cl_mem output, int output_offset, unsigned int data_size)
{
	cl_int err = CL_SUCCESS;
	size_t szGlobalWorkSize[1];
	/* Compute error signal (filter update version) */ 
	szGlobalWorkSize[0] = data_size;

	err = clSetKernelArg(MDF_copy_float_array_kernel, 0, sizeof(cl_mem), (void*)&input);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_copy_float_array_kernel.", err);
	err = clSetKernelArg(MDF_copy_float_array_kernel, 1, sizeof(cl_int), (void*)&input_offset);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_copy_float_array_kernel.", err);
	err = clSetKernelArg(MDF_copy_float_array_kernel, 2, sizeof(cl_mem), (void*)&output);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_copy_float_array_kernel.", err);
	err = clSetKernelArg(MDF_copy_float_array_kernel, 3, sizeof(cl_int), (void*)&output_offset);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_copy_float_array_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clSetKernelArg\n", 3);

	err = clEnqueueNDRangeKernel(*commandQueue, MDF_copy_float_array_kernel, 1, NULL, &szGlobalWorkSize, NULL, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel MDF_copy_float_array_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_copy_float_array_kernel", 3);

	//err=clFinish(*commandQueue);
	//print_on_opencl_error("clFinish failed for MDF_copy_float_array_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_copy_float_array_kernel", err);
}

void mdf_opencl_set_array_to_float_value(cl_mem data, unsigned int data_offset, float value, unsigned int data_size)
{
	//size_t szLocalWorkSize[2];
        size_t szGlobalWorkSize[2];
	cl_int err = CL_SUCCESS;
	//szLocalWorkSize[0] = 32;
	//szLocalWorkSize[1] = 1;
	szGlobalWorkSize[0] = data_size;
	szGlobalWorkSize[1] = 1;

	clSetKernelArg(MDF_set_array_to_float_value_kernel, 0, sizeof(cl_mem), &data);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_set_array_to_float_value_kernel.", err);
	clSetKernelArg(MDF_set_array_to_float_value_kernel, 1, sizeof(cl_uint), &data_offset);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_set_array_to_float_value_kernel.", err);
	clSetKernelArg(MDF_set_array_to_float_value_kernel, 2, sizeof(cl_float), &value);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_set_array_to_float_value_kernel.", err);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_set_array_to_float_value_kernel, 1, NULL, &szGlobalWorkSize, NULL, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_set_array_to_float_value_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_set_array_to_float_value_kernel.", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_set_array_to_float_value_kernel.", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_set_array_to_float_value_kernel", err);
}

void mdf_opencl_response_diff_inner_product(cl_mem e, cl_mem temp_e, cl_mem y, cl_mem input, unsigned int C, unsigned int window_size, unsigned int frame_size, float *Dbf, float *See)
{
	int i;
        size_t szGlobalWorkSize[2];
	size_t szLocalWorkSize[2];
	cl_int err = CL_SUCCESS;
	unsigned int cl_threads = 128;
	unsigned int inner_product_blocks = frame_size / cl_threads; 
	szGlobalWorkSize[0] = frame_size;
	szGlobalWorkSize[1] = 2;
	szLocalWorkSize[0] = cl_threads;
	szLocalWorkSize[1] = 1;

	float *block_sums = (float *)malloc(sizeof(float) * inner_product_blocks * 2);

	clSetKernelArg(MDF_response_diff_inner_product_kernel, 0, sizeof(cl_mem), &e);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 1, sizeof(cl_mem), &temp_e);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 2, sizeof(cl_mem), &y);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 3, sizeof(cl_mem), &input);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 4, sizeof(cl_uint), &C);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 5, sizeof(cl_uint), &window_size);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 6, sizeof(cl_uint), &frame_size);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 7, sizeof(cl_mem), &ocl_temp_reduction_buffer);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 8, sizeof(cl_mem), &ocl_temp_buffer);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_response_diff_inner_product_kernel.", err);
	clSetKernelArg(MDF_response_diff_inner_product_kernel, 9, sizeof(float) * cl_threads, NULL);
	print_on_opencl_error("Failed to set kernel argument 9 of MDF_response_diff_inner_product_kernel.", err);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_response_diff_inner_product_kernel, 2, NULL, &szGlobalWorkSize, &szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_response_diff_inner_product_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_response_diff_inner_product_kernel.", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_response_diff_kernel.", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_response_diff_inner_product_kernel", err);

	// Do a reduction of the sum for each block on the CPU
	err = clEnqueueReadBuffer(*commandQueue, ocl_temp_reduction_buffer, CL_TRUE, 0, inner_product_blocks*sizeof(cl_float)*2, block_sums, 0, NULL, NULL);

	*Dbf = 0.0f;
	*See = 0.0f;
	for(i=0;i<inner_product_blocks;i++)
        {
                *Dbf += block_sums[i];
		//printf("Added %f to Dbf\n", block_sums[i]);
                *See += block_sums[inner_product_blocks + i];
		//printf("Added %f to See\n", block_sums[inner_product_blocks + i]);
        }

	*Dbf += 10*C;
	
	//mdf_opencl_copy_float_array(ocl_temp_reduction_buffer, 0, ocl_e, 0, C*frame_size);

	free(block_sums);
}

void mdf_opencl_compute_error_signal_combined(cl_mem input, cl_mem e, 
						cl_mem y, cl_mem in, cl_mem out, cl_mem saturated, 
				     		float preemph, 
						unsigned int window_size, 
						unsigned int frame_size, 
						unsigned int num_channels, 
				     		float *Sey, float *Syy, float *Sdd, float *h_memE, short *h_out)
{
	size_t szLocalWorkSize[3];
        size_t szGlobalWorkSize[3];
	cl_int err = CL_SUCCESS;
	int n = frame_size >> 1;
	int i, j, chan;
	//int cl_threads = (n < 256) ? nextPow2(n) : 128;
	//int cl_threads = 64;
	int cl_threads = 128;	// IMPORTANT; This is tied to the size of the reduction buffer (global var.)
	//int cl_blocks = (n + cl_threads - 1) / (cl_threads);
	int cl_blocks = frame_size / cl_threads;
	szLocalWorkSize[0] = cl_threads;
	szLocalWorkSize[1] = 1;
	szLocalWorkSize[2] = 1;
	szGlobalWorkSize[0] = cl_threads * cl_blocks;
	szGlobalWorkSize[1] = num_channels;
	szGlobalWorkSize[2] = 4; // three correlations + a loop for output

	//float *block_sums = (float *)malloc(sizeof(float) * cl_blocks * 3 * num_channels);
	// Temporary sum calculated on GPU in order to avoid multiple memory transfers
	//float *temp_error_signal = (float *)malloc(sizeof(float) * frame_size * num_channels);
	//int *h_out_int = (int *)malloc(sizeof(int) * frame_size * num_channels);

	clSetKernelArg(MDF_error_signal_combined_kernel, 0, sizeof(cl_mem), &input);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 1, sizeof(cl_mem), &e);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_error_signal_combined_kernel.", err);
	//clSetKernelArg(MDF_error_signal_combined_kernel, 2, sizeof(cl_mem), &memE);
	//print_on_opencl_error("Failed to set kernel argument 2 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 2, sizeof(cl_mem), &in);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_error_signal_combined_kernel.", err);
	//clSetKernelArg(MDF_error_signal_combined_kernel, 4, sizeof(cl_mem), &out);
	//print_on_opencl_error("Failed to set kernel argument 4 of MDF_error_signal_combined_kernel.", err);
	//clSetKernelArg(MDF_error_signal_combined_kernel, 5, sizeof(cl_float), &preemph);
	//print_on_opencl_error("Failed to set kernel argument 5 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 3, sizeof(cl_mem), &ocl_temp_buffer);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 4, sizeof(cl_mem), &ocl_temp_reduction_buffer);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 5, sizeof(cl_uint), &window_size);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 6, sizeof(cl_uint), &frame_size);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 7, sizeof(cl_uint), &num_channels);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 8, sizeof(cl_mem), &saturated);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 9, sizeof(cl_mem), &y);
	print_on_opencl_error("Failed to set kernel argument 9 of MDF_error_signal_combined_kernel.", err);
	clSetKernelArg(MDF_error_signal_combined_kernel, 10, sizeof(float) * cl_threads, NULL); //Shared mem
	print_on_opencl_error("Failed to set kernel argument 10 of MDF_error_signal_combined_kernel.", err);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_error_signal_combined_kernel, 3, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_error_signal_combined_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_error_signal_combined_kernel", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_inner_product_reduce_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_error_singal_combined_kernel", err);

	// Do a reduction of the sum for each block on the CPU
	err = clEnqueueReadBuffer(*commandQueue, ocl_temp_reduction_buffer, CL_TRUE, 0, cl_blocks*sizeof(float)*3*num_channels, error_signal_block_sums, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy to temporary frequency block sum array in the opencl inner product function.", err);
	err = clEnqueueReadBuffer(*commandQueue, ocl_temp_buffer, CL_TRUE, 0, frame_size*num_channels*sizeof(float), error_signal_temp, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy to temporary error signal array in the opencl inner product function.", err);
	
	// Do error signal calculations on CPU, since they are poorly parallelizable (see memE)
	for (chan = 0; chan < num_channels; chan++)
	{    
		/* Compute error signal (for the output with de-emphasis) */ 
		for (i=0;i<frame_size;i++)
		{
			float tmp_out;
			//printf("temp_error_signal[%d]: %f\n", i*num_channels + chan, temp_error_signal[i*num_channels + chan]);
			tmp_out = error_signal_temp[i*num_channels + chan] + (preemph * h_memE[chan]);
			h_out[i*num_channels+chan] = WORD2INT(tmp_out);
			error_signal_out_int[i*num_channels+chan] = h_out[i*num_channels+chan];
			//printf("h_out[%d]: %hd\n", i*num_channels + chan, h_out[i*num_channels + chan]);
			h_memE[chan] = tmp_out;
		}
	}
	
	err = clEnqueueWriteBuffer(*commandQueue, out, CL_TRUE, 0, frame_size*num_channels*sizeof(int), error_signal_out_int, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy back to output on device.", err);

	for(i=0;i<num_channels;i++)
	{
		for(j=0;j<cl_blocks;j++)
		{
			*Sey += error_signal_block_sums[j*num_channels + i];
			//printf("Adding %f to Sey\n", error_signal_block_sums[j*num_channels + i]);
			*Syy += error_signal_block_sums[cl_blocks*num_channels + j*num_channels + i];
			//printf("Adding %f to Syy\n", error_signal_block_sums[cl_blocks*num_channels + j*num_channels + i]);
			*Sdd += error_signal_block_sums[2*cl_blocks*num_channels + j*num_channels + i];
			//printf("Adding %f to Sdd\n", error_signal_block_sums[2*cl_blocks*num_channels + j*num_channels + i]);
		}
	}
}

/**
 * Specialized combined kernel
 */
void mdf_opencl_inner_prod_power_spec_spectral_mul(cl_mem Y, cl_mem Xf, cl_mem data_time_domain, cl_mem data_freq_domain, cl_mem foreground, cl_mem temp_acc, float *Sxx, int N, int frame_size, int M, int C, int K)
{
	size_t szLocalWorkSize[3];
        size_t szGlobalWorkSize[3];
	cl_int err = CL_SUCCESS;
	szLocalWorkSize[0] = 128;
	szLocalWorkSize[1] = 1;
	szLocalWorkSize[2] = 1;
	szGlobalWorkSize[0] = N/2;	// M blocks used by spectral_mul_accum, 1 block used for power_spectrum_accum
	szGlobalWorkSize[1] = C;
	szGlobalWorkSize[2] = 2;
	
	float *block_sums = (float *)malloc(sizeof(float) * K);
	int i;
	
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 0, sizeof(cl_mem), &Xf);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 1, sizeof(cl_mem), &data_time_domain);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 2, sizeof(cl_mem), &data_freq_domain);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 3, sizeof(cl_mem), &foreground);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 4, sizeof(cl_mem), &ocl_temp_reduction_buffer);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 5, sizeof(cl_mem), &Y);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 6, sizeof(cl_int), &N);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 7, sizeof(cl_int), &K);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 8, sizeof(cl_int), &M);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 9, sizeof(cl_int), &frame_size);
	print_on_opencl_error("Failed to set kernel argument 9 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
	clSetKernelArg(MDF_inner_prod_power_spec_spectral_mul_kernel, 10, sizeof(cl_float) * frame_size, NULL); //Shared mem
	print_on_opencl_error("Failed to set kernel argument 10 of MDF_inner_prod_power_spec_spectral_mul_kernel.", err);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_inner_prod_power_spec_spectral_mul_kernel, 3, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_inner_prod_power_spec_spectral_mul_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_inner_prod_power_spec_spectral_mul_kernel", 3);

	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_inner_prod_power_spec_spectral_mul_kernel", err);
	if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueBarrier of MDF_inner_prod_power_spec_spectral_mul_kernel", 3);
}

void mdf_opencl_prepare_shift_preemph_freq_data(cl_mem input, cl_mem x, cl_mem X, cl_mem X_temp, cl_mem far_end, cl_mem memD, cl_mem memX, float preemph, int window_size, int frame_size, int C, int K, int M)
{
	size_t szLocalWorkSize[3];
        size_t szGlobalWorkSize[3];
	
	cl_int err = CL_SUCCESS;
	
	szGlobalWorkSize[0] = window_size;
	szGlobalWorkSize[1] = M;
	szGlobalWorkSize[2] = K;
	//szGlobalWorkSize[2] = 1;
	
	szLocalWorkSize[0] = 64;	// Should be larger, not allowed on ATI (GPU)
	szLocalWorkSize[1] = 1;
	szLocalWorkSize[2] = 1;

        err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 0, sizeof(cl_mem), (void*)&input);
        print_on_opencl_error("Failed to set kernel argument 0 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
        err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 1, sizeof(cl_mem), (void*)&x);
        print_on_opencl_error("Failed to set kernel argument 1 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
        err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 2, sizeof(cl_mem), (void*)&X);
        print_on_opencl_error("Failed to set kernel argument 2 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
        err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 3, sizeof(cl_mem), (void*)&X_temp);
        print_on_opencl_error("Failed to set kernel argument 3 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
        err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 4, sizeof(cl_mem), (void*)&far_end);
        print_on_opencl_error("Failed to set kernel argument 4 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 5, sizeof(cl_mem), (void*)&memD);
        print_on_opencl_error("Failed to set kernel argument 5 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
        err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 6, sizeof(cl_mem), (void*)&memX);
        print_on_opencl_error("Failed to set kernel argument 6 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 7, sizeof(cl_float), (void*)&preemph);
        print_on_opencl_error("Failed to set kernel argument 7 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 8, sizeof(cl_uint), (void*)&frame_size);
        print_on_opencl_error("Failed to set kernel argument 8 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 9, sizeof(cl_uint), (void*)&window_size);
        print_on_opencl_error("Failed to set kernel argument 9 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 10, sizeof(cl_uint), (void*)&C);
        print_on_opencl_error("Failed to set kernel argument 10 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 11, sizeof(cl_uint), (void*)&K);
        print_on_opencl_error("Failed to set kernel argument 11 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 12, sizeof(cl_int)*K*frame_size, NULL);
        print_on_opencl_error("Failed to set kernel argument 12 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 13, sizeof(cl_float)*C*szLocalWorkSize[0]*(frame_size/szLocalWorkSize[0]), NULL);
        print_on_opencl_error("Failed to set kernel argument 13 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
	//err = clSetKernelArg(MDF_prepare_shift_preemph_freq_data_kernel, 13, sizeof(cl_mem), &ocl_temp_buffer);
        //print_on_opencl_error("Failed to set kernel argument 13 of MDF_prepare_shift_preemph_freq_data_kernel.", err);
        
	if(err == CL_SUCCESS)
                print_debug("SUCCESS: clSetKernelArg\n", 3);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_prepare_shift_preemph_freq_data_kernel, 3, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_prepare_shift_preemph_freq_data_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_prepare_shift_preemph_freq_data_kernel", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_prepare_shift_preemph_freq_data_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_prepare_shift_preemph_freq_data_kernel", err);
}

void mdf_opencl_compute_foreground_filter(cl_mem e, cl_mem input, int window_size, int frame_size, int C)
{
	size_t szLocalWorkSize[3];
        size_t szGlobalWorkSize[3];
	
	cl_int err = CL_SUCCESS;
	
	// Shift data in the frequency domain
	// Input: input, e
	// Output: e
	szLocalWorkSize[0] = 32;
	szLocalWorkSize[1] = 1;
	szLocalWorkSize[2] = 1;
	szGlobalWorkSize[0] = frame_size;
	szGlobalWorkSize[1] = C + 1;
	szGlobalWorkSize[2] = 1;

        err = clSetKernelArg(MDF_compute_foreground_filter_kernel, 0, sizeof(cl_mem), (void*)&e);
        print_on_opencl_error("Failed to set kernel argument 0 of MDF_compute_foreground_filter_kernel.", err);
        err = clSetKernelArg(MDF_compute_foreground_filter_kernel, 1, sizeof(cl_mem), (void*)&input);
        print_on_opencl_error("Failed to set kernel argument 1 of MDF_compute_foreground_filter_kernel.", err);
        err = clSetKernelArg(MDF_compute_foreground_filter_kernel, 2, sizeof(cl_uint), (void*)&window_size);
        print_on_opencl_error("Failed to set kernel argument 2 of MDF_compute_foreground_filter_kernel.", err);
        err = clSetKernelArg(MDF_compute_foreground_filter_kernel, 3, sizeof(cl_uint), (void*)&frame_size);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_compute_foreground_filter_kernel.", err);
        err = clSetKernelArg(MDF_compute_foreground_filter_kernel, 4, sizeof(cl_uint), (void*)&C);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_compute_foreground_filter_kernel.", err);
	
	if(err == CL_SUCCESS)
                print_debug("SUCCESS: clSetKernelArg\n", 3);

        err = clEnqueueNDRangeKernel(*commandQueue, MDF_compute_foreground_filter_kernel, 3, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_compute_foreground_filter_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_compute_foreground_filter_kernel", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_compute_foreground_filter_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_compute_foreground_filter_kernel", err);
}

void mdf_opencl_update_foreground_filter(cl_mem e, cl_mem window, cl_mem y, int window_size, int frame_size, int C)
{
        size_t szGlobalWorkSize[3];
	
	cl_int err = CL_SUCCESS;
	
	/* Apply a smooth transition so as to not introduce blocking artifacts */
	szGlobalWorkSize[0] = frame_size;
	szGlobalWorkSize[1] = C;
	szGlobalWorkSize[2] = 1;

	err = clSetKernelArg(MDF_update_foreground_filter_kernel, 0, sizeof(cl_mem), (void*)&ocl_e);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_update_foreground_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_foreground_filter_kernel, 1, sizeof(cl_mem), (void*)&ocl_window);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_update_foreground_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_foreground_filter_kernel, 2, sizeof(cl_mem), (void*)&ocl_y);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_update_foreground_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_foreground_filter_kernel, 3, sizeof(cl_uint), (void*)&window_size);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_update_foreground_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_foreground_filter_kernel, 4, sizeof(cl_uint), (void*)&frame_size);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_update_foreground_filter_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clSetKernelArg\n", 3);

	err = clEnqueueNDRangeKernel(*commandQueue, MDF_update_foreground_filter_kernel, 2, NULL, szGlobalWorkSize, NULL, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel MDF_update_foreground_filter_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_update_foreground_filter_kernel", 3);

	//err=clFinish(*commandQueue);
	//print_on_opencl_error("clFinish failed for MDF_update_foreground_filter_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_update_foreground_filter_kernel", err);
}

void mdf_opencl_update_background_filter(cl_mem y, cl_mem e, cl_mem input, int window_size, int frame_size, int C)
{
	size_t szGlobalWorkSize[3];
	
	cl_int err = CL_SUCCESS;
	
	/* Apply a smooth transition so as to not introduce blocking artifacts */
	szGlobalWorkSize[0] = frame_size;
	szGlobalWorkSize[1] = C;
	szGlobalWorkSize[2] = 1;

	err = clSetKernelArg(MDF_update_background_filter_kernel, 0, sizeof(cl_mem), (void*)&y);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_update_background_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_background_filter_kernel, 1, sizeof(cl_mem), (void*)&e);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_update_background_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_background_filter_kernel, 2, sizeof(cl_mem), (void*)&input);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_update_background_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_background_filter_kernel, 3, sizeof(cl_uint), (void*)&window_size);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_update_background_filter_kernel.", err);
	err = clSetKernelArg(MDF_update_background_filter_kernel, 4, sizeof(cl_uint), (void*)&frame_size);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_update_background_filter_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clSetKernelArg\n", 3);

	err = clEnqueueNDRangeKernel(*commandQueue, MDF_update_background_filter_kernel, 2, NULL, &szGlobalWorkSize, NULL, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel MDF_update_background_filter_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_update_background_filter_kernel", 3);

	//err=clFinish(*commandQueue);
	//print_on_opencl_error("clFinish failed for MDF_update_background_filter_kernel", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_update_background_filter_kernel", err);
}

void mdf_opencl_smooth_far_end_compute_filtered_spectra(cl_mem power, cl_mem Xf, cl_mem Rf, cl_mem Eh, cl_mem Yf, cl_mem Yh, float ss_1, float ss, float spec_average, float *Pey, float *Pyy, int frame_size)
{
	size_t szLocalWorkSize[3];
        size_t szGlobalWorkSize[3];
	
	cl_int err = CL_SUCCESS;
	
	int cl_threads = 128;
	int cl_blocks = (frame_size + cl_threads) / cl_threads;
	int j;
	
	szGlobalWorkSize[0] = cl_blocks * cl_threads;
	szGlobalWorkSize[1] = 1;
	szGlobalWorkSize[2] = 1;
	szLocalWorkSize[0] = cl_threads;
	szLocalWorkSize[1] = 1;
	szLocalWorkSize[2] = 1;

	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 0, sizeof(cl_mem), &power);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 1, sizeof(cl_mem), &Xf);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 2, sizeof(cl_mem), &Rf);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 3, sizeof(cl_mem), &Eh);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 4, sizeof(cl_mem), &Yf);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 5, sizeof(cl_mem), &Yh);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 6, sizeof(cl_float), &ss_1);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 7, sizeof(cl_float), &ss);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 8, sizeof(cl_float), &spec_average);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 9, sizeof(cl_uint), &frame_size);
	print_on_opencl_error("Failed to set kernel argument 9 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 10, sizeof(cl_mem), &ocl_temp_buffer);
	print_on_opencl_error("Failed to set kernel argument 10 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 11, sizeof(cl_mem), &ocl_temp_reduction_buffer);
	print_on_opencl_error("Failed to set kernel argument 11 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	clSetKernelArg(MDF_smooth_far_end_compute_filtered_spectra_kernel, 12, sizeof(float) * cl_threads * 2, NULL);
	print_on_opencl_error("Failed to set kernel argument 12 of MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	
	// Outputs into ocl_out, ocl_memE and ocl_saturated
        err = clEnqueueNDRangeKernel(*commandQueue, MDF_smooth_far_end_compute_filtered_spectra_kernel, 1, NULL, szGlobalWorkSize, szLocalWorkSize, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_smooth_far_end_compute_filtered_spectra_kernel.", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_smooth_far_end_compute_filtered_spectra_kernel.", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_smooth_far_end_compute_filtered_spectra_kernel", err);
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_temp_buffer, CL_TRUE, 0, sizeof(float)*cl_blocks*2, Pey_Pyy_reduction_host, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueReadBuffer failed on copy of Pey/Pyy reduction buffer back to host.", err);
	
	for(j=0;j<cl_blocks;j++)
	{
		*Pey += Pey_Pyy_reduction_host[2*j];
		//printf("Added %f to Pey\n", Pey_Pyy_reduction_host[2*j]);
		*Pyy += Pey_Pyy_reduction_host[2*j + 1];
		//printf("Added %f to Pyy\n", Pey_Pyy_reduction_host[2*j + 1]);
	}
}

void mdf_opencl_learning_rate_calc(cl_mem Yf, cl_mem Rf, cl_mem power_1, cl_mem power, float RER, float leak_estimate, float adapt_rate, int adapted, int frame_size)
{
	size_t szGlobalWorkSize[3];
	
	cl_int err = CL_SUCCESS;
	
	szGlobalWorkSize[0] = frame_size + 32;
	szGlobalWorkSize[1] = 1;
	szGlobalWorkSize[2] = 1;

	clSetKernelArg(MDF_learning_rate_calc_kernel, 0, sizeof(cl_mem), &Yf);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 1, sizeof(cl_mem), &Rf);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 2, sizeof(cl_mem), &power_1);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 3, sizeof(cl_mem), &power);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 4, sizeof(cl_float), &RER);
	print_on_opencl_error("Failed to set kernel argument 4 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 5, sizeof(cl_uint), &frame_size);
	print_on_opencl_error("Failed to set kernel argument 5 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 6, sizeof(cl_float), &leak_estimate);
	print_on_opencl_error("Failed to set kernel argument 6 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 7, sizeof(cl_float), &adapt_rate);
	print_on_opencl_error("Failed to set kernel argument 7 of MDF_learning_rate_calc_kernel.", err);
	clSetKernelArg(MDF_learning_rate_calc_kernel, 8, sizeof(cl_int), &adapted);
	print_on_opencl_error("Failed to set kernel argument 8 of MDF_learning_rate_calc_kernel.", err);
	
        err = clEnqueueNDRangeKernel(*commandQueue, MDF_learning_rate_calc_kernel, 1, NULL, szGlobalWorkSize, NULL, 0, NULL, NULL);
        print_on_opencl_error("Failed to execute kernel MDF_learning_rate_calc_kernel.", err);
        if(err == CL_SUCCESS)
                print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_learning_rate_calc_kernel.", 3);

        //err=clFinish(*commandQueue);
        //print_on_opencl_error("clFinish failed for MDF_learning_rate_calc_kernel.", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_learning_rate_calc_kernel", err);
}

void mdf_opencl_adapted_copy(cl_mem last_y, cl_mem in, cl_mem out, int frame_size)
{
	size_t szGlobalWorkSize[3];
	
	cl_int err = CL_SUCCESS;
	
	szGlobalWorkSize[0] = frame_size;
	szGlobalWorkSize[1] = 1;
	szGlobalWorkSize[2] = 1;

	clSetKernelArg(MDF_adapted_copy_kernel, 0, sizeof(cl_mem), &last_y);
	print_on_opencl_error("Failed to set kernel argument 0 of MDF_adapted_copy_kernel.", err);
	clSetKernelArg(MDF_adapted_copy_kernel, 1, sizeof(cl_mem), &in);
	print_on_opencl_error("Failed to set kernel argument 1 of MDF_adapted_copy_kernel.", err);
	clSetKernelArg(MDF_adapted_copy_kernel, 2, sizeof(cl_mem), &out);
	print_on_opencl_error("Failed to set kernel argument 2 of MDF_adapted_copy_kernel.", err);
	clSetKernelArg(MDF_adapted_copy_kernel, 3, sizeof(cl_uint), &frame_size);
	print_on_opencl_error("Failed to set kernel argument 3 of MDF_adapted_copy_kernel.", err);
	
	err = clEnqueueNDRangeKernel(*commandQueue, MDF_adapted_copy_kernel, 1, NULL, szGlobalWorkSize, NULL, 0, NULL, NULL);
	print_on_opencl_error("Failed to execute kernel MDF_adapted_copy_kernel.", err);
	if(err == CL_SUCCESS)
		print_debug("SUCCESS: clEnqueueNDRangeKernel of MDF_adapted_copy_kernel.", 3);

	//err=clFinish(*commandQueue);
	//print_on_opencl_error("clFinish failed for MDF_adapted_copy_kernel.", err);
	err = clEnqueueBarrier(*commandQueue);
	print_on_opencl_error("clEnqueueBarrier failed for MDF_adapted_copy_kernel", err);
}

EXPORT void speex_echo_cancellation_opencl(SpeexEchoState *st, const spx_int16_t *in, const spx_int16_t *far_end, spx_int16_t *out) 
{
	struct timeval start, end;
	set_start_reference_opencl();
	size_t szLocalWorkSize[] = {32,0,0};
        size_t szGlobalWorkSize[] = {0,0,0};

	int i,j, chan, speak;
	int N,M, C, K;
	spx_word32_t Syy,See,Sxx,Sdd, Sff;
	int zero = 0;
	#ifdef TWO_PATH
		spx_word32_t Dbf;
		int update_foreground;
	#endif
	spx_word32_t Sey;
	spx_word16_t ss, ss_1;
	spx_float_t Pey = FLOAT_ONE, Pyy=FLOAT_ONE;
	spx_float_t alpha, alpha_1;
	spx_word16_t RER;
	spx_word32_t tmp32;

	N = st->window_size;
	M = st->M;
	C = st->C;
	K = st->K;

	st->cancel_count++;
	ss=.35/M;
	ss_1 = 1-ss;
	
	cl_int err = CL_SUCCESS;

	//printf("Beginning MDF filter with OpenCL.\n");
	//gettimeofday(&start, NULL);

	// Do notch filter on CPU, not very parallelizable, and prone to rounding errorss
	for (chan = 0; chan < st->C; chan++)
	{
		filter_dc_notch16(in+chan, st->notch_radius, st->input+chan*st->frame_size, st->frame_size, st->notch_mem+2*chan, st->C);
	}
	// Find memD, this is done on CPU to avoid a global sync on the device
	// Copied into temporary array because of dep. in first kernel
	for(chan=0;chan<st->C;chan++)
	{
		host_temp_memD[chan] = st->input[chan*st->frame_size + st->frame_size - 1];
	}
	save_new_timingpoint("Start of MDF filter.");
	
	// Initialize data on GPU
	speex_echo_opencl_copy_to_GPU(st, in, far_end, out);
	
	save_new_timingpoint("Done with copy to device.");
	
	check_against_checkpoint_opencl(0, ocl_input, 0, 0);

	// Include in next kernel?
	mdf_opencl_copy_float_array(data_freq_domain, 0, data_freq_domain_temp, 0, st->K*(st->M+1)*st->window_size);
	
	save_new_timingpoint("Done with clMDF_copy_float_array.");
	
	check_against_checkpoint_opencl(8, ocl_memX, 0, 0);
	check_against_checkpoint_opencl(9, data_time_domain, 0, 0);
	
	mdf_opencl_prepare_shift_preemph_freq_data(ocl_input, data_time_domain, data_freq_domain, data_freq_domain_temp, ocl_far_end, ocl_memD, ocl_memX, st->preemph, st->window_size, st->frame_size, st->C, st->K, st->M);
	
	//mdf_opencl_set_array_to_float_value(data_time_domain, 0, 9000.0f, st->frame_size);
	
	save_new_timingpoint("Done with clMDF_prepare_shift_preemph_freq_data.");
	
	check_against_checkpoint_opencl(6, ocl_notch_mem, 0, 0);
	check_against_checkpoint_opencl(10, ocl_memX, 0, 0);
	// Extra for combined kernel
	check_against_checkpoint_opencl(11, data_time_domain, 0, 0);
	check_against_checkpoint_opencl(12, ocl_input, 0, 0);
	
	err = clEnqueueWriteBuffer(*commandQueue, ocl_memD, CL_TRUE, 0, st->C*sizeof(cl_float), host_temp_memD, 0, NULL, NULL);
	print_on_opencl_error("clEnqueueWriteBuffer of ocl_memD failed", err);
	check_against_checkpoint_opencl(13, ocl_memD, 0, 0);
	
	save_new_timingpoint("Copied memD to device.");
	
	Sxx = 0;
	
	speex_echo_opencl_fft_inner_prod_combined(data_time_domain, data_freq_domain, ocl_Xf, ocl_temp_reduction_buffer, &Sxx, st->frame_size, st->K, commandQueue);
	
	check_against_checkpoint_opencl(20, data_freq_domain, 0, 0);
	check_against_checkpoint_opencl(21, data_time_domain, 0, 0);

	save_new_timingpoint("Done with clFFT_custom.");
	
	Sff = 0;
	
	check_against_checkpoint_opencl(22, ocl_foreground, 0, 0);
	mdf_opencl_inner_prod_power_spec_spectral_mul(ocl_Y, ocl_Xf, data_time_domain, data_freq_domain, ocl_foreground, ocl_temp_acc, &Sxx, st->window_size, st->frame_size, st->M*st->K, st->C, st->K);
	
	save_new_timingpoint("Done with clMDF_inner_prod_power_spec_spectral_mul.");
	
	check_against_checkpoint(25, &Sxx);
	check_against_checkpoint_opencl(31, data_freq_domain, 0, 0);
	check_against_checkpoint_opencl(32, ocl_Xf, 0, 0);
	check_against_checkpoint_opencl(38, ocl_Y, 0, 0);

	speex_echo_opencl_ifft_execute2(ocl_Y, 0, ocl_e, 0, st->C, commandQueue);	
	
	check_against_checkpoint_opencl(33, ocl_e, 0, 0);
	
	mdf_opencl_compute_foreground_filter(ocl_e, ocl_input, st->window_size, st->frame_size, st->C);

	save_new_timingpoint("Done with clMDF_compute_foreground_filter.");
	
	check_against_checkpoint_opencl(46, ocl_e, 0, 0);
	// Input: e
	// Output: Sff
	Sff = mdf_opencl_inner_prod(ocl_e, 0, ocl_e, 0, st->frame_size, st->C, st->window_size);
	
	save_new_timingpoint("Done with clMDF_inner_product_reduce.");
	
	check_against_checkpoint_opencl(40, ocl_Y, 0, 0);
	check_against_checkpoint_opencl(41, ocl_e, 0, 0);
	check_against_checkpoint_opencl(45, ocl_input, 0, 0);
	check_against_checkpoint(50, &Sff);

	check_against_checkpoint_opencl(51, ocl_prop, 0, 0);
	
	if(st->adapted)
	{
		// Input: W
		// Output: prop
		//printf("\n\n\nADAPTED FORK OPENCL (cancel count %d)\n\n\n", st->cancel_count);
		mdf_opencl_adjust_prop(ocl_W, st->window_size, st->M, st->C*st->K, ocl_prop);
		save_new_timingpoint("Done with clMDF_adjust_prop.");
	}
	
	check_against_checkpoint_opencl(55, ocl_prop, 0, 0);
	
	// Copy saturated to host for decision
	int temp_saturated;
	err = clEnqueueReadBuffer(*commandQueue, ocl_saturated, CL_TRUE, 0, sizeof(int), &temp_saturated, 0, NULL, NULL);
	save_new_timingpoint("Read saturated back from device.");
	check_against_checkpoint_opencl(56, data_freq_domain, 0, 0);
	check_against_checkpoint_opencl(57, ocl_E, 0, 0);
	check_against_checkpoint_opencl(58, ocl_prop, 0, 0);
	check_against_checkpoint_opencl(59, ocl_power_1, 0, 0);
	if(temp_saturated == 0)
	{
		// Input: power_1, prop, data_freq_domain, E
		// Output: W, PHI
		//printf("\n\n\nRUNNING SPECTRAL MUL CONJ OPENCL\n\n\n");
		mdf_opencl_weighted_spectral_mul_conj(st->C, st->K, st->M, ocl_power_1, ocl_prop, data_freq_domain, ocl_E, ocl_PHI, st->window_size, ocl_W);
		save_new_timingpoint("Done with clMDF_weigthed_spectral_mul_conj.");
		check_against_checkpoint_opencl(60, ocl_PHI, 0, 0);
		check_against_checkpoint_opencl(61, ocl_W, 0, 0);
	}
	else
	{
		temp_saturated--;
	}
	// Copy saturated back
	err = clEnqueueWriteBuffer(*commandQueue, ocl_saturated, CL_TRUE, 0, sizeof(int), &temp_saturated, 0, NULL, NULL);
        print_on_opencl_error("clEnqueueWriteBuffer of ocl_saturated failed", err);

	save_new_timingpoint("Wrote saturated to device.");

	check_against_checkpoint_opencl(64, ocl_W, 0, 0);
	
	speex_echo_opencl_fft_update_weights(ocl_W, ocl_wtmp, st->M, st->C, st->K, commandQueue);
	check_against_checkpoint_opencl(65, ocl_W, 0, 0);
	
	//save_new_timingpoint("Updated weights.");
	save_new_timingpoint("Done with clFFT_update_weigths.");

	Dbf = 0;
	See = 0;
	
	check_against_checkpoint_opencl(70, ocl_Y, 0, 0);
	
	check_against_checkpoint_opencl(66, ocl_W, 0, 0);
	check_against_checkpoint_opencl(74, data_freq_domain, 0, 0);

	mdf_opencl_spectral_mul_accum(data_freq_domain, ocl_W, ocl_Y, ocl_Rf, ocl_Yf, ocl_Xf, st->window_size, st->K*st->M, st->K, st->C);
	
	save_new_timingpoint("Done with clMDF_spectral_mul_accum.");
	
	check_against_checkpoint_opencl(75, ocl_Y, 0, 0);
	
	speex_echo_opencl_ifft_execute2(ocl_Y, 0, ocl_y, 0, C, commandQueue);
	
	save_new_timingpoint("Done with clFFT_custom.");
	
	check_against_checkpoint_opencl(67, ocl_Rf, 0, 0);
	check_against_checkpoint_opencl(68, ocl_Yf, 0, 0);
	check_against_checkpoint_opencl(69, ocl_Xf, 0, 0);
	check_against_checkpoint_opencl(71, ocl_y, 0, 0);
	check_against_checkpoint_opencl(72, ocl_Y, 0, 0);

	check_against_checkpoint_opencl(73, ocl_e, 0, 0);
	
	mdf_opencl_response_diff_inner_product(ocl_e, ocl_temp_e, ocl_y, ocl_input, st->C, st->window_size, st->frame_size, &Dbf, &See);
	
	check_against_checkpoint(76, &Dbf);
	check_against_checkpoint_opencl(77, ocl_e, 0, 0);
	check_against_checkpoint(78, &See);
	
	save_new_timingpoint("Done with clMDF_response_diff_inner_product.");
	
	Sff = See;
	
	// Logic for updating the foreground filter, do on CPU

	// For two time windows, compute the mean of the energy difference, as well as the variance
	// st->Davg1 = .6*st->Davg1 + .4*(Sff-See);
	// st->Davg2 = .85*st->Davg2 + .15*(Sff-See);
	// st->Dvar1 = .36*st->Dvar1 + .16*Sff*Dbf;
	// st->Dvar2 = .7225*st->Dvar2 + .0225*Sff*Dbf;
	check_against_checkpoint(180, &See);
	check_against_checkpoint(181, &Sff);
	check_against_checkpoint(182, &Dbf);
	check_against_checkpoint(183, &st->Davg1);
	check_against_checkpoint(184, &st->Davg2);
	check_against_checkpoint(185, &st->Dvar1);
	check_against_checkpoint(186, &st->Dvar2);
	
	st->Davg1 = ADD32(MULT16_32_Q15(QCONST16(.6f,15),st->Davg1), MULT16_32_Q15(QCONST16(.4f,15),SUB32(Sff,See)));
	st->Davg2 = ADD32(MULT16_32_Q15(QCONST16(.85f,15),st->Davg2), MULT16_32_Q15(QCONST16(.15f,15),SUB32(Sff,See)));
	st->Dvar1 = FLOAT_ADD(FLOAT_MULT(VAR1_SMOOTH, st->Dvar1), FLOAT_MUL32U(MULT16_32_Q15(QCONST16(.4f,15),Sff), MULT16_32_Q15(QCONST16(.4f,15),Dbf)));
	st->Dvar2 = FLOAT_ADD(FLOAT_MULT(VAR2_SMOOTH, st->Dvar2), FLOAT_MUL32U(MULT16_32_Q15(QCONST16(.15f,15),Sff), MULT16_32_Q15(QCONST16(.15f,15),Dbf)));

	check_against_checkpoint(190, &st->Davg1);
	check_against_checkpoint(191, &st->Davg2);
	check_against_checkpoint(192, &st->Dvar1);
	check_against_checkpoint(193, &st->Dvar2);
	
	update_foreground = 0;
	// Check if we have a statistically significant reduction in the residual echo
	// Note that this is *not* Gaussian, so we need to be careful about the longer tail
	// if (((Sff - See) * ABS32(Sff - See)) > (Sff * Dbf))
	// 	update_foreground = 1;
	// else if ((st->Davg1 * ABS32(st->Davg1)) > (VAR1_UPDATE * (st->Dvar1)))
	// 	update_foreground = 1;
	// else if ((st->Davg2 * ABS32(st->Davg2)) > (VAR2_UPDATE * (st->Dvar2)))
	// 	update_foreground = 1;
	if (FLOAT_GT(FLOAT_MUL32U(SUB32(Sff,See),ABS32(SUB32(Sff,See))), FLOAT_MUL32U(Sff,Dbf)))
		update_foreground = 1;
	else if (FLOAT_GT(FLOAT_MUL32U(st->Davg1, ABS32(st->Davg1)), FLOAT_MULT(VAR1_UPDATE,(st->Dvar1))))
		update_foreground = 1;
	else if (FLOAT_GT(FLOAT_MUL32U(st->Davg2, ABS32(st->Davg2)), FLOAT_MULT(VAR2_UPDATE,(st->Dvar2))))
		update_foreground = 1;
	
	check_against_checkpoint_opencl(79, ocl_e, 0, 0);
	
	save_new_timingpoint("Some correlation logic done on CPU.");
	
	// Do we update?
	if (update_foreground)
	{
		st->Davg1 = st->Davg2 = 0;
		st->Dvar1 = st->Dvar2 = FLOAT_ZERO;
		
		mdf_opencl_copy_float_array(ocl_W, 0, ocl_foreground, 0, st->window_size*st->M*st->C*st->K);
		
		save_new_timingpoint("Done with clMDF_copy_float_array.");
		
		// Apply a smooth transition so as to not introduce blocking artifacts
		mdf_opencl_update_foreground_filter(ocl_e, ocl_window, ocl_y, st->window_size, st->frame_size, st->C);
		
		save_new_timingpoint("Done with clMDF_update_foreground_filter.");
		
		check_against_checkpoint_opencl(80, ocl_foreground, 0, 0);
		check_against_checkpoint_opencl(81, ocl_e, 0, 0);
		
		print_debug("OpenCL: Updated foreground", 3);
	}
	else
	{
		int reset_background=0;
			// Otherwise, check if the background filter is significantly worse
		if (FLOAT_GT(FLOAT_MUL32U(NEG32(SUB32(Sff,See)),ABS32(SUB32(Sff,See))), FLOAT_MULT(VAR_BACKTRACK,FLOAT_MUL32U(Sff,Dbf))))
			reset_background = 1;
		if (FLOAT_GT(FLOAT_MUL32U(NEG32(st->Davg1), ABS32(st->Davg1)), FLOAT_MULT(VAR_BACKTRACK,st->Dvar1)))
			reset_background = 1;
		if (FLOAT_GT(FLOAT_MUL32U(NEG32(st->Davg2), ABS32(st->Davg2)), FLOAT_MULT(VAR_BACKTRACK,st->Dvar2)))
			reset_background = 1;
		// Otherwise, check if the background filter is significantly worse
		//if ((-(Sff - See) * ABS32(Sff - See)) > (VAR_BACKTRACK * (Sff * Dbf)))
		//	reset_background = 1;
		//if ((-st->Davg1 * ABS32(st->Davg1)) > (VAR_BACKTRACK * st->Dvar1))
		//	reset_background = 1;
		//if ((-st->Davg2 * ABS32(st->Davg2)) > (VAR_BACKTRACK * st->Dvar2))
		//	reset_background = 1;
		
		if (reset_background)
		{
			// Copy foreground filter to background filter
			//for (i=0;i<N*M*C*K;i++)
			//	st->W[i] = st->foreground[i];
			// Copy background filter to foreground filter 
			mdf_opencl_copy_float_array(ocl_foreground, 0, ocl_W, 0, st->window_size*st->M*st->C*st->K);
			
			save_new_timingpoint("Done with clMDF_copy_float_array.");
			
			// Apply a smooth transition so as to not introduce blocking artifacts
			mdf_opencl_update_background_filter(ocl_y, ocl_e, ocl_input, st->window_size, st->frame_size, st->C);
			
			save_new_timingpoint("Done with clMDF_update_background_filter.");
			
			check_against_checkpoint_opencl(82, ocl_W, 0, 0);
			check_against_checkpoint_opencl(83, ocl_y, 0, 0);
			check_against_checkpoint_opencl(84, ocl_e, 0, 0);
			
			See = Sff;
			st->Davg1 = st->Davg2 = 0;
			st->Dvar1 = st->Dvar2 = FLOAT_ZERO;
			print_debug("OpenCL: Reset background.", 3);
		}
		else
		{
			print_debug("OpenCL: No foreground/background filter update", 3);
		}
	}
	
	Sey = Syy = Sdd = 0;

	check_against_checkpoint_opencl(85, ocl_input, 0, 0);
	check_against_checkpoint_opencl(86, ocl_e, 0, 0);
	// Compute the error signal
	
	// This clFinish increases performance by the power of magic
	err=clFinish(*commandQueue);
        print_on_opencl_error("clFinish before MDF_compute_error_signal_combined", err);
	
	mdf_opencl_compute_error_signal_combined(ocl_input, ocl_e, ocl_y, ocl_in, ocl_out, ocl_saturated,
					st->preemph,
					st->window_size,
					st->frame_size,
					st->C,
					&Sey, &Syy, &Sdd, st->memE, out);
					
	save_new_timingpoint("Done with clMDF_compute_error_signal_combined.");

	check_against_checkpoint_opencl(88, ocl_saturated, 0, 0);
	check_against_checkpoint_opencl(89, ocl_out, 0, 1);
					
	szGlobalWorkSize[0] = st->frame_size;
	szGlobalWorkSize[1] = st->C;
	szGlobalWorkSize[2] = 1;

	// Combined kernel for converting to frequency domain and setting the value of y
	speex_echo_opencl_fft_convert_error(ocl_e, ocl_E, ocl_y, ocl_Y, st->C, commandQueue);

	// Temporary variable because of combined kernels
	float temp_Sxx = Sxx;

	Sxx += mdf_opencl_power_spectrum_inner_prod(data_time_domain, ocl_E, ocl_Y, data_freq_domain, ocl_Rf, ocl_Yf, ocl_Xf, 0, 0, 0, st->C, st->K, st->window_size, st->frame_size, st->frame_size, st->window_size, st->frame_size);

	save_new_timingpoint("Done with clMDF_power_spectrum_inner_prod.");
	
	check_against_checkpoint(93, (float *)&Sey);
	check_against_checkpoint(95, (float *)&Syy);
	check_against_checkpoint(97, (float *)&Sdd);
	check_against_checkpoint_opencl(100, ocl_e, 0, 0);
	check_against_checkpoint_opencl(101, ocl_E, 0, 0);
	check_against_checkpoint_opencl(102, ocl_y, 0, 0);
	check_against_checkpoint_opencl(103, ocl_Y, 0, 0);
	check_against_checkpoint_opencl(105, ocl_Rf, 0, 0);
	check_against_checkpoint_opencl(106, ocl_Yf, 0, 0);

	// Do some sanity check
	if (!(Syy>=0 && temp_Sxx>=0 && See >= 0))
	{
		// Things have gone really bad
		st->screwed_up += 50;
		
		mdf_opencl_set_array_to_float_value(ocl_out, 0, 0, st->frame_size*st->C);
	} 
	else if (SHR32(Sff, 2) > ADD32(Sdd, SHR32(MULT16_16(N, 10000),6)))
	{
		// AEC seems to add lots of echo instead of removing it, let's see if it will improve
		st->screwed_up++;
	} 
	else 
	{
		//speex_warning("The OpenCL filter passed the sanity test.");
		// Everything's fine
		st->screwed_up=0;
	}
	if (st->screwed_up>=50)
	{
		speex_warning("The OpenCL echo canceller started acting funny and got slapped (reset). It swears it will behave now.");
		speex_echo_state_reset(st);
		return;
	}
	
	// Add a small noise floor to make sure not to have problems when dividing
	See = MAX32(See, N * 100);
	
	save_new_timingpoint("Done with sanity check.");
	check_against_checkpoint(110, &See);
	
	check_against_checkpoint(115, &Sxx);
	check_against_checkpoint_opencl(116, ocl_Xf, 0, 0);
	check_against_checkpoint_opencl(117, ocl_Yf, 0, 0);
	check_against_checkpoint(120, &Pey);
	check_against_checkpoint(121, &Pyy);
	
	mdf_opencl_smooth_far_end_compute_filtered_spectra(ocl_power, ocl_Xf, ocl_Rf, ocl_Eh, ocl_Yf, ocl_Yh, ss_1, ss, st->spec_average, &Pey, &Pyy, st->frame_size);
	
	save_new_timingpoint("Read back temporary buffer for correlations (Pey/Pyy).");
	
	check_against_checkpoint(125, &Pey);
	check_against_checkpoint(126, &Pyy);
	check_against_checkpoint_opencl(127, ocl_power, 0, 0);
	check_against_checkpoint_opencl(128, ocl_Eh, 0, 0);
	check_against_checkpoint_opencl(129, ocl_Yh, 0, 0);
	
	Pyy = FLOAT_SQRT(Pyy);
	Pey = FLOAT_DIVU(Pey,Pyy);
	
	// Compute correlation update rate
	tmp32 = st->beta0 * Syy;
	if (tmp32 > (st->beta_max * See))
		tmp32 = st->beta_max * See;
	alpha = tmp32 / See;
	alpha_1 = FLOAT_ONE - alpha;
	check_against_checkpoint(130, &alpha);
	check_against_checkpoint(131, &alpha_1);
	
	// Update correlations (recursive average)
	st->Pey = (alpha_1 * st->Pey) + (alpha * Pey);
	st->Pyy = (alpha_1 * st->Pyy) + (alpha * Pyy);
	if (st->Pyy < FLOAT_ONE)
		st->Pyy = FLOAT_ONE;
	check_against_checkpoint(135, &st->Pyy);
	
	// We don't really hope to get better than 33 dB (MIN_LEAK-3dB) attenuation anyway
	if (st->Pey < (MIN_LEAK * st->Pyy))
		st->Pey = MIN_LEAK * st->Pyy;
	if (st->Pey > st->Pyy)
		st->Pey = st->Pyy;
	check_against_checkpoint(140, &st->Pey);
	
	// leak_estimate is the linear regression result
	st->leak_estimate = st->Pey / st->Pyy;
	// This looks like a stupid bug, but it's right (because we convert from Q14 to Q15)
	if (st->leak_estimate > 16383)
		st->leak_estimate = 32767;
	else
		st->leak_estimate = st->leak_estimate;
	check_against_checkpoint(145, &st->leak_estimate);
	
	RER = (0.0001 * Sxx + 3.0 * (st->leak_estimate * Syy)) / See;
	// Check for y in e (lower bound on RER)
	if (RER < Sey*Sey/(1+See*Syy))
		RER = Sey*Sey/(1+See*Syy);
	if (RER > .5)
		RER = .5;
	check_against_checkpoint(150, &RER);
	
	// We consider that the filter has had minimal adaptation if the following is true
	if (!st->adapted && st->sum_adapt > M && (st->leak_estimate * Syy) > (0.03f * Syy))
	{
		st->adapted = 1;
	}

	check_against_checkpoint_opencl(151, ocl_power, 0, 0);
	check_against_checkpoint_opencl(152, ocl_power_1, 0, 0);
	check_against_checkpoint(153, &Sxx);
	check_against_checkpoint(154, &See);
	
	// Temporary adaption rate if filter is not yet adapted enough 
	float adapt_rate = 0;
	if(!st->adapted)
	{
		if (Sxx > (st->window_size * 1000)) 
		{
			tmp32 = 0.25f * Sxx;

			if (tmp32 > 0.25*See)
				tmp32 = 0.25*See;
			adapt_rate = tmp32 / See;
		}
		// How much have we adapted so far?
		st->sum_adapt = st->sum_adapt + adapt_rate;
	}

	save_new_timingpoint("Done with CPU control logic for the filter.");

	check_against_checkpoint(155, &adapt_rate);
	check_against_checkpoint(156, &st->sum_adapt);
	
	mdf_opencl_learning_rate_calc(ocl_Yf, ocl_Rf, ocl_power_1, ocl_power, RER, st->leak_estimate, adapt_rate, st->adapted, st->frame_size);
	
	save_new_timingpoint("Done wite clMDF_learning_rate_calc.");
	
	check_against_checkpoint_opencl(160, ocl_power_1, 0, 0);
	check_against_checkpoint_opencl(161, ocl_last_y, 0, 0);
	
	// FIXME: MC conversion required 
	mdf_opencl_copy_float_array(ocl_last_y, st->frame_size, ocl_last_y, 0, st->frame_size);
	
	save_new_timingpoint("Done with clMDF_copy_float_array.");
	
	check_against_checkpoint_opencl(162, ocl_last_y, 0, 0);
	
	if(st->adapted)
	{
		mdf_opencl_adapted_copy(ocl_last_y, ocl_in, ocl_out, st->frame_size);
		
		save_new_timingpoint("Done with clMDF_adapted_copy_kernel.");
	}
	check_against_checkpoint_opencl(165, ocl_last_y, 0, 0);
	
	save_new_timingpoint("Output data done on GPU.");
	
	err = clEnqueueReadBuffer(*commandQueue, ocl_out, CL_TRUE, 0, st->C*st->frame_size*sizeof(cl_int), out_host_dummy, 0, NULL, NULL);
	
	err=clFinish(*commandQueue);
	print_on_opencl_error("Ending clFinish failed.", err);
	
	for(i=0;i<st->frame_size*st->C;i++)
	{
		out[i] = out_host_dummy[i];
	}
	check_against_checkpoint(166, (short *)out);
	
	save_new_timingpoint("Output data copied back to host.");
	
	speex_echo_opencl_store_execution_time(print_and_reset_timingpoints("MDF filter with OpenCL."));
	//print_and_reset_timingpoints("MDF filter with OpenCL.");
	
	//gettimeofday(&end, NULL);
	
	//printf("Stopped GPU execution!\n");
	
	//long long tusecs = (long long) (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec);
	//double t = (double)(tusecs/1000000.0);
	//printf("Computation time on GPU (seconds): %f\n", t);
}
