/* Copyright (C) 2004 Jean-Marc Valin

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// This file contains utility functions used in the OpenCL MDF implementation.
// Here you will find timing and validation functions, as well as debug printing functions.

// Debug level 1-3
// Level 3: Everything, including verification (lots of overhead, do not use this when timing)
// Level 2: Mostly unused
// Level 1; Only timing output (adds very little overhead inside the iteration itself)
// Level 0: No debug output

#define SPEEX_OCL_DEBUG_LEVEL 3

int debug_level = SPEEX_OCL_DEBUG_LEVEL;
static const int MAX_NUM_CHECKPOINTS = 260;
static const int MAX_NUM_TIMINGPOINTS = 100;
float **checkpoint_data = NULL;
double *timingpoint_data = NULL;
char **timingpoint_strings = NULL;
struct timeval current_time;

// For statistics (for all iterations of the filter)
double *exec_times_ocl = NULL;
double *exec_times_cpu = NULL;
int exec_num_ocl;
int exec_num_cpu;

// Numbering of short and float checkpoints are interleaved
int *checkpoint_sizes = NULL;
int num_timingpoints = NULL;

short **checkpoint_data_short = NULL;

void print_debug(char *string, int level) 
{
	if(level <= debug_level)
		printf("%s\n", string);
} 

void print_on_opencl_error(char *string, cl_int err) 
{
	if(err != CL_SUCCESS)
	{
		printf("%s\n", string);
		print_opencl_error(err);
	}
}

void print_error(char *string, int stop_condition)
{
	//printf("%s\n", string);
	print_debug(string, 1);
	if(stop_condition > 0)
	{
		printf("Fatal error, exiting.\n");
		exit(-1);
	}
}

#if SPEEX_OCL_DEBUG_LEVEL >= 1
//#if 0

EXPORT void init_timingpoints()
{
	int i;
	
	timingpoint_strings = (char **) malloc(sizeof(char *) * MAX_NUM_TIMINGPOINTS);
	timingpoint_data = (double *) malloc(sizeof(double) * MAX_NUM_TIMINGPOINTS);
	
	num_timingpoints = 0;
	
	for(i = 0;i < MAX_NUM_TIMINGPOINTS;i++) 
	{
		timingpoint_data[i] = -1.0;
		timingpoint_strings[i] = NULL;
	}
}

// Prints values and resets the timingpoints, NOT a cleanup function
// Returns total time (in seconds), or -1 if no points are stored
double print_and_reset_timingpoints(const char *title)
{
	int i;
	//double orig_first_timingpoint;
	struct timeval end_time;
	gettimeofday(&end_time, NULL);
	double end_time_usecs;
	double current_time = 0.0;
	double last_time = 0.0;
	
	if(num_timingpoints > 0)
	{
		end_time_usecs = end_time.tv_sec * 1000000 + end_time.tv_usec - timingpoint_data[0];
		printf("**** Start of timing printout for %s ****\n", title);
		for(i=0;i<num_timingpoints;i++)
		{
			current_time = (timingpoint_data[i] - timingpoint_data[0]) / 1000000.0;
			// Line up diffs in console, presuming we never exceed 100 timingpoints
			if(i < 10)
				printf("Time %d   +%f (%s): %f\n", i, current_time - last_time, timingpoint_strings[i], current_time);
			else
				printf("Time %d  +%f (%s): %f\n", i, current_time - last_time, timingpoint_strings[i], current_time);
			last_time = current_time;
			timingpoint_data[i] = -1.0;
		}
		printf("Time on printout: %f\n", end_time_usecs / 1000000.0);
		printf("**** End of timing printout ****\n");
		for(i=0;i<num_timingpoints;i++)
		{
			free(timingpoint_strings[i]);
			timingpoint_strings[i] = NULL;
		}
		num_timingpoints = 0;
		
		return end_time_usecs / 1000000.0;
	}
	else
	{
	    return 0.0;
	}
}

int save_new_timingpoint(const char *desc)
{
	cl_int err = CL_SUCCESS;
	if(commandQueue != NULL)
	{
		err=clFinish(*commandQueue);
		print_on_opencl_error("clFinish failed during save_new_timingpoint.", err);
	}
	
	// Read the timer value
	gettimeofday(&current_time, NULL);
	
	if(num_timingpoints < MAX_NUM_CHECKPOINTS)
	{
		if(num_timingpoints > 0)
		{
			timingpoint_data[num_timingpoints] = current_time.tv_sec * 1000000 + current_time.tv_usec - timingpoint_data[0];
		}
		else
		{
			timingpoint_data[num_timingpoints] = current_time.tv_sec * 1000000 + current_time.tv_usec;
		}
		// Copy the string
		int input_string_len = strlen(desc);
		//printf("Adding timingpoint num %d with string length: %d\n", num_timingpoints, input_string_len);
		timingpoint_strings[num_timingpoints] = (char *) malloc(sizeof(char) * input_string_len + 1);	
		strcpy(timingpoint_strings[num_timingpoints], desc);
		num_timingpoints++;
		// Return index of the newly stored timing point
		return num_timingpoints - 1;
	}
	else
	{
		return -1;
	}
}

// Disabled when using regular timingpoints
EXPORT void set_start_reference_opencl() {}
EXPORT void set_start_reference_cpu() {}

EXPORT void cleanup_timingpoints()
{
	int i;
	for(i = 0;i < MAX_NUM_TIMINGPOINTS;i++)
	{
		if(timingpoint_strings[i] != NULL)
		{
			free(timingpoint_strings[i]);
		}
	}
	free(timingpoint_data);
	free(timingpoint_strings);
}
#else
	double start_ref_time_cpu;
	double start_ref_time_ocl;
	
	EXPORT void init_timingpoints() {}
	double print_and_reset_timingpoints(const char *title) {}
	int save_new_timingpoint(const char *desc) {}
	EXPORT void cleanup_timingpoints() {}
	
	// Sets a start reference for use when not using timing points, but only calculating average
	EXPORT void set_start_reference_opencl()
	{
		cl_int err = CL_SUCCESS;
		if(commandQueue != NULL)
		{
			err=clFinish(*commandQueue);
			print_on_opencl_error("clFinish failed during set_start_reference_opencl.", err);
		}
		
		// Read the timer value
		gettimeofday(&current_time, NULL);
		
		start_ref_time_ocl = current_time.tv_sec * 1000000 + current_time.tv_usec;
	}
	
	EXPORT void set_start_reference_cpu()
	{
		cl_int err = CL_SUCCESS;
		if(commandQueue != NULL)
		{
			err=clFinish(*commandQueue);
			print_on_opencl_error("clFinish failed during set_start_reference_cpu.", err);
		}
		
		// Read the timer value
		gettimeofday(&current_time, NULL);
		
		start_ref_time_cpu = current_time.tv_sec * 1000000 + current_time.tv_usec;
	}
#endif

/**
 * Adds a new execution time for the filter using OpenCL
 * Used for statistics/benchmarking
 * This functions executes regardless of debug level (only adds overhead after timings are done)
 */
EXPORT void speex_echo_opencl_store_execution_time(double exec_time)
{
	// If timing points are used, accept the input
	#if SPEEX_OCL_DEBUG_LEVEL >= 1
	if(exec_times_ocl != NULL)
	{
		// Bounds check
		if(exec_num_ocl < MAX_EXEC_TIMINGS)
		{
			// Add the newest time
			exec_times_ocl[exec_num_ocl] = exec_time;
			printf("Adding exec_time OpenCL: %f\n", exec_time);
			exec_num_ocl++;
		}
	}
	// Otherwise, measure the time and use the start reference
	#else
	// Bounds check
	if(exec_num_ocl < MAX_EXEC_TIMINGS)
	{
		cl_int err = CL_SUCCESS;
		double exec_time_real;
		if(commandQueue != NULL)
		{
			err=clFinish(*commandQueue);
			print_on_opencl_error("clFinish failed during speex_echo_opencl_store_execution_time_cpu.", err);
		}
		
		// Read the timer value
		gettimeofday(&current_time, NULL);
		exec_time_real = current_time.tv_sec * 1000000 + current_time.tv_usec - start_ref_time_ocl;
		exec_times_ocl[exec_num_ocl] = exec_time_real;
		//printf("Adding exec_time OpenCL: %f\n", exec_time_real);
		exec_num_ocl++;
	}
	#endif
}

/**
 * Adds a new execution time for the filter on CPU
 * Used for statistics/benchmarking
 * This functions executes regardless of debug level (only adds overhead after timings are done)
 */
EXPORT void speex_echo_opencl_store_execution_time_cpu(double exec_time)
{
	// If timing points are used, accept the input
	#if SPEEX_OCL_DEBUG_LEVEL >= 1
	if(exec_times_cpu != NULL)
	{
		// Bounds check
		if(exec_num_cpu < MAX_EXEC_TIMINGS)
		{
			// Add the newest time
			exec_times_cpu[exec_num_cpu] = exec_time;
			printf("Adding exec_time CPU: %f\n", exec_time);
			exec_num_cpu++;
		}
	}
	// Otherwise, measure the time and use the start reference
	#else
	// Bounds check
	if(exec_num_ocl < MAX_EXEC_TIMINGS)
	{
		cl_int err = CL_SUCCESS;
		double exec_time_real;
		if(commandQueue != NULL)
		{
			err=clFinish(*commandQueue);
			print_on_opencl_error("clFinish failed during speex_echo_opencl_store_execution_time_cpu.", err);
		}
		
		// Read the timer value
		gettimeofday(&current_time, NULL);
		exec_time_real = current_time.tv_sec * 1000000 + current_time.tv_usec - start_ref_time_cpu;
		exec_times_cpu[exec_num_cpu] = exec_time_real;
		//printf("Adding exec_time CPU: %f\n", exec_time_real);
		exec_num_cpu++;
	}
	#endif
}

/**
 * Simple bubblesort algorithm used to sort timing values
 * Do not use to sort large arrays, or generally if speed matters...
 * http://www.algorithmist.com/index.php/Bubble_sort.c
 */
void bubble_sort(double *numbers, int array_size)
{
	int i, j;
	double temp;

	for (i = (array_size - 1); i >= 0; i--)
	{
		for (j = 1; j <= i; j++)
		{
			if (numbers[j-1] > numbers[j])
			{
				temp = numbers[j-1];
				numbers[j-1] = numbers[j];
				numbers[j] = temp;
			}
		}
	}
}

/**
 * Print execution time stats (run after all iterations are finished)
 * This functions executes regardless of debug level (only adds overhead after timings are done)
 */
EXPORT void speex_echo_opencl_print_execution_time_stats()
{
	int i;
	int cutoff_tail = 2;	// Number of elements to cut off either end of the array (warm-up etc.)
	double average = 0.0;
	double median;
	double min_value, max_value;
	double variance = 0.0;
	double stddev;
	
	/*printf("Unsorted run times with OpenCL:\n");
	for(i=0;i<exec_num_ocl_sorted;i++)
	{
		printf("%d: %f\t", i, exec_times_ocl_sorted[i]);
		if(i % 5 == 0)
			printf("\n");
	}
	printf("\n");*/
		
	if(exec_times_ocl != NULL)
	{
		// Only print average if run 10 iterations or more
		if(exec_num_ocl > 10)
		{
			// Make sorted copies to find median
			int exec_num_ocl_sorted = exec_num_ocl - 2*cutoff_tail;
			double *exec_times_ocl_sorted = (double *) malloc(sizeof(double) * exec_num_ocl_sorted);
			for(i=0;i<exec_num_ocl_sorted;i++)
			{
				exec_times_ocl_sorted[i] = exec_times_ocl[i + cutoff_tail];
			}
			// Use bubble sort because of simplicity
			// Performance is not as important as it is only run once, and a limited number of timings are used
			bubble_sort(exec_times_ocl_sorted, exec_num_ocl_sorted);
			
			min_value = exec_times_ocl[cutoff_tail];
			max_value = exec_times_ocl[cutoff_tail];
			// If even number of elements, take average
			if(exec_num_ocl_sorted % 2 == 0)
				median = (exec_times_ocl_sorted[exec_num_ocl_sorted/2 - 1] + exec_times_ocl_sorted[exec_num_ocl_sorted/2]) / 2;
			else
				median = exec_times_ocl_sorted[exec_num_ocl_sorted/2];
			
			// Cut off two values at either end
			for(i=cutoff_tail;i<exec_num_ocl-cutoff_tail;i++)
			{
				if(exec_times_ocl[i] < min_value)
					min_value = exec_times_ocl[i];
				if(exec_times_ocl[i] > max_value)
					max_value = exec_times_ocl[i];
				average += exec_times_ocl[i];
			}
			average = average / (exec_num_ocl - 2*cutoff_tail);
			// Pass 2, to calculate variance
			for(i=cutoff_tail;i<exec_num_ocl-cutoff_tail;i++)
			{
				variance += (exec_times_ocl[i] - average) * (exec_times_ocl[i] - average);
			}
			stddev = sqrt(variance / (exec_num_ocl - 2*cutoff_tail));
			
			printf("Average execution time for the OpenCL filter was %f (median: %f, min: %f, max: %f, stddev: %f, N: %d)\n", average, median, min_value, max_value, stddev, exec_num_ocl - cutoff_tail*2);
			free(exec_times_ocl_sorted);
		}
		else
			printf("The OpenCL filter was not run enough time to compute an average.\n");
	}
	
	average = 0.0;
	variance = 0.0;
	stddev = 0.0;
	
	if(exec_times_cpu != NULL)
	{
		// Only print average if run 10 iterations or more
		if(exec_num_cpu > 10)
		{
			int exec_num_cpu_sorted = exec_num_cpu - 2*cutoff_tail;
			double *exec_times_cpu_sorted = (double *) malloc(sizeof(double) * exec_num_cpu_sorted);
			for(i=0;i<exec_num_cpu_sorted;i++)
			{
				exec_times_cpu_sorted[i] = exec_times_cpu[i + cutoff_tail];
			}
			bubble_sort(exec_times_cpu_sorted, exec_num_cpu_sorted);
			
			min_value = exec_times_cpu[cutoff_tail];
			max_value = exec_times_cpu[cutoff_tail];
			// If even number of elements, take average
			if(exec_num_cpu_sorted % 2 == 0)
				median = (exec_times_cpu_sorted[exec_num_cpu_sorted/2 - 1] + exec_times_cpu_sorted[exec_num_cpu_sorted/2]) / 2;
			else
				median = exec_times_cpu_sorted[exec_num_cpu_sorted/2];
			// Cut off two values at either end
			for(i=cutoff_tail;i<exec_num_cpu-cutoff_tail;i++)
			{
				if(exec_times_cpu[i] < min_value)
					min_value = exec_times_cpu[i];
				if(exec_times_cpu[i] > max_value)
					max_value = exec_times_cpu[i];
				average += exec_times_cpu[i];
			}
			average = average / (exec_num_cpu - 2*cutoff_tail);
			for(i=cutoff_tail;i<exec_num_cpu-cutoff_tail;i++)
			{
				variance += (exec_times_cpu[i] - average) * (exec_times_cpu[i] - average);
			}
			stddev = sqrt(variance / (exec_num_cpu - 2*cutoff_tail));
			
			printf("Average execution time for the CPU Speex filter was %f (median: %f, min: %f, max: %f, stddev: %f, N: %d)\n", average, median, min_value, max_value, stddev, exec_num_cpu - 2*cutoff_tail);
			free(exec_times_cpu_sorted);
		}
		else
			printf("The CPU filter was not run enough time to compute an average.\n");
	}
}

EXPORT void speex_echo_opencl_execution_time_init()
{
	int i;
	exec_times_ocl = (double *) malloc(sizeof(double) * MAX_EXEC_TIMINGS);
	exec_times_cpu = (double *) malloc(sizeof(double) * MAX_EXEC_TIMINGS);
	for(i=0;i<MAX_EXEC_TIMINGS;i++)
	{
		exec_times_ocl[i] = 0.0;
		exec_times_cpu[i] = 0.0;
	}
	exec_num_ocl = 0;
	exec_num_cpu = 0;
}

EXPORT void speex_echo_opencl_execution_time_cleanup()
{
	//if(exec_times_ocl != NULL)
	//	free(exec_times_ocl);
	//if(exec_times_cpu != NULL)
	//	free(exec_times_cpu);
}

#if SPEEX_OCL_DEBUG_LEVEL >= 2
//#if 0
/**
 * Checkpoint system for equiavalent functions
 */
EXPORT void init_checkpoints()
{
	int i;
	checkpoint_data = (float **) malloc(sizeof(float *) * MAX_NUM_CHECKPOINTS);
	checkpoint_data_short = (short **) malloc(sizeof(short *) * MAX_NUM_CHECKPOINTS);
	checkpoint_sizes = (int *) malloc(sizeof(int) * MAX_NUM_CHECKPOINTS);	

	// All checkpoint initialized to zero size
	// If the wrong index is given, will not segfault
	for(i = 0;i < MAX_NUM_CHECKPOINTS;i++)
	{
		checkpoint_sizes[i] = 0;
		checkpoint_data[i] = NULL;
		checkpoint_data_short[i] = NULL;
	}
}

void save_new_checkpoint(int index, const float *data, int size)
{
	int i;
	if(checkpoint_sizes[index] != NULL)
		free(checkpoint_data[index]);
	checkpoint_data[index] = (float *) malloc(sizeof(float) * size);
	for(i = 0;i < size;i++)
	{
		checkpoint_data[index][i] = data[i];
	}
	checkpoint_sizes[index] = size;
}

void save_new_checkpoint_short(int index, const short *data, int size)
{
	int i;
	if(checkpoint_sizes[index] != NULL)
		free(checkpoint_data[index]);
	checkpoint_data_short[index] = (float *) malloc(sizeof(short) * size);
	for(i = 0;i < size;i++)
	{
		checkpoint_data_short[index][i] = data[i];
	}
	checkpoint_sizes[index] = size;
}

/**
 * Returns 1 if the given data corresponds to the given checkpoint index
 */
void check_against_checkpoint(int index, const void *data)
{
	int i;
	int is_float_checkpoint = 1;
	int is_the_same = 1;
	int first_error = -1;
	float first_float_diff = 0.0f;
	float float_diff;
	short first_short_diff = 0;
	short short_diff;
	char *print_message = (char *) malloc(sizeof(char) * 5000);
	float *float_data = NULL;
	short *short_data = NULL;
	
	// Temporary host array
	if(checkpoint_data[index] == NULL)
		is_float_checkpoint = 0;
	
	if(is_float_checkpoint == 1)
		float_data = (float *)data;
	else
		short_data = (short *)data;
	
	for(i = 0;i < checkpoint_sizes[index];i++)
	{		
		if(is_float_checkpoint == 1)
		{
			float_diff = checkpoint_data[index][i] - float_data[i];
			// Check for NaN values, if found, difference is set to a large value
			if((isnan(checkpoint_data[index][i]) && !isnan(float_data[i])) ||
				(!isnan(checkpoint_data[index][i]) && isnan(float_data[i])))
			{
				float_diff = 10000.0f;
			}
		}
		else
			short_diff = checkpoint_data_short[index][i] - short_data[i];
		
		if(checkpoint_sizes[index] > 1) 
		{
			if(is_float_checkpoint == 1 && (float_diff > 0.0001 || float_diff < -0.0001)) {
				if(first_error < 0) 
				{
					first_error = i;	
					first_float_diff = float_diff;
				}
				is_the_same = 0;
			}
			else if(is_float_checkpoint == 0 && short_diff != 0)
			{
				if(first_error < 0) 
				{
					first_error = i;
					first_short_diff = short_diff;
				}
				is_the_same = 0;
			}
		}
		// Tolerate larger error for single values (often aggregated values)
		else 
		{
			if(is_float_checkpoint == 1 && (float_diff > 0.01 || float_diff < -0.01)) 
			{
				if(first_error < 0) {
					first_error = i;	
					first_float_diff = float_diff;
				}
				is_the_same = 0;
			}
			// Still do not tolerate any error for shorts
			else if(is_float_checkpoint == 0 && short_diff != 0)
			{
				if(first_error < 0) {
					first_error = i;	
					first_short_diff = short_diff;
				}
				is_the_same = 0;
			}
		}
	}	
	if(is_the_same == 1)
	{
		sprintf(print_message, "Checkpoint %d correctly validated", index);
		print_debug(print_message, 3);
	}
	else
	{
		sprintf(print_message, "*** Checkpoint %d failed ***", index);
		print_error(print_message, 0);
		
		sprintf(print_message, "Printout of reference data checkpoint %d", index);
		print_error(print_message, 0);
		//printf("Printout of reference data checkpoint %d", index);

		for(i=0;i<checkpoint_sizes[index];i++)
		{
			if(is_float_checkpoint == 1)
				printf("%d: %2.7f\t", i, checkpoint_data[index][i]);	
			else
				printf("%d: %d\t", i, checkpoint_data_short[index][i]);	
			if(i % 5 == 0)
				printf("\n");
		}

		if(is_float_checkpoint == 1)
			sprintf(print_message, "\n\nPrintout of erronous data checkpoint %d (first error found at index %d with the error %f)", index, first_error, first_float_diff);
		else
			sprintf(print_message, "\n\nPrintout of erronous data checkpoint %d (first error found at index %d with the error %d)", index, first_error, first_short_diff);

		print_error(print_message, 0);

		for(i=0;i<checkpoint_sizes[index];i++)
		{
			if(is_float_checkpoint == 1)
				printf("%d: %2.7f\t", i, float_data[i]);	
			else
				printf("%d: %d\t", i, short_data[i]);	
			if(i % 5 == 0)
				printf("\n");
		}
		printf("\n");
	}

	free(print_message);
}

/**
 * Copies an OpenCL array of floats/shorts to temporary host memory and checks it
 * Returns 1 if the given data corresponds to the given checkpoint index
 */
void check_against_checkpoint_opencl(int index, cl_mem cl_data, size_t cl_data_offset, int input_is_int)
{
	int i;
	int is_float_checkpoint = 1;
	int is_the_same = 1;
	int first_error = -1;
	float first_float_diff = 0.0f;
	float float_diff;
	short first_short_diff = 0;
	short short_diff = 0;
	int int_diff = 0;
	int first_int_diff = 0;
	char *print_message = (char *) malloc(sizeof(char) * 2000);
	cl_int err = CL_SUCCESS;
	float *float_data = NULL;
	short *short_data = NULL;
	int *int_data = NULL;
	
	// Put in a clFinish to make sure that state is consitent
	err=clFinish(*commandQueue);
        print_on_opencl_error("clFinish failed during check_against_checkpoint_opencl.", err);
	
	// Temporary host array
	if(checkpoint_data[index] == NULL)
		is_float_checkpoint = 0;
	
	if(is_float_checkpoint == 1)
		float_data = (float *)malloc(sizeof(float) * checkpoint_sizes[index]);
	else if(input_is_int == 0)
		short_data = (short *)malloc(sizeof(short) * checkpoint_sizes[index]);
	else
		int_data = (int *)malloc(sizeof(int) * checkpoint_sizes[index]);

	if(cl_data == NULL)
	{
		sprintf(print_message, "*** Checkpoint %d failed, non-existent OpenCL data ***", index);
		print_error(print_message, 0);
	}

	if(is_float_checkpoint == 1)
	{	
		err = clEnqueueReadBuffer(*commandQueue, cl_data, CL_TRUE, sizeof(float)*cl_data_offset, sizeof(float)*checkpoint_sizes[index], float_data, 0, NULL, NULL);
	}
	else if(input_is_int == 0)
	{
		err = clEnqueueReadBuffer(*commandQueue, cl_data, CL_TRUE, sizeof(short)*cl_data_offset, sizeof(short)*checkpoint_sizes[index], short_data, 0, NULL, NULL);
	}
	else
	{
		err = clEnqueueReadBuffer(*commandQueue, cl_data, CL_TRUE, sizeof(int)*cl_data_offset, sizeof(int)*checkpoint_sizes[index], int_data, 0, NULL, NULL);
	}
        print_on_opencl_error("clEnqueueReadBuffer failed when checking checkpoint.", err);
	
	for(i = 0;i < checkpoint_sizes[index];i++)
	{
		if(is_float_checkpoint == 1)
		{
			float_diff = checkpoint_data[index][i] - float_data[i];
			// Check for NaN values, if found, difference is set to a large value
			if((isnan(checkpoint_data[index][i]) && !isnan(float_data[i])) ||
				(!isnan(checkpoint_data[index][i]) && isnan(float_data[i])))
			{
				float_diff = 10000.0f;
			}
		}
		else if(input_is_int == 0)
			short_diff = checkpoint_data_short[index][i] - short_data[i];
		else
			int_diff = checkpoint_data_short[index][i] - int_data[i];
		
		if(checkpoint_sizes[index] > 1) 
		{
			if(is_float_checkpoint == 1 && (float_diff > 0.01 || float_diff < -0.01)) {
				if(first_error < 0) 
				{
					first_error = i;	
					first_float_diff = float_diff;
				}
				is_the_same = 0;
			}
			else if(is_float_checkpoint == 0 && short_diff != 0)
			{
				if(first_error < 0) 
				{
					first_error = i;
					first_short_diff = short_diff;
				}
				is_the_same = 0;
			}
			else if(is_float_checkpoint == 0 && int_diff != 0)
			{
				if(first_error < 0) 
				{
					first_error = i;
					first_int_diff = int_diff;
				}
				is_the_same = 0;
			}
		}
		// Tolerate larger error for single values (often aggregated values)
		else
		{
			if(is_float_checkpoint == 1 && (float_diff > 0.1 || float_diff < -0.1)) 
			{
				if(first_error < 0) {
					first_error = i;	
					first_float_diff = float_diff;
				}
				is_the_same = 0;
			}
			// Still do not tolerate any error for shorts
			else if(is_float_checkpoint == 0 && short_diff != 0)
			{
				if(first_error < 0) {
					first_error = i;	
					first_short_diff = short_diff;
				}
				is_the_same = 0;
			}
			else if(is_float_checkpoint == 0 && int_diff != 0)
			{
				if(first_error < 0) {
					first_error = i;	
					first_int_diff = int_diff;
				}
				is_the_same = 0;
			}
		}
	}
	if(is_the_same == 1)
	{
		sprintf(print_message, "Checkpoint %d correctly validated", index);
		print_debug(print_message, 3);
	}
	else
	{
		sprintf(print_message, "*** Checkpoint %d failed ***", index);
		print_error(print_message, 0);
		
		sprintf(print_message, "Printout of reference data checkpoint %d", index);
		print_error(print_message, 0);

		for(i=0;i<checkpoint_sizes[index];i++)
		{
			if(is_float_checkpoint == 1)
				printf("%d: %2.7f\t", i, checkpoint_data[index][i]);	
			else
				printf("%d: %d\t", i, checkpoint_data_short[index][i]);	
			if(i % 5 == 0)
				printf("\n");
		}

		if(is_float_checkpoint == 1)
			sprintf(print_message, "\n\nPrintout of erronous data checkpoint %d (first error found at index %d with the error %f)", index, first_error, first_float_diff);
		else if(input_is_int == 0)
			sprintf(print_message, "\n\nPrintout of erronous data checkpoint %d (first error found at index %d with the error %d)", index, first_error, first_short_diff);
		else
			sprintf(print_message, "\n\nPrintout of erronous data checkpoint %d (first error found at index %d with the error %d)", index, first_error, first_int_diff);

		print_error(print_message, 0);

		for(i=0;i<checkpoint_sizes[index];i++)
		{
			if(is_float_checkpoint == 1)
				printf("%d: %2.7f\t", i, float_data[i]);	
			else if(input_is_int == 0)
				printf("%d: %d\t", i, short_data[i]);	
			else
				printf("%d: %d\t", i, int_data[i]);	
			if(i % 5 == 0)
				printf("\n");
		}
		printf("\n");
	}

	free(print_message);

	if(float_data != NULL)
		free(float_data);
	if(short_data != NULL)
		free(short_data);
	if(int_data != NULL)
		free(int_data);
}

EXPORT void cleanup_checkpoints()
{
	int i;
	if(checkpoint_sizes != NULL)
		free(checkpoint_sizes);
	
	for(i = 0;i < MAX_NUM_CHECKPOINTS;i++)
	{
		if(checkpoint_data[i] != NULL)
		{
			free(checkpoint_data[i]);
		}
		if(checkpoint_data_short[i] != NULL)
		{
			free(checkpoint_data_short[i]);
		}
	}
	if(checkpoint_data != NULL)
		free(checkpoint_data);
	if(checkpoint_data_short != NULL)
		free(checkpoint_data_short);
}
#else
	EXPORT void init_checkpoints() {}
	void save_new_checkpoint(int index, const float *data, int size) {}
	void save_new_checkpoint_short(int index, const short *data, int size) {}
	void check_against_checkpoint(int index, const void *data) {}
	void check_against_checkpoint_opencl(int index, cl_mem cl_data, size_t cl_data_offset, int input_is_int) {}
	EXPORT void cleanup_checkpoints() {}
#endif

struct fftw_config2 {
  float *in;
  float *out;
  fftwf_plan fft;
  fftwf_plan ifft;
  int N;
};


/**
 * Copies an echo state, for testing two MDF filters against eachother
 */
EXPORT void copy_echo_state(SpeexEchoState *input, SpeexEchoState *output)
{
        // Copy the variables
        //memcpy((void *)output, (void *)input, sizeof(SpeexEchoState));

	output->frame_size = input->frame_size;
	output->window_size = input->window_size;
	output->M = input->M;
	output->cancel_count = input->cancel_count;
	output->adapted = input->adapted;
	output->saturated = input->saturated;
	output->screwed_up = input->screwed_up;
	output->C = input->C;
	output->K = input->K;
	output->sampling_rate = input->sampling_rate;
	output->spec_average = input->spec_average;
	output->beta0 = input->beta0;
	output->beta_max = input->beta_max;
	output->sum_adapt = input->sum_adapt;
	output->leak_estimate = input->leak_estimate;
	output->Davg1 = input->Davg1;
	output->Davg2 = input->Davg2;
	output->Dvar1 = input->Dvar1;
	output->Dvar2 = input->Dvar2;
	output->Pey = input->Pey;
	output->Pyy = input->Pyy;
	output->preemph = input->preemph;
	output->notch_radius = input->notch_radius;
	output->play_buf_pos = input->play_buf_pos;
	output->play_buf_started = input->play_buf_started;

        // Copy the arrays	
	// Do not copy the FFT, since we dont need it in the copy used with OpenCL
	//struct fftw_config2 *input_table = input->fft_table;
	//struct fftw_config2 *output_table = output->fft_table;
	//output_table->N = input_table->N;

	//output->fft_table = spx_fft_init(output->window_size);

	memcpy(output->e, input->e, input->C*input->window_size*sizeof(spx_word16_t));
	memcpy(output->x, input->x, input->K*input->window_size*sizeof(spx_word16_t));
	memcpy(output->input, input->input, input->C*input->frame_size*sizeof(spx_word16_t));
	memcpy(output->y, input->y, input->C*input->window_size*sizeof(spx_word16_t));
	memcpy(output->last_y, input->last_y, input->C*input->window_size*sizeof(spx_word16_t));

	memcpy(output->Yf, input->Yf, (input->frame_size+1)*sizeof(spx_word32_t));
	memcpy(output->Rf, input->Rf, (input->frame_size+1)*sizeof(spx_word32_t));
	memcpy(output->Xf, input->Xf, (input->frame_size+1)*sizeof(spx_word32_t));
	memcpy(output->Yh, input->Yh, (input->frame_size+1)*sizeof(spx_word32_t));
	memcpy(output->Eh, input->Eh, (input->frame_size+1)*sizeof(spx_word32_t));

	memcpy(output->X, input->X, input->K*(input->M+1)*input->window_size*sizeof(spx_word16_t));
	memcpy(output->Y, input->Y, input->C*input->window_size*sizeof(spx_word16_t));
	memcpy(output->E, input->E, input->C*input->window_size*sizeof(spx_word16_t));
	memcpy(output->W, input->W, input->C*input->K*input->M*input->window_size*sizeof(spx_word32_t));
	
//#ifdef TWO_PATH
	memcpy(output->foreground, input->foreground, input->M*input->window_size*input->C*input->K*sizeof(spx_word16_t));
//#endif
	memcpy(output->PHI, input->PHI, input->window_size*sizeof(spx_word32_t));
	memcpy(output->power, input->power, (input->frame_size+1)*sizeof(spx_word32_t));
	memcpy(output->power_1, input->power_1, (input->frame_size+1)*sizeof(spx_float_t));
	memcpy(output->window, input->window, input->window_size*sizeof(spx_word16_t));
	memcpy(output->prop, input->prop, input->M*sizeof(spx_word16_t));
	memcpy(output->wtmp, input->wtmp, input->window_size*sizeof(spx_word16_t));

	memcpy(output->memX, input->memX, input->K*sizeof(spx_word16_t));
	memcpy(output->memD, input->memD, input->C*sizeof(spx_word16_t));
	memcpy(output->memE, input->memE, input->C*sizeof(spx_word16_t));

	memcpy(output->notch_mem, input->notch_mem, 2*input->C*sizeof(spx_mem_t));
	memcpy(output->play_buf, input->play_buf, input->K*(1)*input->frame_size*sizeof(spx_word16_t));
}

/*
 * Returns filesize of the file with the given name
 */
int getFileSize(const char *filename)
{
        FILE *fp = NULL;
        int size = 0;

        if((fp = fopen(filename, "r"))!=NULL) {
                fseek(fp, 0, SEEK_END);
                size = ftell(fp);
                return size;
        }
        return -1;
}

/*
 * Converts the contents of a file into a string
 * Length is an output parameter
 */
int convertToString(const char *filename, char *str, size_t *length)
{
        FILE *fp = NULL;
        int size = 0;
        int i = 0;

        if((fp = fopen(filename, "rb"))!=NULL) {
                fseek(fp, 0, SEEK_END);
                size = ftell(fp);
                fseek(fp, 0, SEEK_SET);
                printf("The size of the kernel source file: %d\n", size);
                *length = size;
                if(size > 0) {
                        for(i=0;i<size;i++) {
                                str[i] = fgetc(fp);
                        }
                        str[size] = '\0';
                        //printf("Read file, returning the following string:\n%s\n", str);
                        fclose(fp);
                        return 0;
                }
                fclose(fp);
                return -2;
        }
        return -1;
}

/**
 * Prints useful information from the compilation of the given program
 */
void print_opencl_build_info(cl_program *program)
{
        cl_int err = CL_SUCCESS;
        unsigned int i;
        cl_uint numProgramDevices;
        cl_device_id *dIds;
        cl_build_status buildStatus;
        char *buildLog;
        err = clGetProgramInfo(*program, CL_PROGRAM_NUM_DEVICES, sizeof(cl_uint), &numProgramDevices, NULL);
        if(err != CL_SUCCESS)
        {
                printf("OPENCL BUILD INFO: clGetProgramInfo of program failed\n");
                print_opencl_error(err);
                return;
        }
        else
        {
                printf("OPENCL_BUILD_INFO: Number of devices associated with the program: %d\n", (unsigned int)numProgramDevices);
        }
        dIds = (cl_device_id *)malloc(sizeof(cl_device_id) * numProgramDevices);

        err = clGetProgramInfo(*program, CL_PROGRAM_DEVICES, sizeof(cl_device_id) * numProgramDevices, dIds, NULL);
        if(err != CL_SUCCESS)
        {
                printf("OPENCL BUILD INFO: clGetProgramInfo of program failed\n");
                print_opencl_error(err);
                return;
        }

        buildLog = (char *)malloc(sizeof(char) * 6000);

	for(i=0;i<numProgramDevices;i++)
        {
                err = clGetProgramBuildInfo(*program, dIds[i], CL_PROGRAM_BUILD_STATUS, sizeof(cl_build_status), &buildStatus, NULL);
                if(err != CL_SUCCESS)
                {
                        printf("OPENCL_BUILD_INFO: clGetProgramBuildInfo of program failed for device %d\n", i);
                        print_opencl_error(err);
                }
                else
                {
                        printf("OPENCL_BUILD_INFO: Build status of the program on device %d: ", i);
                        if(buildStatus == CL_BUILD_NONE)
                                printf("CL_BUILD_NONE");
                        else if(buildStatus == CL_BUILD_ERROR)
                                printf("CL_BUILD_ERROR");
                        else if(buildStatus == CL_BUILD_SUCCESS)
                                printf("CL_BUILD_SUCCESS");
                        else if(buildStatus == CL_BUILD_IN_PROGRESS)
                                printf("CL_BUILD_IN_PROGRESS");
                        else
                                printf("Undefined build code returned.");
                        printf("\n");
                }
                err = clGetProgramBuildInfo(*program, dIds[i], CL_PROGRAM_BUILD_LOG, sizeof(char) * 5000, buildLog, NULL);
                if(err != CL_SUCCESS)
                {
			// ATI returns error code on successful build and no messages
			if(buildStatus == CL_BUILD_SUCCESS && err == CL_INVALID_VALUE)
			{
				printf("OPENCL_BUILD_INFO: No build info available for device %d\n", i);
			}
			else
			{
				printf("OPENCL_BUILD_INFO: clGetProgramBuildInfo of program failed for device %d\n", i);
				print_opencl_error(err);
			}
                }
                else
                {
                        printf("OPENCL_BUILD_INFO: Build log for the program on device %d: \n%s\n", i, buildLog);
                }
        }
        free(dIds);
}

/**
 * Outputs OpenCL errorcodes to console
 */
void print_opencl_error(cl_int error)
{
        if(error == CL_INVALID_COMMAND_QUEUE)
                printf("CL_INVALID_COMMAND_QUEUE\n");
        else if(error == CL_INVALID_CONTEXT)
                printf("CL_INVALID_CONTEXT\n");
        else if(error == CL_INVALID_MEM_OBJECT)
                printf("CL_INVALID_MEM_OBJECT\n");
        else if(error == CL_INVALID_VALUE)
                printf("CL_INVALID_VALUE\n");
        else if(error == CL_INVALID_EVENT_WAIT_LIST)
                printf("CL_INVALID_EVENT_WAIT_LIST\n");
        else if(error == CL_MEM_OBJECT_ALLOCATION_FAILURE)
                printf("CL_MEM_OBJECT_ALLOCATION_FAILURE\n");
        else if(error == CL_OUT_OF_HOST_MEMORY)
                printf("CL_OUT_OF_HOST_MEMORY\n");
        else if(error == CL_INVALID_PROGRAM_EXECUTABLE)
                printf("CL_INVALID_PROGRAM_EXECUTABLE\n");
        else if(error == CL_INVALID_COMMAND_QUEUE)
                printf("CL_INVALID_COMMAND_QUEUE\n");
        else if(error == CL_INVALID_KERNEL)
                printf("CL_INVALID_KERNEL\n");
	else if(error == CL_INVALID_KERNEL_NAME)
                printf("CL_INVALID_KERNEL_NAME\n");
	else if(error == CL_INVALID_KERNEL_DEFINITION)
                printf("CL_INVALID_KERNEL_DEFINITION\n");
        else if(error == CL_INVALID_KERNEL_ARGS)
                printf("CL_INVALID_KERNEL_ARGS\n");
        else if(error == CL_INVALID_WORK_DIMENSION)
                printf("CL_INVALID_WORK_DIMENSION\n");
        else if(error == CL_INVALID_WORK_GROUP_SIZE)
                printf("CL_INVALID_WORK_GROUP_SIZE\n");
        else if(error == CL_INVALID_WORK_ITEM_SIZE)
                printf("CL_INVALID_WORK_ITEM_SIZE\n");
        else if(error == CL_INVALID_GLOBAL_OFFSET)
                printf("CL_INVALID_GLOBAL_OFFSET\n");
        else if(error == CL_OUT_OF_RESOURCES)
                printf("CL_OUT_OF_RESOURCES\n");
        else if(error == CL_MEM_OBJECT_ALLOCATION_FAILURE)
                printf("CL_MEM_OBJECT_ALLOCATION_FAILURE\n");
        else if(error == CL_INVALID_EVENT_WAIT_LIST)
                printf("CL_INVALID_EVENT_WAIT_LIST\n");
        else if(error == CL_OUT_OF_HOST_MEMORY)
                printf("CL_OUT_OF_HOST_MEMORY\n");
        else if(error == CL_INVALID_PROGRAM)
                printf("CL_INVALID_PROGRAM\n");
        else if(error == CL_INVALID_DEVICE)
                printf("CL_INVALID_DEVICE\n");
        else if(error == CL_INVALID_BINARY)
                printf("CL_INVALID_BINARY\n");
        else if(error == CL_INVALID_BUILD_OPTIONS)
                printf("CL_INVALID_BUILD_OPTIONS\n");
        else if(error == CL_INVALID_OPERATION)
                printf("CL_INVALID_OPERATION\n");
        else if(error == CL_COMPILER_NOT_AVAILABLE)
                printf("CL_COMPILER_NOT_AVAILABLE\n");
        else if(error == CL_BUILD_PROGRAM_FAILURE)
                printf("CL_BUILD_PROGRAM_FAILURE\n");
	else if(error == CL_INVALID_ARG_INDEX)
                printf("CL_INVALID_ARG_INDEX\n");
	else if(error == CL_INVALID_ARG_VALUE)
                printf("CL_INVALID_ARG_VALUE\n");
	else if(error == CL_INVALID_MEM_OBJECT)
                printf("CL_INVALID_MEM_OBJECT\n");
	else if(error == CL_INVALID_ARG_SIZE)
                printf("CL_INVALID_ARG_SIZE\n");
        else
                printf("print_opencl_error stumbled on an unknown error code: %d\n", error);
}
